/*
 * Atmel maXTouch Touchscreen driver
 *
 * Copyright (C) 2010 Samsung Electronics Co.Ltd
 * Copyright (C) 2011-2012 Atmel Corporation
 * Copyright (C) 2012 Google, Inc.
 *
 * Author: Joonyoung Shim <jy0922.shim@samsung.com>
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 *
 */
/* #define DEBUG */
#include <linux/module.h>
#include <linux/init.h>
#include <linux/completion.h>
#include <linux/delay.h>
#include <linux/firmware.h>
#include <linux/i2c.h>
#include <linux/input/mt.h>
#include <linux/interrupt.h>
#include <linux/slab.h>
#include <linux/timer.h>
#include <linux/regulator/consumer.h>
#include <linux/gpio.h>
#include <linux/pinctrl/consumer.h>
#include <linux/semaphore.h>
#include <linux/atomic.h>
#include <asm/uaccess.h>
#include <linux/fs.h>
#include <linux/file.h>
#include <linux/notifier.h>
#include <linux/kfifo.h>
#include <linux/jiffies.h>
#include "./atmel_mxt_ts.h"
#include "../../../../kernel/msm-4.4/drivers/video/fbdev/msm/mdss_panel.h"

enum {
	STATE_UNKNOWN,
	STATE_ACTIVE,
	STATE_SUSPEND,
	STATE_STANDBY = 4,
	STATE_BL,
	STATE_INIT,
	STATE_FLASH,
	STATE_QUERY,
	STATE_INVALID
};

#define PROC_COORD_EX

#define USE_REQUEST_FIRMWARE 0

#define STATE_UI	(STATE_ACTIVE | STATE_SUSPEND)

#ifdef CONFIG_OF
#include <linux/of.h>
#include <linux/of_gpio.h>
#endif

#ifdef CONFIG_FB
#include <linux/notifier.h>
#include <linux/fb.h>
#include <linux/workqueue.h>

static int fb_notifier_callback(struct notifier_block *self,
		unsigned long event, void *data);
#endif

static int nfc_notifier_callback(struct notifier_block *this,
		unsigned long event, void *ptr);

extern int register_nfc_notifier(struct notifier_block*);
extern int unregister_nfc_notifier(struct notifier_block*);

static bool istpenable;

struct kfifo  nfc_event_fifo ;

enum{
	TP_DISABLE,
	TP_ENABLE
};

#define PINCTRL_STATE_ACTIVE	"mxt_ts_active"
#define PINCTRL_STATE_SUSPEND	"mxt_ts_suspend"

#define DRIVER_NAME "atmel_mxt_ts"
#define MXT_MAX_RETRIES		10
#define MXT_MAX_BLOCK_WRITE	256
#define MXT_MAX_BUTTONS		8
#define DEBUG_MSG_MAX		200

/* Configuration file */
#define MXT_CFG_MAGIC		"OBP_RAW V1"
/* Registers */
#define MXT_OBJECT_START	0x07
#define MXT_OBJECT_SIZE		6
#define MXT_INFO_CHECKSUM_SIZE	3
#define MXT_T37_REVISION	21

/* Object types */
#define MXT_DEBUG_DIAGNOSTIC_T37	37
#define MXT_GEN_MESSAGE_T5		5
#define MXT_GEN_COMMAND_T6		6
#define MXT_GEN_POWER_T7		7
#define MXT_GEN_ACQUIRE_T8		8
#define MXT_GEN_DATASOURCE_T53		53
#define MXT_TOUCH_MULTI_T9		9
#define MXT_TOUCH_KEYARRAY_T15		15
#define MXT_TOUCH_PROXIMITY_T23		23
#define MXT_TOUCH_PROXKEY_T52		52
#define MXT_PROCI_GRIPFACE_T20		20
#define MXT_PROCG_NOISE_T22		22
#define MXT_PROCI_ONETOUCH_T24		24
#define MXT_PROCI_TWOTOUCH_T27		27
#define MXT_PROCI_GRIP_T40		40
#define MXT_PROCI_PALM_T41		41
#define MXT_PROCI_TOUCHSUPPRESSION_T42	42
#define MXT_PROCI_STYLUS_T47		47
#define MXT_PROCG_NOISESUPPRESSION_T48	48
#define MXT_SPT_COMMSCONFIG_T18		18
#define MXT_SPT_GPIOPWM_T19		19
#define MXT_SPT_SELFTEST_T25		25
#define MXT_SPT_CTECONFIG_T28		28
#define MXT_SPT_USERDATA_T38		38
#define MXT_SPT_DIGITIZER_T43		43
#define MXT_SPT_MESSAGECOUNT_T44	44
#define MXT_SPT_CTECONFIG_T46		46
#define MXT_SPT_TOUCHEVENTTRIGGER_T79	79
#define MXT_PROCI_ACTIVE_STYLUS_T63	63
#define MXT_PROCI_LENSEBENDING_T65	65
#define MXT_PROCG_NOISESUPPRESSION_T72	72
#define MXT_PROCI_TOUCHSEQUENCELOGGER_T93 93
#define MXT_TOUCH_MULTITOUCHSCREEN_T100 100
#define MXT_PROCI_GLOVEDETECTION_T78	78
#define MXT_SPT_TOUCHSCREENHOVER_T101	101
#define MXT_PROCI_ACTIVESTYLUS_T107	107
#define MXT_PROCI_RETRANSMISSIONCOMPENSATION_T80	80
#define MXT_PROCG_NOISESUPSELFCAT_T108	108
#define MXT_SPT_DYNAMICCONFIGURATIONCONTROLLER_T70	70
#define MXT_SPT_SELFCAPCONFIG_T111	111

/* MXT_GEN_MESSAGE_T5 object */
#define MXT_RPTID_NOMSG		0xff

/* MXT_GEN_COMMAND_T6 field */
#define MXT_COMMAND_RESET	0
#define MXT_COMMAND_BACKUPNV	1
#define MXT_COMMAND_CALIBRATE	2
#define MXT_COMMAND_REPORTALL	3
#define MXT_COMMAND_DIAGNOSTIC	5

/* Define for T6 status byte */
#define MXT_T6_STATUS_RESET	(1 << 7)
#define MXT_T6_STATUS_OFL	(1 << 6)
#define MXT_T6_STATUS_SIGERR	(1 << 5)
#define MXT_T6_STATUS_CAL	(1 << 4)
#define MXT_T6_STATUS_CFGERR	(1 << 3)
#define MXT_T6_STATUS_COMSERR	(1 << 2)
#define MXT_T6_STATUS_DEVERR	\
	(MXT_T6_STATUS_SIGERR|MXT_T6_STATUS_CFGERR|MXT_T6_STATUS_COMSERR)

/* MXT_GEN_POWER_T7 field */
struct t7_config {
	u8 idle;
	u8 active;
	u8 actv2idle;
} __packed;

#define MXT_POWER_CFG_RUN		0
#define MXT_POWER_CFG_DEEPSLEEP		1
#define MXT_POWER_CFG_RUN_FAST		2

/* MXT_TOUCH_MULTI_T9 field */
#define MXT_T9_ORIENT		9
#define MXT_T9_RANGE		18

/* MXT_TOUCH_MULTI_T9 status */
#define MXT_T9_UNGRIP		(1 << 0)
#define MXT_T9_SUPPRESS		(1 << 1)
#define MXT_T9_AMP		(1 << 2)
#define MXT_T9_VECTOR		(1 << 3)
#define MXT_T9_MOVE		(1 << 4)
#define MXT_T9_RELEASE		(1 << 5)
#define MXT_T9_PRESS		(1 << 6)
#define MXT_T9_DETECT		(1 << 7)

struct t9_range {
	u16 x;
	u16 y;
} __packed;

/* MXT_TOUCH_MULTI_T9 orient */
#define MXT_T9_ORIENT_SWITCH	(1 << 0)

/* MXT_SPT_COMMSCONFIG_T18 */
#define MXT_COMMS_CTRL		0
#define MXT_COMMS_CMD		1
#define MXT_COMMS_RETRIGEN      (1 << 6)

/* Define for MXT_GEN_COMMAND_T6 */
#define MXT_BOOT_VALUE		0xa5
#define MXT_RESET_VALUE		0x01
#define MXT_BACKUP_VALUE	0x55
#define MXT_IDENT_VALUE		0x80

/* Define for MXT_PROCI_TOUCHSUPPRESSION_T42 */
#define MXT_T42_MSG_TCHSUP	(1 << 0)

/* T47 Stylus */
#define MXT_TOUCH_MAJOR_T47_STYLUS	1

/* T63 Stylus */
#define MXT_T63_STYLUS_PRESS	(1 << 0)
#define MXT_T63_STYLUS_RELEASE	(1 << 1)
#define MXT_T63_STYLUS_MOVE		(1 << 2)
#define MXT_T63_STYLUS_SUPPRESS	(1 << 3)

#define MXT_T63_STYLUS_DETECT	(1 << 4)
#define MXT_T63_STYLUS_TIP		(1 << 5)
#define MXT_T63_STYLUS_ERASER	(1 << 6)
#define MXT_T63_STYLUS_BARREL	(1 << 7)

#define MXT_T63_STYLUS_PRESSURE_MASK	0x3F

/* T100 Multiple Touch Touchscreen */
#define MXT_T100_CTRL		0
#define MXT_T100_CFG1		1
#define MXT_T100_TCHAUX		3
#define MXT_T100_NUMTCH		6
#define MXT_T100_XSIZE		9
#define MXT_T100_XPITCH		10
#define MXT_T100_XRANGE		13
#define MXT_T100_YSIZE		20
#define MXT_T100_YPITCH		21
#define MXT_T100_YRANGE		24
#define MXT_T100_GAIN		28
#define MXT_T100_TCHTHR		30
#define MXT_T100_TCHTHRHYS	31
#define MXT_T100_INTTHR		32
#define MXT_T100_MOVHYSTI	47
#define MXT_T100_MOVHYSTN	49

#define MXT_T100_CFG_SWITCHXY	(1 << 5)

#define MXT_T100_TCHAUX_VECT	(1 << 0)
#define MXT_T100_TCHAUX_AMPL	(1 << 1)
#define MXT_T100_TCHAUX_AREA	(1 << 2)

#define MXT_T100_DETECT		(1 << 7)
#define MXT_T100_TYPE_MASK		0x70
#define MXT_T100_TYPE_FINGER		0x10
#define MXT_T100_TYPE_PASSIVE_STYLUS	0x20
#define MXT_T100_TYPE_ACTIVE_STYLUS	0x30
#define MXT_T100_TYPE_GLOVE		0x50

#define MXT_T100_ACTIVE_STYLUS_HOVER	(0x01)
#define MXT_T100_ACTIVE_STYLUS_TIPSWITH	(0x02)
#define MXT_T100_ACTIVE_STYLUS_BUTTON0	(0x04)
#define MXT_T100_ACTIVE_STYLUS_BUTTON1	(0x08)

#define MXT_T100_STATE_NONE	(0)
#define MXT_T100_STATE_HOVER	(1)
#define MXT_T100_STATE_TOUCH	(2)

/* T38 User Data */
#define MXT_T38_VENDOR_VER	(4)
#define MXT_T38_VENDOR_REV	(8)
#define MXT_T38_STATUS		(63)
#define MXT_T38_STATUS_DATA	('P')

/* T25 Selftest */
#define MXT_T25_SELFTEST_PIN_FAULT	(0x12)
#define MXT_T25_SELFTEST_SIGNAL_LIMIT	(0x17)
#define MXT_T25_CMD			(1)
#define MXT_T25_SIGNAL_UPPER_LSB	(2)
#define MXT_T25_SIGNAL_UPPER_MSB	(3)
#define MXT_T25_SIGNAL_LOWER_LSB	(4)
#define MXT_T25_SIGNAL_LOWER_MSB	(5)
#define MXT_T25_SIGNAL_RANGE_LSB	(15)
#define MXT_T25_SIGNAL_RANGE_MSB	(16)

/* T107 Stylus */
#define MXT_T107_CTRL	(0)
#define MXT_T107_CFG1	(1)

/* Delay times */
#define MXT_BACKUP_TIME		50	/* msec */
#define MXT_RESET_TIME		200	/* msec */
#define MXT_RESET_TIMEOUT	3000	/* msec */
#define MXT_CRC_TIMEOUT		2000	/* msec */
#define MXT_FW_RESET_TIME	3500	/* msec */
#define MXT_FW_CHG_TIMEOUT	300	/* msec */
#define MXT_WAKEUP_TIME		25	/* msec */
#define MXT_RESET_CTRL_TIME	10	/* msec */
#define MXT_REGULATOR_DELAY	150	/* msec */
#define MXT_POWERON_DELAY	150	/* msec */
#define MXT_HEALTHCHK_INTERVAL	5000    /* msec */
#define MXT_NFCSENING_INTERVAL  400     /* msec */
#define MXT_REBOOT_TIME		1000	/* msec */
#define MXT_SELFTEST_TIMEOUT	1000	/* msec */

/* Command to unlock bootloader */
#define MXT_UNLOCK_CMD_MSB	0xaa
#define MXT_UNLOCK_CMD_LSB	0xdc

/* Bootloader mode status */
#define MXT_WAITING_BOOTLOAD_CMD	0xc0	/* valid 7 6 bit only */
#define MXT_WAITING_FRAME_DATA	0x80	/* valid 7 6 bit only */
#define MXT_FRAME_CRC_CHECK	0x02
#define MXT_FRAME_CRC_FAIL	0x03
#define MXT_FRAME_CRC_PASS	0x04
#define MXT_APP_CRC_FAIL	0x40	/* valid 7 8 bit only */
#define MXT_BOOT_STATUS_MASK	0x3f
#define MXT_BOOT_EXTENDED_ID	(1 << 5)
#define MXT_BOOT_ID_MASK	0x1f

/* Touchscreen absolute values */
#define MXT_MAX_AREA		0x3C
#define MXT_MAX_PRESSURE	0x3ff

/* awaitable flag for suspend function call */
#define MXT_SUSPEND_SYNC	(1)
#define MXT_SUSPEND_ASYNC	(0)

#define MXT_PIXELS_PER_MM	20

#define NFC_MSG_RF_INTF_ACTIVATED 5
#define NFC_MSG_RF_DEACTIVATE 6

#define MXT_DEFAULT_TP_MODE 	0x0001

/*for recognize LCD panel*/
#define TYPEB    "dsi:0:qcom,mdss_dsi_ili9881c_video:1:none:cfg:single_dsi"
#define TYPEA    "dsi:0:qcom,mdss_dsi_otm1282a_cmd:1:none:cfg:single_dsi"

struct mxt_obj_patch {
	u8 number;
	u8 instance;
	u8 offset;
	u8 value;
	struct list_head link;
};

struct mxt_patch {
	const char *name;
	int	cfg_num;
	struct list_head cfg_head;
};


struct mxt_patchset {
	int	patch_num;
	struct mxt_patch *patch_data;
};

struct mxt_object {
	u8 type;
	u16 start_address;
	u8 size_minus_one;
	u8 instances_minus_one;
	u8 num_report_ids;
} __packed;

struct mxt_tdat_section {
	const unsigned char *data;
	size_t size;
};

#define ACTIVE_IDX	0
#define SUSPEND_IDX	1
#define MAX_NUM_STATES	2

#define MAX_VALUE_COUNT 128

struct mxt_tp_mode {
	uint16_t mode_id ;
	uint8_t value[MAX_VALUE_COUNT];
	struct list_head list_node;
};

struct mxt_tp_conf{
	uint16_t conf_register_num;
	uint16_t mode_num;
	uint16_t* address;
	struct list_head mode_conf_list;
};



/* Each client has this additional data */
struct mxt_data {
	struct i2c_client *client;
	struct input_dev *input_dev;
	char phys[64];		/* device physical location */
	struct mxt_platform_data *pdata;
	struct mxt_object *object_table;
	struct mxt_info *info;
	void *raw_info_block;
	atomic_t state;
	unsigned int irq;
	unsigned int max_x;
	unsigned int max_y;
	bool in_bootloader;
	u16 mem_size;
	u32 config_id;
	u32 vendor_config_id;
	u32 vendor_local_rev;
	u32 firmware_id;
	u8 t100_aux_ampl;
	u8 t100_aux_area;
	u8 t100_aux_vect;
	bool mem_access_created;
	struct bin_attribute mem_access_attr;
	bool debug_enabled;
	bool debug_v2_enabled;
	u8 *debug_msg_data;

	u16 debug_msg_count;
	struct bin_attribute debug_msg_attr;
	struct mutex debug_msg_lock;
	u8 revision_id;
	u8 max_reportid;
	u32 config_crc;
	u32 info_crc;
	u8 bootloader_addr;
	u8 *msg_buf;
	u8 t6_status;
	bool update_input;
	u8 last_message_count;
	u8 num_touchids;
	u8 num_stylusids;
	unsigned long t15_keystatus;
	bool use_retrigen_workaround;
	bool use_regulator;
	bool poweron;
	bool input_registered;
	bool buttons_enabled;

	bool mode_is_wakeable;
	bool mode_is_persistent;

	struct mxt_patchset *default_mode;
	struct mxt_patchset *alternate_mode;
	struct mxt_patchset *current_mode;

	bool irq_enabled;
	struct regulator *reg_vdd;
	struct regulator *reg_avdd;
	char *fw_name;
	char *cfg_name;
	u8 *T100_data;

	/* Cached parameters from object table */
	u16 T5_address;
	u8 T5_msg_size;
	u8 T6_reportid;
	u16 T6_address;
	u16 T7_address;
	u16 T38_address;
	u8 T9_reportid_min;
	u8 T9_reportid_max;
	u8 T15_reportid_min;
	u8 T15_reportid_max;
	u16 T18_address;
	u8 T19_reportid;
	u8 T42_reportid_min;
	u8 T42_reportid_max;
	u16 T44_address;
	u8 T48_reportid;
	u8 T63_reportid_min;
	u8 T63_reportid_max;
	u8 T93_reportid;
	u8 T100_reportid_min;
	u8 T100_reportid_max;
	u8 T25_reportid;

	/* for fw update in bootloader */
	struct completion bl_completion;

	/* for reset handling */
	struct completion reset_completion;

	/* for reset handling */
	struct completion crc_completion;

	/* for selftest handling */
	struct completion selftest_completion;

	/* Enable reporting of input events */
	bool enable_reporting;

	/* Indicates whether device is in suspend */
	atomic_t suspended;

#ifdef CONFIG_FB
	struct notifier_block fb_notif;
#endif
	struct notifier_block nfc_notifier;

	struct semaphore crit_section_lock;
	const struct firmware *tdat;
	struct mxt_tdat_section fw;
	struct mxt_tdat_section tsett;

	struct workqueue_struct *main_wq;
	struct delayed_work irq_work;
#ifdef CONFIG_FB
	struct delayed_work resume_work;
#endif
	struct delayed_work healthcheck_work;
	struct delayed_work plugin_work;
	struct delayed_work fwupdate_work;
	struct delayed_work nfc_sensing_work;
	struct delayed_work parse_mode_conf_work;

	uint32_t restart_wait_time ;
	uint32_t force_restart_time ;

	atomic_t enable_tp;
	atomic_t nfc_enable_tp;

	struct mxt_tp_conf tp_conf;

	uint16_t mxt_tp_mode;

	int num_touched;
	u8 *touch_state;
	int enable_fw_update;
	int init_completed;
	int finger_sensitivity;
	int glove_enabled;
	int glove_conf;
	int glove_sensitivity;
	int hover_enabled;
	int waterproof_enabled;
	int waterproof_conf;
	int waterResistance;
	int passive_stylus_enabled;
	int active_stylus_enabled;
	int stylus_conf;
	int active_stylus_pressure;

	int running_healthcheck;

	u8 T25_result[6];

	struct work_struct eval_noise_work;
	struct timer_list eval_noise_timer;
};
#define mxt_unlock(s)	{\
		up(s);\
		pr_debug("critical section RELEASED (count %d)\n",\
				((struct semaphore *)(s))->count);\
	}

static void mxt_lock(struct semaphore *sem)
{
	ktime_t after, before;
	int retval, elapsed_time;

	retval = down_trylock(sem);
	if (!retval)
		goto done;
	/* will have to wait */
	before = ktime_get();
	retval = down_interruptible(sem);
	if (retval) {
		pr_err("cannot lock critical section\n");
		return;
	}
	after = ktime_get();
	elapsed_time = (ktime_to_timeval(after).tv_sec -
			ktime_to_timeval(before).tv_sec) * 1000;
	elapsed_time += (ktime_to_timeval(after).tv_usec -
			ktime_to_timeval(before).tv_usec) / 1000;
	if (elapsed_time > 10)
		pr_info("lock delayed by %d ms\n", elapsed_time);
done:
	pr_debug("critical section LOCKED (count %d)\n", sem->count);
}

#define ENABLE_CFGDRV
#ifdef ENABLE_CFGDRV
#include <linux/cfgdrv.h>
#include <linux/hcm_eep.h>
#endif
#include <linux/proc_fs.h>
#include <linux/poll.h>
#include <linux/miscdevice.h>

#define TP_MAX_POS (10)
#define TP_Z_SHIFT (2)
#define TP_MAX_Z (0x3FC) /* 0xFF << 2 */

#define TP_POS_NONE          (0x00000000LU) /**< no detection  */
#define TP_POS_FIN           (0x00000001LU) /**< finger        */
#define TP_POS_HOVER         (0x00000002LU) /**< hover         */
#define TP_POS_GLOVE         (0x00000004LU) /**< glove         */
#define TP_POS_PEN           (0x00000040LU) /**< stylus        */
#define TP_POS_APEN          (0x00000080LU) /**< active stylus */
#define TP_BTN_BTN1          (0x00000100LU) /**< active stylus button1 state */
#define TP_BTN_BTN2          (0x00000200LU) /**< active stylus button2 state */

#define TP_POS_TYPE(op) ((op)&(TP_POS_FIN|TP_POS_PEN|TP_POS_APEN|TP_POS_GLOVE|TP_POS_HOVER))
#define TP_POS_IS_VALID(op) (TP_POS_TYPE((op)))

#define TP_ON_RELEASE (0)
#define TP_ON_TOUCH   (1)
#define TP_ON_HOVER   (2)

#define TP_TYPE_PEN         (1)
#define TP_TYPE_HOVER       (2)
#define TP_TYPE_HOVER_CONT  (3)
#define TP_TYPE_GLOVE       (4)
#define TP_TYPE_WATER_PROOF (5)

#define TP_FUNC_SET_GLOVE_SENS		(0)
#define TP_FUNC_SET_TOUCH_INT_TRIGGER	(1)
#define TP_FUNC_SET_WATER_RESISTANCE	(2)
#define TP_FUNC_SET_PEN_TYPE		(3)
#define TP_FUNC_SET_PEN_ACTIVE_PRESSURE	(4)
#define TP_FUNC_SET_FINGER_SENS		(5)

#define TP_GLOVE_OFF  (0)
#define TP_GLOVE_ON   (1)
#define TP_GLOVE_AUTO (2)

#define TP_GLOVE_SENS_DEFAULT (0)
#define TP_GLOVE_SENS_LOW     (1)
#define TP_GLOVE_SENS_MID     (2)
#define TP_GLOVE_SENS_HIGH    (3)

#define TP_WATER_PROOF_OFF (0)
#define TP_WATER_PROOF_ON  (1)

#define TP_WATER_RESISTANCE_NORMAL (0)
#define TP_WATER_RESISTANCE_HIGH   (1)

#define TP_PEN_OFF  (0)
#define TP_PEN_ON   (1)

#define TP_PEN_TYPE_ACTIVE         (0x1)
#define TP_PEN_TYPE_PASSIVE        (0x2)
#define TP_PEN_TYPE_ACTIVE_PASSIVE (TP_PEN_TYPE_ACTIVE|TP_PEN_TYPE_PASSIVE)

#define TP_PEN_ACTIVE_PRESSURE_NORMAL	(0x0)
#define TP_PEN_ACTIVE_PRESSURE_LIGHT	(0x1)

#define TP_FINGER_SENS_LOW     (1)
#define TP_FINGER_SENS_MID     (2)
#define TP_FINGER_SENS_HIGH    (3)

#define TP_MM2DOT_SHIFT (10)
static unsigned int tp_mm2dot_factor = (12711); /* (720[dot]/58[mm]) << SHIFT */
#define TP_MM2DOT(mm) (((mm) * tp_mm2dot_factor) >> TP_MM2DOT_SHIFT)

#define TP_NVM_INIT_WAIT (1000) /* ms */
#define TP_MAX_RETRY_NVM_INIT (100)

#define TPIF_PROCFS_DIR		"tpif"
#define TPIF_PROCFS_SELFTEST12	"selftest12"
#define TPIF_PROCFS_SELFTEST17	"selftest17"

#define TPIF_DEV_NAME      "tpif"

#define TPIF_FW_UPDATE_ALL (0)
#define TPIF_FW_UPDATE_CFG (1)
#define TPIF_FW_FILE_NAME_SIZE (256)
#define TPIF_PANEL_NAME_SIZE (256)
#define TPIF_TOUCH_NAME_SIZE (256)

#define TPIF_IOCTL_GET_VERSION	(100)
#define TPIF_IOCTL_UPDATE_FW	(101)
#define TPIF_IOCTL_GET_PANEL	(102)
#define TPIF_IOCTL_GET_TOUCH	(103)

#define TP_CFG_CMD_POWER		(0)
#define TP_CFG_CMD_GLOVE		(1)
#define TP_CFG_CMD_WATERPROOF		(2)
#define TP_CFG_CMD_ACTIVE_STYLUS	(3)
#define TP_CFG_CMD_PASSIVE_STYLUS	(4)
#define TP_CFG_CMD_CALIB		(5)

struct TpPos_t {
	short x;
	short y;
	short z;
	short wx;
	short wy;
	short area;
	short misc;
	unsigned long option;
	short slot;
};

struct TpEvent_t {
	struct TpPos_t pos[TP_MAX_POS];
	unsigned long time;
	unsigned short posState;
	int numFinger;
	int numHover;
	int isValid;
	int keepCount;
	unsigned long maxTime;
	int maxValue;
	int threshold;
	int isValidEvPos;
	int touchCount;
};

enum TP_SPEC {
	TP_SPEC_MT_X = 0,
	TP_SPEC_MT_Y,
	TP_SPEC_MT_PRESSURE,
	TP_SPEC_MT_WIDTH,
	TP_NUM_SPEC
};

struct ParamSpec_t {
	int min;
	int max;
	int fuzz;
	int flat;
};

struct DefaultParams_t {
	struct ParamSpec_t spec[TP_NUM_SPEC];
};

struct TouchPanelPlugin {
	int (*process)(struct TpEvent_t *ev, int numValid, int maxEntries);
	void (*postProcess)(struct TpEvent_t *ev);
	void (*reset)(void);
};

struct TouchPanelRegister {
    int (*read)(unsigned int addr, unsigned char *buf, int size);
    int (*write)(unsigned int addr, const unsigned char *buf, int size);
};

struct TouchPanelParameter {
	const struct DefaultParams_t *(*getDefault)(void);
	const struct HoverZCalib_t *(*getHoverZCalib)(void);
};

struct TouchPanelFirmware {
    int (*update)(const unsigned char *img, int target, int force);
};

struct TouchPanelDevCalib {
    void (*refresh)(int type);
    void (*sendRefreshCmd)(int type);
};

struct TouchPanelDevFilter {
    unsigned char backup;
    void (*enableDevFilter)(int enable, unsigned char *setting);
};

struct TouchPanelFeature {
	void (*getFeature)(int type, int *mode);
	void (*setFeature)(int type, int mode);
	int (*funcFeature)(int func, unsigned long arg);
};

struct TouchPanelController {
    void (*setIntMode)(int isEnable);
    void (*setFullAwake)(int mode);
    void (*startWakeUpPalm)(void);
};

struct TouchPanelInput {
    void (*enqueueTouchEvent)(const struct TpEvent_t *ev);
    int (*enqueueTouchEventAsync)(const struct TpEvent_t *ev, int num);
    void (*enqueueBackKey)(void);
};

struct TouchPanelHandler {
	struct TouchPanelRegister reg;
	struct TouchPanelParameter params;
	struct TouchPanelFirmware fw;
	struct TouchPanelDevCalib devCalib;
	struct TouchPanelDevFilter devFilter;
	struct TouchPanelFeature devFeature;
	struct TouchPanelController devController;
	struct TouchPanelInput input;
};

struct PluginInstaller {
	int (*install)(struct TouchPanelPlugin *plugin,
		       struct TouchPanelHandler **handler);
	void (*uninstall)(int id);
};

struct TouchPanelFwVersion_t {
	u16 base;
	u32 config;
	u16 config_rev;
};
struct TouchPanelFwUpdateInfo_t {
	int result;
	int target;
	int force;
	char fw_name[TPIF_FW_FILE_NAME_SIZE];
	char cfg_name[TPIF_FW_FILE_NAME_SIZE];
	char panel_name[TPIF_PANEL_NAME_SIZE];
	char touch_name[TPIF_TOUCH_NAME_SIZE];
};

struct TouchPanelCommand_t {
	struct mxt_data *pdata;
	struct semaphore lock;
	int gloveMode;
	int gloveSens;
	int waterProofMode;
	int waterResistance;
	int stylusMode;
	int stylusType;
	int stylusActivePressure;
	int fingerSens;
	struct TouchPanelFwUpdateInfo_t *fwUpdate;
};

struct glove_config_t {
	u8 enable;
	u8 glove_self_thr;
	u8 glove_hys;
	u8 discr_thr;
	u8 active_mutual_self;
	u8 next_move_filter;
	u8 max_touch_amp;
	u8 z_touch_thr;
	u8 z_touch_thr_hys;
	u8 init_move_filter;
	u8 noise_on_cal;
	u8 noise_on;
	u8 ins1_dynamic_en;
	u8 ins2_dynamic_en;
	u8 ins5_dynamic_en;
	u8 ins7_dynamic_en;
	u8 ins8_dynamic_en;
	u8 long_cal_en;
	u8 dc_filter;
	u8 dc_cal_rec_str;
};

struct active_stylus_config_t {
	u8 enable;
	u8 detect_mode;
};

struct passive_stylus_config_t {
	u8 enable;
	u8 touch_debounce;
	u8 release_debounce;
	u8 smooth_level;
};

struct waterproof_config_t {
	u8 report_mode;
	u8 touch_gain;
	u8 delta_touch_thr;
	u8 anti_moisture;
};

struct calib_config_t {
	int force_calib_cnt; /* unit: healthcheck interval */
};

struct touch_dev_config_t {
	struct t7_config t7_finger_cfg;
	struct t7_config t7_stylus_cfg;
	struct glove_config_t glove_off_cfg;
	struct glove_config_t glove_on_cfg_low;
	struct glove_config_t glove_on_cfg_mid;
	struct glove_config_t glove_on_cfg_high;
	struct active_stylus_config_t active_stylus_off_cfg;
	struct active_stylus_config_t active_stylus_on_cfg_normal;
	struct active_stylus_config_t active_stylus_on_cfg_light;
	struct passive_stylus_config_t passive_stylus_off_cfg;
	struct passive_stylus_config_t passive_stylus_on_cfg;
	struct waterproof_config_t waterproof_off_cfg;
	struct waterproof_config_t waterproof_on_cfg_light;
	struct waterproof_config_t waterproof_on_cfg_heavy;
	struct calib_config_t calib_cfg;
};
static struct touch_dev_config_t touch_dev_cfg = {
	.t7_finger_cfg = {
		.idle			= 0x10,
		.active			= 0x04,
		.actv2idle		= 0x0A,
	},
	.t7_stylus_cfg = {
		.idle			= 0x08,
		.active			= 0xFF,
		.actv2idle		= 0x0A,
	},
	.glove_off_cfg = {
		.enable			= 0x80,
		.glove_self_thr		= 0x00,
		.glove_hys		= 0x00,
		.discr_thr		= 0x32,
		.active_mutual_self	= 0x01,
		.next_move_filter	= 0x01,
		.max_touch_amp		= 0x00,
		.z_touch_thr		= 0x46,
		.z_touch_thr_hys	= 0x14,
		.init_move_filter	= 0x00,
		.noise_on_cal		= 0x00,
		.noise_on		= 0x00,
		.ins1_dynamic_en	= 0x00,
		.ins2_dynamic_en	= 0x00,
		.ins5_dynamic_en	= 0x00,
		.ins7_dynamic_en	= 0x00,
		.ins8_dynamic_en	= 0x00,
		.long_cal_en		= 0x04,
		.dc_filter		= 0x33,
		.dc_cal_rec_str		= 0x00,
	},
	.glove_on_cfg_low = {
		.enable			= 0x81,
		.glove_self_thr		= 0x00,
		.glove_hys		= 0x00,
		.discr_thr		= 0x32,
		.active_mutual_self	= 0x01,
		.next_move_filter	= 0x01,
		.max_touch_amp		= 0x00,
		.z_touch_thr		= 0x46,
		.z_touch_thr_hys	= 0x14,
		.init_move_filter	= 0x00,
		.noise_on_cal		= 0x00,
		.noise_on		= 0x00,
		.ins1_dynamic_en	= 0x00,
		.ins2_dynamic_en	= 0x00,
		.ins5_dynamic_en	= 0x00,
		.ins7_dynamic_en	= 0x00,
		.ins8_dynamic_en	= 0x00,
		.long_cal_en		= 0x04,
		.dc_filter		= 0x33,
		.dc_cal_rec_str		= 0x00,
	},
	.glove_on_cfg_mid = {
		.enable			= 0x81,
		.glove_self_thr		= 0x46,
		.glove_hys		= 0x14,
		.discr_thr		= 0x32,
		.active_mutual_self	= 0x01,
		.next_move_filter	= 0x01,
		.max_touch_amp		= 0x01,
		.z_touch_thr		= 0x1E,
		.z_touch_thr_hys	= 0x14,
		.init_move_filter	= 0x32,
		.noise_on_cal		= 0x00,
		.noise_on		= 0x00,
		.ins1_dynamic_en	= 0x00,
		.ins2_dynamic_en	= 0x00,
		.ins5_dynamic_en	= 0x00,
		.ins7_dynamic_en	= 0x00,
		.ins8_dynamic_en	= 0x00,
		.long_cal_en		= 0x04,
		.dc_filter		= 0x33,
		.dc_cal_rec_str		= 0x00,
	},
	.glove_on_cfg_high = {
		.enable			= 0x81,
		.glove_self_thr		= 0x14,
		.glove_hys		= 0x05,
		.discr_thr		= 0x32,
		.active_mutual_self	= 0x01,
		.next_move_filter	= 0x01,
		.max_touch_amp		= 0x01,
		.z_touch_thr		= 0x1E,
		.z_touch_thr_hys	= 0x14,
		.init_move_filter	= 0x32,
		.noise_on_cal		= 0x02,
		.noise_on		= 0x00,
		.ins1_dynamic_en	= 0x01,
		.ins2_dynamic_en	= 0x01,
		.ins5_dynamic_en	= 0x01,
		.ins7_dynamic_en	= 0x01,
		.ins8_dynamic_en	= 0x00,
		.long_cal_en		= 0x04,
		.dc_filter		= 0x37,
		.dc_cal_rec_str		= 0x64,
	},
	.active_stylus_off_cfg = {
		.enable			= 0x0B,
		.detect_mode		= 0x82,
	},
	.active_stylus_on_cfg_normal = {
		.enable			= 0x1B,
		.detect_mode		= 0x80,
	},
	.active_stylus_on_cfg_light = {
		.enable			= 0x1B,
		.detect_mode		= 0x80,
	},
	.passive_stylus_off_cfg = {
		.enable			= 0x08,
		.touch_debounce		= 0x00,
		.release_debounce	= 0x00,
		.smooth_level		= 0x96,
	},
	.passive_stylus_on_cfg = {
		.enable			= 0x09,
		.touch_debounce		= 0x03,
		.release_debounce	= 0x03,
		.smooth_level		= 0xE6,
	},
	.waterproof_off_cfg = {
		.report_mode		= 0x00,
		.touch_gain		= 0x0E,
		.delta_touch_thr	= 0x0F,
		.anti_moisture		= 0x01,
	},
	.waterproof_on_cfg_light = {
		.report_mode		= 0x01,
		.touch_gain		= 0x0E,
		.delta_touch_thr	= 0x0F,
		.anti_moisture		= 0x01,
	},
	.waterproof_on_cfg_heavy = {
		.report_mode		= 0x01,
		.touch_gain		= 0x0E,
		.delta_touch_thr	= 0x0F,
		.anti_moisture		= 0x01,
	},
	.calib_cfg = {
		.force_calib_cnt	= 12,
	},
};

/* porting layer function */
static void TpIf_init(struct mxt_data *data);
static void TpIf_deinit(struct mxt_data *data);
static void TpIf_queued_plugin_cb_func(struct work_struct *w);
static void TpIf_convertPos(struct TpPos_t *pos, int slot, u8 *message);
static void InputSubsystem_store(const struct TpPos_t *pos);
static void InputSubsystem_update(void);
static void InputSubsystem_enqueue(struct input_dev *input_dev,
				   const struct TpEvent_t *ev);

/* common interface */
static int  TpIfMain_install(struct TouchPanelPlugin *plugin,
                             struct TouchPanelHandler **handler);
static void TpIfMain_uninstall(int id);

static const struct DefaultParams_t *TpIfMainAndroid_getDefault(void);
static void TpIfMainAndroid_getFeature(int type, int *mode);
static void TpIfMainAndroid_setFeature(int type, int mode);
static int  TpIfMainAndroid_funcFeature(int func, unsigned long arg);

/* common fops (nop) */
static unsigned int tpif_fops_poll(struct file *file, poll_table *wait);
static int tpif_fops_open(struct inode *inode, struct file *file);
static int tpif_fops_close(struct inode *inode, struct file *file);

/* ioctl */
static long tpif_dev_ioctl(struct file *filp, unsigned int cmd,
			   unsigned long arg);

/* procfs entry */
static ssize_t tpif_read_pin_fault_test(struct file *file,
					char __user *buf, size_t count,
					loff_t *off);
static ssize_t tpif_write_pin_fault_test(struct file *file,
					 const char __user *buf, size_t count,
					 loff_t *off);
static ssize_t tpif_read_signal_limit_test(struct file *file,
					   char __user *buf, size_t count,
					   loff_t *off);
static ssize_t tpif_write_signal_limit_test(struct file *file,
					    const char __user *buf, size_t count,
					    loff_t *off);

/* static variables */
static int mId = -1;

struct PluginInstaller tpPluginInstaller = {
	.install   = TpIfMain_install,
	.uninstall = TpIfMain_uninstall
};
EXPORT_SYMBOL(tpPluginInstaller);

static struct TouchPanelPlugin *mPlugin = NULL;

static struct TouchPanelHandler mHandler = {
	.params = {
		.getDefault = TpIfMainAndroid_getDefault,
	},
	.devFeature = {
		.getFeature = TpIfMainAndroid_getFeature,
		.setFeature = TpIfMainAndroid_setFeature,
		.funcFeature = TpIfMainAndroid_funcFeature,
	},
};

struct TpEvent_t mCurrentEvent;
static struct TouchPanelCommand_t mTpCmd;
static int mForceCalibCnt = 12;

static struct file_operations tpif_dev_fops = {
	.owner		= THIS_MODULE,
	.llseek		= no_llseek,
	.unlocked_ioctl	= tpif_dev_ioctl,
	.poll		= tpif_fops_poll,
	.open		= tpif_fops_open,
	.release	= tpif_fops_close,
};
static struct miscdevice tpif_dev = {
	.minor	= MISC_DYNAMIC_MINOR,
	.name	= TPIF_DEV_NAME,
	.fops	= &tpif_dev_fops
};

static const struct file_operations tpif_pin_fault_test_fops = {
	.owner		= THIS_MODULE,
	.llseek		= no_llseek,
	.read		= tpif_read_pin_fault_test,
	.write		= tpif_write_pin_fault_test,
	.poll		= tpif_fops_poll,
	.open		= tpif_fops_open,
	.release	= tpif_fops_close
};
static const struct file_operations tpif_signal_limit_test_fops = {
	.owner		= THIS_MODULE,
	.llseek		= no_llseek,
	.read		= tpif_read_signal_limit_test,
	.write		= tpif_write_signal_limit_test,
	.poll		= tpif_fops_poll,
	.open		= tpif_fops_open,
	.release	= tpif_fops_close
};
static struct proc_dir_entry *tpif_procfs_dir = NULL;
static struct proc_dir_entry *tpif_pin_fault_test_entry = NULL;
static struct proc_dir_entry *tpif_signal_limit_test_entry = NULL;

/* debugging function */
#define ATMEL_DEBUG_TAG		"atmel_mxt_ts"
#define ATMEL_DEBUG_MSG		(0x00000001)
#define ATMEL_DEBUG_TRACE	(0x00000002)
#define ATMEL_DEBUG_DBG		(0x00000004)
#define ATMEL_DEBUG_DAT		(0x00000008)

#define ATMEL_ERR(fmt, arg...)				\
printk(KERN_ERR ATMEL_DEBUG_TAG":ERROR: [%s():%d] "fmt,	\
	__FUNCTION__, __LINE__, ## arg)


#define ATMEL_MSG(fmt, arg...)					\
if (atmel_debug_flag & ATMEL_DEBUG_MSG) {			\
	printk(KERN_INFO ATMEL_DEBUG_TAG":MSG: [%s():%d] "fmt,	\
		__FUNCTION__, __LINE__, ## arg);		\
}

#define ATMEL_TRACE()							\
if (atmel_debug_flag & ATMEL_DEBUG_TRACE) {				\
	printk(KERN_INFO ATMEL_DEBUG_TAG":TRC: %s()\n", __FUNCTION__);	\
}

#define ATMEL_DBG(fmt, arg...)					\
if (atmel_debug_flag & ATMEL_DEBUG_DBG) {			\
	printk(KERN_INFO ATMEL_DEBUG_TAG":DBG: [%s():%d] "fmt,	\
		__FUNCTION__, __LINE__, ## arg);		\
}

#define ATMEL_DAT(fmt, arg...)					\
if (atmel_debug_flag & ATMEL_DEBUG_DAT) {			\
	printk(KERN_INFO ATMEL_DEBUG_TAG":DAT: [%s():%d] "fmt,	\
		__FUNCTION__, __LINE__, ## arg);		\
}

static unsigned long atmel_debug_flag = 0x0;
/* static unsigned long atmel_debug_flag = 0xF; */
module_param(atmel_debug_flag, ulong, S_IRUGO|S_IWUSR);

static unsigned long atmel_healthcheck_force_disable = 0x0;

static void mxt_irq_enable(struct mxt_data *data, bool enable);
static int mxt_hw_reset(struct mxt_data *data);
static int mxt_suspend(struct device *dev, int awaitable);
static int mxt_resume(struct device *dev);
static int mxt_get_sensor_state(struct mxt_data *data);
static int mxt_init_t7_power_cfg(struct mxt_data *data);
static void mxt_regulator_disable(struct mxt_data *data);
static void mxt_regulator_enable(struct mxt_data *data);
static void mxt_reset_slots(struct mxt_data *data);
static void mxt_wait_for_idle(struct mxt_data *data);
static int mxt_update_file_name(struct device *dev, char **file_name,
				const char *buf, size_t count);
static u32 mxt_update_config_id(struct mxt_data *data);
static void mxt_update_vendor_ver_id(struct mxt_data *data);
static int mxt_set_glove(struct mxt_data *data, int enable);
static int mxt_set_hover(struct mxt_data *data, int enable);
static int mxt_set_passive_stylus(struct mxt_data *data, int enable);
static int mxt_set_active_stylus(struct mxt_data *data, int enable);
static int mxt_set_waterproof(struct mxt_data *data, int enable);
static void mxt_update_t7_config(struct mxt_data *data);
static void mxt_init_feature(struct mxt_data *data);
static void mxt_healthcheck(struct mxt_data *data);
static void mxt_create_task(struct mxt_data *data);
static void mxt_stop_all_task(struct mxt_data *data);
static void mxt_destroy_task(struct mxt_data *data);
static void mxt_queue_irq(struct mxt_data *data);
static void mxt_work_irq(struct work_struct *work);
static void mxt_parse_mode_conf(struct mxt_data *data);
#ifdef CONFIG_FB
static void mxt_queue_resume(struct mxt_data *data);
static void mxt_work_resume(struct work_struct *work);
#endif
static void mxt_queue_healthcheck(struct mxt_data *data);
static void mxt_stop_healthcheck(struct mxt_data *data);
static void mxt_work_healthcheck(struct work_struct *work);
static void mxt_queue_nfc_sening(struct mxt_data *data);
static void mxt_stop_nfc_sensing(struct mxt_data *data);
static void mxt_work_nfc_sensing(struct work_struct *work);
static void mxt_nfc_sensing(struct mxt_data *data);
static void mxt_queue_plugin(struct mxt_data *data);
static void mxt_work_plugin(struct work_struct *work);
static int mxt_read_fw_version(struct mxt_data *data, u16 *ver);
static int mxt_read_cfg_version(struct mxt_data *data, u32 *ver, u16 *rev);
static int mxt_update_firmware(struct mxt_data *data, char *filename);
static int mxt_update_config(struct mxt_data *data, char *filename);
static void mxt_queue_fwupdate(struct mxt_data *data);
static void mxt_work_fwupdate(struct work_struct *work);
static void mxt_work_parse_conf_work(struct work_struct *work);
static int mxt_set_detect_mode(struct mxt_data *data);

struct debug_section {
	unsigned long j;
	unsigned int count;
};

#ifndef PROC_COORD_EX
static struct debug_section mxt_proc_t100_dbg;
#endif
static struct debug_section mxt_dump_message_dbg;

/* Function: throttle_dbgout
 *  Print no more than n messages in msec interval
 *  Return
 *  - false if printing is allowed
 *  - true if printing is NOT allowed
 */
bool throttle_dbgout(struct debug_section *dbg, int n, unsigned int msec)
{
	if (printk_timed_ratelimit(&dbg->j, msec))
		dbg->count = 0;

	if (dbg->count < n) {
		++dbg->count;
		return false;
	} else
		return true;
}

struct touch_up_down {
	int	mismatch;
	u8	up_down;
	unsigned int counter;
};

struct touch_area_stats {
	struct touch_up_down *ud;
	ssize_t ud_len;
	ssize_t ud_id;
	ssize_t unknown_counter;
	const char *name;
};

static struct touch_up_down display_ud[20];
static struct touch_area_stats display_ud_stats = {
	.ud = display_ud,
	.ud_len = ARRAY_SIZE(display_ud),
	.name = "ts"
};

static struct touch_up_down button_ud[10];
static struct touch_area_stats button_ud_stats = {
	.ud = button_ud,
	.ud_len = ARRAY_SIZE(button_ud),
	.name = "btn"
};

static void ud_set_id(struct touch_area_stats *tas, int id)
{
	/* ATMEL_TRACE(); */
	tas->ud_id = id;
}

static void ud_log_status(struct touch_area_stats *tas, bool down)
{
	struct touch_up_down *ud = tas->ud;
	ssize_t id = tas->ud_id;

	/* ATMEL_TRACE(); */

	if (id >= tas->ud_len)
		tas->unknown_counter++;

	if (!down) { /* up */
		if (ud[id].up_down == 0x10) {
			pr_debug("%s UP[%zd]\n", tas->name, id);
			ud[id].up_down |= 1;
			ud[id].mismatch--;
		}
	} else if (down) { /* down */
		if (ud[id].up_down == 0) {
			ud[id].up_down |= (1 << 4);
			pr_debug("%s DOWN[%zd]\n", tas->name, id);
			ud[id].mismatch++;
		} else if (ud[id].up_down == 0x10)
			return;
	}

	if (ud[id].up_down == 0x11) {
		pr_debug("%s CLEAR[%zd]\n", tas->name, id);
		ud[id].up_down = 0;
		ud[id].counter++;
	}
}

#if 0
static void TSI_state(struct input_dev *dev, unsigned int tool, bool status)
{
	ud_log_status(&display_ud_stats, status);
	input_mt_report_slot_state(dev, tool, status);
}

static void TSI_id(struct input_dev *dev, int id)
{
	ud_set_id(&display_ud_stats, id);
	input_mt_slot(dev, id);
}

#define input_mt_report_slot_state	TSI_state
#define input_mt_slot	TSI_id
#endif

static ssize_t mxt_ud_show(struct device *dev,
				   struct device_attribute *attr, char *buf)
{
	int i;
	struct mxt_data *data = dev_get_drvdata(dev);
	ssize_t total = 0;

	/* ATMEL_TRACE(); */

	total += scnprintf(buf + total, PAGE_SIZE - total, "display:\n");
	for (i = 0; i < display_ud_stats.ud_len; i++)
		total += scnprintf(buf + total, PAGE_SIZE - total,
				"[%d]: full cycles-%u, mismatch-%d\n", i,
				display_ud[i].counter,
				display_ud[i].mismatch);

	if (data->buttons_enabled) {
		total += scnprintf(buf + total, PAGE_SIZE - total,
			"buttons:\n");
		for (i = 0; i < button_ud_stats.ud_len; i++)
			total += scnprintf(buf + total, PAGE_SIZE - total,
					"[%d]: full cycles-%u, mismatch-%d\n",
					i,
					button_ud[i].counter,
					button_ud[i].mismatch);
	}
	return total;
}

static ssize_t mxt_ud_stat(char *buf, ssize_t size)
{
	int i;
	ssize_t total = 0;

	/* ATMEL_TRACE(); */

	total += scnprintf(buf + total, size - total, "screen: ");
	for (i = 0; i < display_ud_stats.ud_len; i++)
		if (display_ud[i].mismatch)
			total += scnprintf(buf + total, size - total,
				"%d)%u,%d ", i,
				display_ud[i].counter,
				display_ud[i].mismatch);
		else if (display_ud[i].counter)
			total += scnprintf(buf + total, size - total,
				"%d)%u ", i,
				display_ud[i].counter);

	total += scnprintf(buf + total, size - total, "buttons: ");
	for (i = 0; i < button_ud_stats.ud_len; i++)
		if (button_ud[i].mismatch)
			total += scnprintf(buf + total, size - total,
				"%d)%u,%d ", i,
				button_ud[i].counter,
				button_ud[i].mismatch);
		else if (button_ud[i].counter)
			total += scnprintf(buf + total, size - total,
				"%d)%u ", i,
				button_ud[i].counter);

	return total;
}

static char *mxt_find_patch(char *head, char *delimiters, char **next)
{
	char *patch = head;

	ATMEL_TRACE();

	for (; patch; patch = *next) {
		*next = strpbrk(patch, delimiters);
		if (*next)
			*(*next)++ = '\0';

		patch = skip_spaces(patch);
		if (!patch || !*patch || *patch == '#')
			continue;
		else
			break;
	}

	return patch;
}

static int mxt_parse_patch(int object, int instance, char *query,
		struct mxt_patch *patch_ptr)
{
	int i, error;
	char *next, *value_p, *pair = query;
	long offset_v, value_v;
	struct mxt_obj_patch *patch;

	ATMEL_TRACE();

	for (i = 0; pair; pair = next, i++) {
		pair = mxt_find_patch(pair, ",", &next);

		value_p = strpbrk(pair, "=");
		if (!value_p) {
			pr_err("T%d[%d]: invalid syntax '%s'\n",
				object, i, pair);
			continue;
		}

		/* make sure string is null terminated */
		*value_p = '\0';

		error = kstrtol(pair, 10, &offset_v);
		if (error) {
			pr_err("T%d[%d]: dec conversion error\n", object, i);
			continue;
		}

		error = kstrtol(++value_p, 16, &value_v);
		if (error) {
			pr_err("T%d[%d]: hex conversion error\n", object, i);
			continue;
		}

		/* primitive data validation */
		if (object <= 0 || object > 255 ||
		    offset_v < 0 || offset_v > 255) {
			pr_err("T%d[%d]: invalid values\n", object, i);
			continue;
		}

		patch = kzalloc(sizeof(*patch), GFP_KERNEL);
		if (!patch) {
			pr_err("failed to alloc mem\n");
			return -ENOMEM;
		}

		patch->number = (u8)object;
		patch->instance = (u8)instance;
		patch->offset = (u8)offset_v;
		patch->value = (u8)value_v;

		pr_debug("T%d[%d], instance %d, offset %d, value 0x%02x\n",
			patch->number, i,
			patch->instance, patch->offset, patch->value);
		patch_ptr->cfg_num++;
		list_add_tail(&patch->link, &patch_ptr->cfg_head);
	}

	return 0;
}

/*
 * Special settings can be passed to the driver via kernel cmd line
 * Example: atmxt="T100@0=1f,1=a0;T72@47=b2;T110-3@26=a"
 *   Where:
 *      T100    - decimal object number
 *      @       - delimits object number and following patch sets
 *      0=1f    - patch set decimal offset and hex value
 *      110-3   - object number and instance
 */
static void mxt_parse_setup_string(struct mxt_data *data,
		const char *patch_ptr, struct mxt_patch *patch)
{
	struct device *dev = &data->client->dev;
	long number_v, instance_v;
	char *patch_string;
	char *config_p, *instance_p, *next, *patch_set;
	int i, error;

	ATMEL_TRACE();

	patch_string = kstrdup(patch_ptr, GFP_KERNEL);
	for (i = 0, patch_set = patch_string; patch_set; patch_set = next) {
		patch_set = mxt_find_patch(patch_set, ";\n", &next);
		if (!patch_set)
			break;

		dev_dbg(dev, "patch set %d: \"%s\"\n", i, patch_set);

		config_p = strpbrk(patch_set, "@");

		if ((*patch_set != 'T' && *patch_set != 't') || !config_p) {
			dev_err(dev, "invalid syntax '%s'\n", patch_set);
			continue;
		}

		/* strip non digits */
		*config_p++ = '\0';

		instance_v = 0L;
		instance_p = strpbrk(patch_set, "-");
		if (instance_p) {
			*instance_p++ = '\0';
			error = kstrtol(instance_p, 10, &instance_v);
			if (error)
				dev_err(dev, "kstrtol error %d\n", error);
		}

		error = kstrtol(++patch_set, 10, &number_v);
		if (error) {
			dev_err(dev, "kstrtol error %d\n", error);
			continue;
		}

		error = mxt_parse_patch((int)number_v, (int)instance_v,
				config_p, patch);
		if (error < 0) {
			dev_err(dev, "invalid patch; parse error %d\n", error);
			continue;
		}

		i++;
	}
	kfree(patch_string);
	if (patch->cfg_num)
		dev_info(dev, "processed %d patch sets for %d objects\n",
			patch->cfg_num, i);
	else
		dev_info(dev, "no valid patch sets found\n");
}

static inline size_t mxt_obj_size(const struct mxt_object *obj)
{
	/* ATMEL_TRACE(); */

	return obj->size_minus_one + 1;
}

static inline size_t mxt_obj_instances(const struct mxt_object *obj)
{
	/* ATMEL_TRACE(); */

	return obj->instances_minus_one + 1;
}

static bool mxt_object_readable(unsigned int type)
{
	ATMEL_TRACE();

	switch (type) {
	case MXT_GEN_POWER_T7:
	case MXT_GEN_ACQUIRE_T8:
	case MXT_SPT_SELFTEST_T25:
	case MXT_SPT_USERDATA_T38:
	case MXT_SPT_CTECONFIG_T46:
	case MXT_PROCI_LENSEBENDING_T65:
	case MXT_PROCG_NOISESUPPRESSION_T72:
	case MXT_TOUCH_MULTITOUCHSCREEN_T100:
	case MXT_PROCI_GLOVEDETECTION_T78:
	case MXT_PROCI_ACTIVESTYLUS_T107:
	case MXT_PROCI_STYLUS_T47:
		return true;
	case MXT_GEN_DATASOURCE_T53:
	case MXT_TOUCH_MULTI_T9:
	case MXT_TOUCH_KEYARRAY_T15:
	case MXT_TOUCH_PROXIMITY_T23:
	case MXT_TOUCH_PROXKEY_T52:
	case MXT_PROCI_GRIPFACE_T20:
	case MXT_PROCG_NOISE_T22:
	case MXT_PROCI_ONETOUCH_T24:
	case MXT_PROCI_TWOTOUCH_T27:
	case MXT_PROCI_GRIP_T40:
	case MXT_PROCI_PALM_T41:
	case MXT_PROCI_TOUCHSUPPRESSION_T42:
	case MXT_PROCG_NOISESUPPRESSION_T48:
	case MXT_SPT_COMMSCONFIG_T18:
	case MXT_SPT_GPIOPWM_T19:
	case MXT_SPT_CTECONFIG_T28:
	case MXT_SPT_DIGITIZER_T43:
	default:
		return false;
	}
}

static void mxt_dump_message(struct mxt_data *data, u8 *message)
{
	ATMEL_TRACE();

	if (!throttle_dbgout(&mxt_dump_message_dbg, 20, 300000))
		print_hex_dump_bytes("atmel_mxt_ts_mmi: MXT MSG:",
			DUMP_PREFIX_NONE, message, data->T5_msg_size);
}

static void mxt_debug_msg_enable(struct mxt_data *data)
{
	struct device *dev = &data->client->dev;

	ATMEL_TRACE();

	if (data->debug_v2_enabled)
		return;

	mutex_lock(&data->debug_msg_lock);

	data->debug_msg_data = kcalloc(DEBUG_MSG_MAX,
				data->T5_msg_size, GFP_KERNEL);
	if (!data->debug_msg_data) {
		mutex_unlock(&data->debug_msg_lock);
		dev_err(&data->client->dev, "Failed to allocate buffer\n");
		return;
	}

	data->debug_v2_enabled = true;
	mutex_unlock(&data->debug_msg_lock);

	dev_info(dev, "Enabled message output\n");
}

static void mxt_debug_msg_disable(struct mxt_data *data)
{
	struct device *dev = &data->client->dev;

	ATMEL_TRACE();

	if (!data->debug_v2_enabled)
		return;

	dev_info(dev, "disabling message output\n");
	data->debug_v2_enabled = false;

	mutex_lock(&data->debug_msg_lock);
	kfree(data->debug_msg_data);
	data->debug_msg_data = NULL;
	data->debug_msg_count = 0;
	mutex_unlock(&data->debug_msg_lock);
	dev_info(dev, "Disabled message output\n");
}

static void mxt_debug_msg_add(struct mxt_data *data, u8 *msg)
{
	struct device *dev = &data->client->dev;

	ATMEL_TRACE();

	mutex_lock(&data->debug_msg_lock);

	if (!data->debug_msg_data) {
		mutex_unlock(&data->debug_msg_lock);
		dev_err(dev, "No buffer!\n");
		return;
	}

	if (data->debug_msg_count < DEBUG_MSG_MAX) {
		memcpy(data->debug_msg_data +
		       data->debug_msg_count * data->T5_msg_size,
		       msg, data->T5_msg_size);
		data->debug_msg_count++;
	} else {
		dev_dbg(dev, "Discarding %u messages\n", data->debug_msg_count);
		data->debug_msg_count = 0;
	}

	mutex_unlock(&data->debug_msg_lock);

	sysfs_notify(&data->client->dev.kobj, NULL, "debug_notify");
}

static ssize_t mxt_debug_msg_write(struct file *filp, struct kobject *kobj,
	struct bin_attribute *bin_attr, char *buf, loff_t off,
	size_t count)
{
	ATMEL_TRACE();

	return -EIO;
}

static ssize_t mxt_debug_msg_read(struct file *filp, struct kobject *kobj,
	struct bin_attribute *bin_attr, char *buf, loff_t off, size_t bytes)
{
	struct device *dev = container_of(kobj, struct device, kobj);
	struct mxt_data *data = dev_get_drvdata(dev);
	size_t count, bytes_read;

	ATMEL_TRACE();

	if (!data->debug_msg_data) {
		dev_err(dev, "No buffer!\n");
		return 0;
	}

	count = bytes / data->T5_msg_size;

	if (count > DEBUG_MSG_MAX)
		count = DEBUG_MSG_MAX;

	mutex_lock(&data->debug_msg_lock);

	if (count > data->debug_msg_count)
		count = data->debug_msg_count;

	bytes_read = count * data->T5_msg_size;

	memcpy(buf, data->debug_msg_data, bytes_read);
	data->debug_msg_count = 0;

	mutex_unlock(&data->debug_msg_lock);

	return bytes_read;
}

static int mxt_debug_msg_init(struct mxt_data *data)
{
	ATMEL_TRACE();

	sysfs_bin_attr_init(&data->debug_msg_attr);
	data->debug_msg_attr.attr.name = "debug_msg";
	data->debug_msg_attr.attr.mode = S_IRUGO | S_IWUSR | S_IWGRP;
	data->debug_msg_attr.read = mxt_debug_msg_read;
	data->debug_msg_attr.write = mxt_debug_msg_write;
	data->debug_msg_attr.size = data->T5_msg_size * DEBUG_MSG_MAX;

	if (sysfs_create_bin_file(&data->client->dev.kobj,
				  &data->debug_msg_attr) < 0) {
		dev_err(&data->client->dev, "Failed to create %s\n",
			data->debug_msg_attr.attr.name);
		return -EINVAL;
	}

	return 0;
}

static void mxt_debug_msg_remove(struct mxt_data *data)
{
	ATMEL_TRACE();

	if (data->debug_msg_attr.attr.name) {
		sysfs_remove_bin_file(&data->client->dev.kobj,
				      &data->debug_msg_attr);
		data->debug_msg_attr.attr.name = NULL;
	}
}

static int mxt_wait_for_completion(struct mxt_data *data,
			struct completion *comp, unsigned int timeout_ms)
{
	struct device *dev = &data->client->dev;
	unsigned long timeout = msecs_to_jiffies(timeout_ms);
	long ret;

	ATMEL_TRACE();

	ret = wait_for_completion_interruptible_timeout(comp, timeout);
	if (ret < 0) {
		dev_err(dev, "Wait for completion interrupted.\n");
		return -EINTR;
	} else if (ret == 0) {
		dev_err(dev, "Wait for completion timed out.\n");
		return -ETIMEDOUT;
	}
	return 0;
}

static int mxt_bootloader_read(struct mxt_data *data,
			       u8 *val, unsigned int count)
{
	int ret;
	struct i2c_msg msg;

	ATMEL_TRACE();

	msg.addr = data->bootloader_addr;
	msg.flags = data->client->flags & I2C_M_TEN;
	msg.flags |= I2C_M_RD;
	msg.len = count;
	msg.buf = val;

	ret = i2c_transfer(data->client->adapter, &msg, 1);

	if (ret == 1) {
		ret = 0;
	} else {
		ret = (ret < 0) ? ret : -EIO;
		dev_err(&data->client->dev, "%s: i2c recv failed (%d)\n",
			__func__, ret);
	}

	return ret;
}

static int mxt_bootloader_write(struct mxt_data *data,
				const u8 * const val, unsigned int count)
{
	int ret;
	struct i2c_msg msg;

	ATMEL_TRACE();

	msg.addr = data->bootloader_addr;
	msg.flags = data->client->flags & I2C_M_TEN;
	msg.len = count;
	msg.buf = (u8 *)val;

	ret = i2c_transfer(data->client->adapter, &msg, 1);
	if (ret == 1) {
		ret = 0;
	} else {
		ret = (ret < 0) ? ret : -EIO;
		dev_err(&data->client->dev, "%s: i2c send failed (%d)\n",
			__func__, ret);
	}

	return ret;
}

static int mxt_lookup_bootloader_address(struct mxt_data *data, bool retry)
{
	u8 appmode = data->client->addr;
	u8 bootloader;
	u8 family_id = 0;

	ATMEL_TRACE();

	if (data->info)
		family_id = data->info->family_id;
	else
		family_id = data->pdata->dt_info.family_id;

	switch (appmode) {
	case 0x4a:
	case 0x4b:
		/* Chips after 1664S use different scheme */
		if (retry || family_id >= 0xa2) {
			bootloader = appmode - 0x24;
			break;
		}
		/* Fall through for normal case */
	case 0x4c:
	case 0x4d:
	case 0x5a:
	case 0x5b:
		bootloader = appmode - 0x26;
		break;
	default:
		dev_err(&data->client->dev,
			"Appmode i2c address 0x%02x not found\n",
			appmode);
		return -EINVAL;
	}

	data->bootloader_addr = bootloader;
	ATMEL_DBG("Bootloader address set to %x\n", data->bootloader_addr);

	return 0;
}

static int mxt_probe_bootloader(struct mxt_data *data, bool retry)
{
	struct device *dev = &data->client->dev;
	int ret;
	u8 val;
	bool crc_failure;

	ATMEL_TRACE();

	ret = mxt_lookup_bootloader_address(data, retry);
	if (ret)
		return ret;

	ret = mxt_bootloader_read(data, &val, 1);
	if (ret)
		return ret;

	/* Check app crc fail mode */
	crc_failure = (val & ~MXT_BOOT_STATUS_MASK) == MXT_APP_CRC_FAIL;

	dev_err(dev, "Detected bootloader, status:%02X%s\n",
			val, crc_failure ? ", APP_CRC_FAIL" : "");

	return 0;
}

static u8 mxt_get_bootloader_version(struct mxt_data *data, u8 val)
{
	struct device *dev = &data->client->dev;
	u8 buf[3];

	ATMEL_TRACE();

	if (val & MXT_BOOT_EXTENDED_ID) {
		if (mxt_bootloader_read(data, &buf[0], 3) != 0) {
			dev_err(dev, "%s: i2c failure\n", __func__);
			return -EIO;
		}

		dev_info(dev, "Bootloader ID:%d Version:%d\n", buf[1], buf[2]);

		return buf[0];
	} else {
		dev_info(dev, "Bootloader ID:%d\n", val & MXT_BOOT_ID_MASK);

		return val;
	}
}

static int mxt_check_bootloader(struct mxt_data *data, unsigned int state,
				bool wait)
{
	struct device *dev = &data->client->dev;
	u8 val;
	int ret;

	ATMEL_TRACE();

recheck:
	if (wait) {
		/*
		 * In application update mode, the interrupt
		 * line signals state transitions. We must wait for the
		 * CHG assertion before reading the status byte.
		 * Once the status byte has been read, the line is deasserted.
		 */
		ret = mxt_wait_for_completion(data, &data->bl_completion,
					      MXT_FW_CHG_TIMEOUT);
		if (ret) {
			if (gpio_get_value(data->pdata->gpio_irq) == 0) {
				ATMEL_DBG("CHG is asserted (timeout)\n");
			} else {

			/*
			 * TODO: handle -EINTR better by terminating fw update
			 * process before returning to userspace by writing
			 * length 0x000 to device (if we are in
			 * WAITING_FRAME_DATA state).
			 */
			dev_err(dev, "Update wait error %d\n", ret);
			return ret;
			}
		}
	}

	ret = mxt_bootloader_read(data, &val, 1);
	if (ret)
		return ret;

	if (state == MXT_WAITING_BOOTLOAD_CMD)
		val = mxt_get_bootloader_version(data, val);

	switch (state) {
	case MXT_WAITING_BOOTLOAD_CMD:
	case MXT_WAITING_FRAME_DATA:
	case MXT_APP_CRC_FAIL:
		val &= ~MXT_BOOT_STATUS_MASK;
		break;
	case MXT_FRAME_CRC_PASS:
		if (val == MXT_FRAME_CRC_CHECK) {
			goto recheck;
		} else if (val == MXT_FRAME_CRC_FAIL) {
			dev_err(dev, "Bootloader CRC fail\n");
			return -EINVAL;
		}
		break;
	default:
		return -EINVAL;
	}

	if (val != state) {
		dev_err(dev, "Invalid bootloader state %02X != %02X\n",
			val, state);
		return -EINVAL;
	}

	return 0;
}

static int mxt_send_bootloader_cmd(struct mxt_data *data, bool unlock)
{
	int ret;
	u8 buf[2];

	ATMEL_TRACE();

	if (unlock) {
		buf[0] = MXT_UNLOCK_CMD_LSB;
		buf[1] = MXT_UNLOCK_CMD_MSB;
	} else {
		buf[0] = 0x01;
		buf[1] = 0x01;
	}

	ret = mxt_bootloader_write(data, buf, 2);
	if (ret)
		return ret;

	return 0;
}

static int __mxt_read_reg(struct i2c_client *client,
			       u16 reg, u16 len, void *val)
{
	struct i2c_msg xfer[2];
	u8 buf[2];
	int ret;
	int retry = MXT_MAX_RETRIES;

	buf[0] = reg & 0xff;
	buf[1] = (reg >> 8) & 0xff;

	/* Write register */
	xfer[0].addr = client->addr;
	xfer[0].flags = 0;
	xfer[0].len = 2;
	xfer[0].buf = buf;

	/* Read data */
	xfer[1].addr = client->addr;
	xfer[1].flags = I2C_M_RD;
	xfer[1].len = len;
	xfer[1].buf = val;

retry_read:
	ret = i2c_transfer(client->adapter, xfer, ARRAY_SIZE(xfer));
	if (ret != ARRAY_SIZE(xfer)) {
		if (--retry) {
			dev_dbg(&client->dev, "%s: i2c retry %d\n",
				__func__, MXT_MAX_RETRIES - retry);
			msleep(MXT_WAKEUP_TIME);
			goto retry_read;
		} else {
			dev_err(&client->dev, "%s: i2c read @%d failed (%d)\n",
				__func__, reg, ret);
			return -EIO;
		}
	}

        ATMEL_DBG("ret=%d, reg=0x%04X, len=%d, data=0x%02X...\n", ret, reg, len, xfer[1].buf[0]);

	return 0;
}

static int __mxt_write_reg(struct i2c_client *client, u16 reg, u16 len,
			   const void *val)
{
	u8 *buf;
	size_t count;
	int ret;
	int retry = MXT_MAX_RETRIES;

	count = len + 2;
	buf = kmalloc(count, GFP_KERNEL);
	if (!buf)
		return -ENOMEM;

	buf[0] = reg & 0xff;
	buf[1] = (reg >> 8) & 0xff;
	memcpy(&buf[2], val, len);

        ATMEL_DBG("reg=0x%04X, len=%d, data=0x%02X...\n", reg, len, buf[2]);

retry_write:
	ret = i2c_master_send(client, buf, count);
	if (ret != count) {
		if (--retry) {
			dev_dbg(&client->dev, "%s: i2c retry %d\n",
				__func__, MXT_MAX_RETRIES - retry);
			msleep(MXT_WAKEUP_TIME);
			goto retry_write;
		} else {
			dev_err(&client->dev, "%s: i2c write @%d failed (%d)\n",
				__func__, reg, ret);
			ret = -EIO;
		}
	} else {
		ret = 0;
	}

	kfree(buf);
	return ret;
}

static int mxt_write_reg(struct i2c_client *client, u16 reg, u8 val)
{
	return __mxt_write_reg(client, reg, 1, &val);
}

static struct mxt_object *mxt_get_object(struct mxt_data *data, u8 type)
{
	struct mxt_object *object;
	int i;

	/* ATMEL_TRACE(); */

        if (data->object_table) {
		for (i = 0; i < data->info->object_num; i++) {
			object = data->object_table + i;
			if (object->type == type)
				return object;
		}
        }

	dev_warn(&data->client->dev, "Invalid object type T%u\n", type);
	return NULL;
}

static void mxt_proc_t6_messages(struct mxt_data *data, u8 *msg)
{
	struct device *dev = &data->client->dev;
	u8 status = msg[1];
	u32 crc = msg[2] | (msg[3] << 8) | (msg[4] << 16);

	ATMEL_TRACE();

	if (crc != data->config_crc) {
		data->config_crc = crc;
		dev_dbg(dev, "T6 Config Checksum: 0x%06X\n", crc);
		complete(&data->crc_completion);
	}

	/* Detect transition out of reset */
	if ((data->t6_status & MXT_T6_STATUS_RESET) &&
	    !(status & MXT_T6_STATUS_RESET)) {
		ATMEL_DBG("T6 reset completion, status=%s%s%s%s%s%s%s\n",
			(status == 0) ? " OK" : "",
			(status & MXT_T6_STATUS_RESET) ? " RESET" : "",
			(status & MXT_T6_STATUS_OFL) ? " OFL" : "",
			(status & MXT_T6_STATUS_SIGERR) ? " SIGERR" : "",
			(status & MXT_T6_STATUS_CAL) ? " CAL" : "",
			(status & MXT_T6_STATUS_CFGERR) ? " CFGERR" : "",
			(status & MXT_T6_STATUS_COMSERR) ? " COMSERR" : "");
		complete(&data->reset_completion);
        }

	/* Output debug if status has changed */
	if (status != data->t6_status) {
		ATMEL_DBG("T6 Status 0x%02X%s%s%s%s%s%s%s\n",
			status,
			(status == 0) ? " OK" : "",
			(status & MXT_T6_STATUS_RESET) ? " RESET" : "",
			(status & MXT_T6_STATUS_OFL) ? " OFL" : "",
			(status & MXT_T6_STATUS_SIGERR) ? " SIGERR" : "",
			(status & MXT_T6_STATUS_CAL) ? " CAL" : "",
			(status & MXT_T6_STATUS_CFGERR) ? " CFGERR" : "",
			(status & MXT_T6_STATUS_COMSERR) ? " COMSERR" : "");
	}

	/* Save current status */
	data->t6_status = status;
}

static void mxt_input_button(struct mxt_data *data, u8 *message)
{
	struct input_dev *input = data->input_dev;
	const struct mxt_platform_data *pdata = data->pdata;
	bool button;
	int i;

	ATMEL_TRACE();

	/* do not report events if input device not yet registered */
	if (!data->enable_reporting)
		return;

	/* Active-low switch */
	for (i = 0; i < pdata->t19_num_keys; i++) {
		if (pdata->t19_keymap[i] == KEY_RESERVED)
			continue;
		button = !(message[1] & (1 << i));
		input_report_key(input, pdata->t19_keymap[i], button);
	}
}

static void mxt_input_sync(struct input_dev *input_dev, int num_touched)
{
	ATMEL_TRACE();

#if 0
	input_mt_report_pointer_emulation(input_dev, false);
#else
	input_event(input_dev, EV_KEY, BTN_TOUCH, num_touched > 0 ? 1 : 0);
#endif
	input_sync(input_dev);
}

static void mxt_proc_t9_message(struct mxt_data *data, u8 *message)
{
	struct device *dev = &data->client->dev;
	struct input_dev *input_dev = data->input_dev;
	int id;
	u8 status;
	int x;
	int y;
	int area;
	int amplitude;
	u8 vector;
	int tool;

	ATMEL_TRACE();

	/* do not report events if input device not yet registered */
	if (!data->enable_reporting)
		return;

	id = message[0] - data->T9_reportid_min;
	status = message[1];
	x = (message[2] << 4) | ((message[4] >> 4) & 0xf);
	y = (message[3] << 4) | ((message[4] & 0xf));

	/* Handle 10/12 bit switching */
	if (data->max_x < 1024)
		x >>= 2;
	if (data->max_y < 1024)
		y >>= 2;

	area = message[5];

	amplitude = message[6];
	vector = message[7];

	dev_dbg(dev,
		"[%u] %c%c%c%c%c%c%c%c x: %5u y: %5u area: %3u amp: %3u vector: %02X\n",
		id,
		(status & MXT_T9_DETECT) ? 'D' : '.',
		(status & MXT_T9_PRESS) ? 'P' : '.',
		(status & MXT_T9_RELEASE) ? 'R' : '.',
		(status & MXT_T9_MOVE) ? 'M' : '.',
		(status & MXT_T9_VECTOR) ? 'V' : '.',
		(status & MXT_T9_AMP) ? 'A' : '.',
		(status & MXT_T9_SUPPRESS) ? 'S' : '.',
		(status & MXT_T9_UNGRIP) ? 'U' : '.',
		x, y, area, amplitude, vector);

	input_mt_slot(input_dev, id);

	if (status & MXT_T9_DETECT) {
		/* Multiple bits may be set if the host is slow to read the
		 * status messages, indicating all the events that have
		 * happened */
		if (status & MXT_T9_RELEASE) {
			input_mt_report_slot_state(input_dev,
						   MT_TOOL_FINGER, 0);
			mxt_input_sync(input_dev, data->num_touched);
		}

		/* A reported size of zero indicates that the reported touch
		 * is a stylus from a linked Stylus T47 object. */
		if (area == 0) {
			area = MXT_TOUCH_MAJOR_T47_STYLUS;
			tool = MT_TOOL_PEN;
		} else {
			tool = MT_TOOL_FINGER;
		}

		/* Touch active */
		input_mt_report_slot_state(input_dev, tool, 1);
		input_report_abs(input_dev, ABS_MT_POSITION_X, x);
		input_report_abs(input_dev, ABS_MT_POSITION_Y, y);
		input_report_abs(input_dev, ABS_MT_PRESSURE, amplitude);
		input_report_abs(input_dev, ABS_MT_TOUCH_MAJOR, area);
		input_report_abs(input_dev, ABS_MT_ORIENTATION, vector);
	} else {
		/* Touch no longer active, close out slot */
		input_mt_report_slot_state(input_dev, MT_TOOL_FINGER, 0);
	}

	data->update_input = true;
}

static void mxt_proc_t100_message(struct mxt_data *data, u8 *message)
{
#ifndef PROC_COORD_EX
	struct input_dev *input_dev = data->input_dev;
	int id;
	u8 status;
	u8 tool_type;
	int x;
	int y;
	int tool;
	int ignore = 0;

	ATMEL_TRACE();

	/* do not report events if input device not yet registered */
	if (!data->enable_reporting)
		return;

	id = message[0] - data->T100_reportid_min - 2;

	/* ignore SCRSTATUS events */
	if (id < 0)
		return;

	status = message[1];
	tool_type = (status & MXT_T100_TYPE_MASK);
	x = (message[3] << 8) | message[2];
	y = (message[5] << 8) | message[4];

	if (!throttle_dbgout(&mxt_proc_t100_dbg, 20, 60000))
		dev_dbg(&data->client->dev,
			"[%u] status:%02X x:%u y:%u area:%02X amp:%02X vec:%02X\n",
			id,
			status,
			x, y,
			(data->t100_aux_area) ? message[data->t100_aux_area] : 0,
			(data->t100_aux_ampl) ? message[data->t100_aux_ampl] : 0,
			(data->t100_aux_vect) ? message[data->t100_aux_vect] : 0);

#ifdef FORCE_INVERT_XT
	/* invert x/y respectively */
	x = data->max_x - x;
	y = data->max_y - y;
#endif

	input_mt_slot(input_dev, id);

	if (status & MXT_T100_DETECT) {
		/* A reported size of zero indicates that the reported touch
		 * is a stylus from a linked Stylus T47 object. */
		if ((tool_type == MXT_T100_TYPE_ACTIVE_STYLUS)
		    || (tool_type == MXT_T100_TYPE_PASSIVE_STYLUS)) {
			tool = MT_TOOL_PEN;
		} else {
			tool = MT_TOOL_FINGER;
		}

		/* Touch active */
		if ((tool_type == MXT_T100_TYPE_FINGER)
		    || (tool_type == MXT_T100_TYPE_GLOVE)) {
			input_mt_report_slot_state(input_dev, tool, 1);
			input_report_abs(input_dev, ABS_MT_POSITION_X, x);
			input_report_abs(input_dev, ABS_MT_POSITION_Y, y);

			if (data->t100_aux_ampl)
				input_report_abs(input_dev, ABS_MT_PRESSURE,
						 message[data->t100_aux_ampl]);

			if (data->t100_aux_area)
				input_report_abs(input_dev, ABS_MT_TOUCH_MAJOR,
						 message[data->t100_aux_area]);

			if (data->t100_aux_vect)
				input_report_abs(input_dev, ABS_MT_ORIENTATION,
						 message[data->t100_aux_vect]);

			if (data->touch_state[id] != MXT_T100_STATE_TOUCH) {
				data->num_touched++;
			}
			data->touch_state[id] = MXT_T100_STATE_TOUCH;

			ATMEL_DAT("[%u] sts:%02X(%s) x:%u y:%u area:%02X amp:%02X vec:%02X\n",
				id,
				status,
				(tool_type == MXT_T100_TYPE_GLOVE) ? "glove" : "finger",
				x, y,
				(data->t100_aux_area) ? message[data->t100_aux_area]: 0,
				(data->t100_aux_ampl) ? message[data->t100_aux_ampl]: 0,
				(data->t100_aux_vect) ? message[data->t100_aux_vect]: 0);

		} else if (tool_type == MXT_T100_TYPE_ACTIVE_STYLUS) {
			/* active stylus */
			u8 active_stylus_status = message[6];

			tool = MT_TOOL_FINGER;

			if (active_stylus_status & MXT_T100_ACTIVE_STYLUS_HOVER) {
				unsigned short pressure;
				unsigned short touch_major;
				u8 button0 = 0, button1 = 0;

				input_mt_report_slot_state(input_dev, tool, 1);
				input_report_abs(input_dev, ABS_MT_POSITION_X, x);
				input_report_abs(input_dev, ABS_MT_POSITION_Y, y);

				/* this event is available */
				if (active_stylus_status & MXT_T100_ACTIVE_STYLUS_TIPSWITH) {
					/* stylus is touching */
					pressure = (message[7] << 2) + (0x3 & (message[6] >> 6));
					touch_major = MXT_TOUCH_MAJOR_T47_STYLUS;

					if (data->touch_state[id] != MXT_T100_STATE_TOUCH) {
						data->num_touched++;
					}
					data->touch_state[id] = MXT_T100_STATE_TOUCH;
				} else {
					/* stylus is hovering */
					pressure = 0;
					touch_major = 0;

					if (data->touch_state[id] == MXT_T100_STATE_TOUCH) {
						data->num_touched--;
					}
					data->touch_state[id] = MXT_T100_STATE_HOVER;
				}
				input_report_abs(input_dev, ABS_MT_PRESSURE, pressure);
				input_report_abs(input_dev, ABS_MT_TOUCH_MAJOR, touch_major);

				/* buttons */
				if (message[6] & MXT_T100_ACTIVE_STYLUS_BUTTON0) {
					button0 = 1;
				}
				input_report_key(input_dev, BTN_STYLUS, button0);

				if (message[6] & MXT_T100_ACTIVE_STYLUS_BUTTON1) {
					button1 = 1;
				}
				input_report_key(input_dev, BTN_STYLUS2, button1);

				ATMEL_DAT("[%u] sts:%02X(%s) x:%u y:%u pressure:%u batt:%u, b0=%u,b1=%u\n",
					id,
					status, "apen",
					x, y,
					pressure, message[8],
					button0, button1);
			} else {
				/* this event should be ignored */
				ignore = 1;

				ATMEL_DBG("[%u] sts:%02X(%s) aux=0x%02X, 0x%02X, 0x%02X\n",
					id,
					status, "apen-out",
					message[6], message[7], message[8]);
			}

		} else if (tool_type == MXT_T100_TYPE_PASSIVE_STYLUS) {
			/* passive stylus */
			input_mt_report_slot_state(input_dev, tool, 1);
			input_report_abs(input_dev, ABS_MT_POSITION_X, x);
			input_report_abs(input_dev, ABS_MT_POSITION_Y, y);
			input_report_abs(input_dev, ABS_MT_PRESSURE, 50);
			input_report_abs(input_dev, ABS_MT_TOUCH_MAJOR,
					 MXT_TOUCH_MAJOR_T47_STYLUS);

			if (data->touch_state[id] != MXT_T100_STATE_TOUCH) {
				data->num_touched++;
			}
			data->touch_state[id] = MXT_T100_STATE_TOUCH;

			ATMEL_DAT("[%u] sts:%02X(%s) x:%u y:%u\n",
				  id,
				  status, "ppen",
				  x, y);
		} else {
			/* nop */
		}
	} else {
		ATMEL_DAT("[%u] sts:%02X touch not detected\n",
			id,
			status);

		/* Touch no longer active, close out slot */
		input_mt_report_slot_state(input_dev, MT_TOOL_FINGER, 0);

		if (data->touch_state[id] == MXT_T100_STATE_TOUCH) {
			data->num_touched--;
		}
		data->touch_state[id] = MXT_T100_STATE_NONE;
	}

	if (ignore == 0) {
		data->update_input = true;
	}
#else
	int slot;
	struct TpPos_t pos;

	ATMEL_TRACE();

	/* do not report events if input device not yet registered */
	if (!data->enable_reporting)
		return;

	slot = message[0] - data->T100_reportid_min - 2;

	/* ignore SCRSTATUS events */
	if (slot < 0)
		return;

	TpIf_convertPos(&pos, slot, message);
	InputSubsystem_store(&pos);
	data->update_input = true;
#endif
}

static void mxt_proc_t15_messages(struct mxt_data *data, u8 *msg)
{
	struct input_dev *input_dev = data->input_dev;
	int key;
	bool curr_state, new_state;
	bool sync = false;
	unsigned long keystates = le32_to_cpu(msg[2]);

	ATMEL_TRACE();

	/* do not report events if input device not yet registered */
	if (!data->enable_reporting)
		return;

	for (key = 0; key < data->pdata->t15_num_keys; key++) {
		curr_state = test_bit(key, &data->t15_keystatus);
		new_state = test_bit(key, &keystates);

		ud_set_id(&button_ud_stats, key);

		if (!curr_state && new_state) {
			__set_bit(key, &data->t15_keystatus);
			input_event(input_dev, EV_KEY,
				    data->pdata->t15_keymap[key], 1);
			sync = true;
			ud_log_status(&button_ud_stats, true);
		} else if (curr_state && !new_state) {
			__clear_bit(key, &data->t15_keystatus);
			input_event(input_dev, EV_KEY,
				    data->pdata->t15_keymap[key], 0);
			sync = true;
			ud_log_status(&button_ud_stats, false);
		}
	}

	if (sync)
		input_sync(input_dev);
}

static void mxt_proc_t42_messages(struct mxt_data *data, u8 *msg)
{
	struct device *dev = &data->client->dev;
	u8 status = msg[1];

	ATMEL_TRACE();

	if (status & MXT_T42_MSG_TCHSUP)
		dev_info(dev, "T42 suppress\n");
	else
		dev_info(dev, "T42 normal\n");
}

static void mxt_proc_t93_messages(struct mxt_data *data, u8 *msg)
{
	u8 status = msg[1];

	ATMEL_TRACE();

	if (status & 0x2) {
		struct device *dev = &data->client->dev;
		struct input_dev *input_dev = data->input_dev;

		input_report_key(input_dev, KEY_POWER, 1);
		input_report_key(input_dev, KEY_POWER, 0);
		input_sync(input_dev);

		dev_dbg(dev, "T93 status %s\n", (status & 0x2) ? "DBLTAP" : "");
	}
}

static int mxt_proc_t48_messages(struct mxt_data *data, u8 *msg)
{
	struct device *dev = &data->client->dev;
	u8 status, state;

	ATMEL_TRACE();

	status = msg[1];
	state  = msg[4];

	dev_dbg(dev, "T48 state %d status %02X %s%s%s%s%s\n",
			state,
			status,
			(status & 0x01) ? "FREQCHG " : "",
			(status & 0x02) ? "APXCHG " : "",
			(status & 0x04) ? "ALGOERR " : "",
			(status & 0x10) ? "STATCHG " : "",
			(status & 0x20) ? "NLVLCHG " : "");

	return 0;
}

static void mxt_proc_t63_messages(struct mxt_data *data, u8 *msg)
{
	struct device *dev = &data->client->dev;
	struct input_dev *input_dev = data->input_dev;
	u8 id;
	u16 x, y;
	u8 pressure;

	ATMEL_TRACE();

	/* do not report events if input device not yet registered */
	if (!data->enable_reporting)
		return;

	/* stylus slots come after touch slots */
	id = data->num_touchids + (msg[0] - data->T63_reportid_min);

	if (id < 0 || id > (data->num_touchids + data->num_stylusids)) {
		dev_err(dev, "invalid stylus id %d, max slot is %d\n",
			id, data->num_stylusids);
		return;
	}

	x = msg[3] | (msg[4] << 8);
	y = msg[5] | (msg[6] << 8);
	pressure = msg[7] & MXT_T63_STYLUS_PRESSURE_MASK;

	dev_dbg(dev,
		"[%d] %c%c%c%c x: %d y: %d pressure: %d stylus:%c%c%c%c\n",
		id,
		(msg[1] & MXT_T63_STYLUS_SUPPRESS) ? 'S' : '.',
		(msg[1] & MXT_T63_STYLUS_MOVE)     ? 'M' : '.',
		(msg[1] & MXT_T63_STYLUS_RELEASE)  ? 'R' : '.',
		(msg[1] & MXT_T63_STYLUS_PRESS)    ? 'P' : '.',
		x, y, pressure,
		(msg[2] & MXT_T63_STYLUS_BARREL) ? 'B' : '.',
		(msg[2] & MXT_T63_STYLUS_ERASER) ? 'E' : '.',
		(msg[2] & MXT_T63_STYLUS_TIP)    ? 'T' : '.',
		(msg[2] & MXT_T63_STYLUS_DETECT) ? 'D' : '.');

	input_mt_slot(input_dev, id);

	if (msg[2] & MXT_T63_STYLUS_DETECT) {
		input_mt_report_slot_state(input_dev, MT_TOOL_PEN, 1);
		input_report_abs(input_dev, ABS_MT_POSITION_X, x);
		input_report_abs(input_dev, ABS_MT_POSITION_Y, y);
		input_report_abs(input_dev, ABS_MT_PRESSURE, pressure);
	} else {
		input_mt_report_slot_state(input_dev, MT_TOOL_PEN, 0);
	}

	input_report_key(input_dev, BTN_STYLUS,
			 (msg[2] & MXT_T63_STYLUS_ERASER));
	input_report_key(input_dev, BTN_STYLUS2,
			 (msg[2] & MXT_T63_STYLUS_BARREL));

	mxt_input_sync(input_dev, data->num_touched);
}

static void mxt_proc_t25_messages(struct mxt_data *data, u8 *msg)
{
	ATMEL_TRACE();

	memcpy(data->T25_result, &msg[1], sizeof(data->T25_result));
	complete(&data->selftest_completion);

	return;
}

static int mxt_proc_message(struct mxt_data *data, u8 *message)
{
	u8 report_id = message[0];
	bool dump = data->debug_enabled;

	ATMEL_TRACE();

	if (report_id == MXT_RPTID_NOMSG)
		return 0;

	if (report_id == data->T6_reportid) {
		mxt_proc_t6_messages(data, message);
	} else if (report_id >= data->T9_reportid_min
	    && report_id <= data->T9_reportid_max) {
		mxt_proc_t9_message(data, message);
	} else if (report_id >= data->T100_reportid_min
	    && report_id <= data->T100_reportid_max) {
		mxt_proc_t100_message(data, message);
	} else if (report_id == data->T19_reportid) {
		mxt_input_button(data, message);
		data->update_input = true;
	} else if (report_id >= data->T63_reportid_min
		   && report_id <= data->T63_reportid_max) {
		mxt_proc_t63_messages(data, message);
	} else if (report_id >= data->T42_reportid_min
		   && report_id <= data->T42_reportid_max) {
		mxt_proc_t42_messages(data, message);
	} else if (report_id == data->T48_reportid) {
		mxt_proc_t48_messages(data, message);
	} else if (report_id >= data->T15_reportid_min
		   && report_id <= data->T15_reportid_max) {
		mxt_proc_t15_messages(data, message);
	} else if (report_id == data->T93_reportid) {
		mxt_proc_t93_messages(data, message);
	} else if (report_id == data->T25_reportid) {
		mxt_proc_t25_messages(data, message);
	} else {
		dump = true;
	}

	if (dump)
		mxt_dump_message(data, message);

	if (data->debug_v2_enabled)
		mxt_debug_msg_add(data, message);

	return 1;
}

static int mxt_read_and_process_messages(struct mxt_data *data, u8 count)
{
	struct device *dev = &data->client->dev;
	int ret;
	int i;
	u8 num_valid = 0;

	ATMEL_TRACE();

	/* Safety check for msg_buf */
	if (count > data->max_reportid)
		return -EINVAL;

	/* Process remaining messages if necessary */
	ret = __mxt_read_reg(data->client, data->T5_address,
				data->T5_msg_size * count, data->msg_buf);
	if (ret) {
		dev_err(dev, "Failed to read %u messages (%d)\n", count, ret);
		return ret;
	}

	for (i = 0;  i < count; i++) {
		ret = mxt_proc_message(data,
			data->msg_buf + data->T5_msg_size * i);

		if (ret == 1)
			num_valid++;
	}

	/* return number of messages read */
	return num_valid;
}

static int zeroIntCnt = 0;

static irqreturn_t mxt_process_messages_t44(struct mxt_data *data)
{
	struct device *dev = &data->client->dev;
	int ret;
	u8 count, num_left;

	ATMEL_TRACE();

	if (!data->msg_buf) {
		dev_err(dev, "Message buffer not allocated!!!\n");
		return IRQ_NONE;
	}

	/* Read T44 and T5 together */
	ret = __mxt_read_reg(data->client, data->T44_address,
		data->T5_msg_size + 1, data->msg_buf);
	if (ret) {
		dev_err(dev, "Failed to read T44 and T5 (%d)\n", ret);
		return IRQ_NONE;
	}

	count = data->msg_buf[0];

	if (count == 0) {
		zeroIntCnt++;
		if (zeroIntCnt > 100) {
			dev_warn(dev, "Interrupt triggered but zero messages\n");
			mxt_irq_enable(data, false);
			return IRQ_HANDLED;
		}
		/* dev_warn(dev, "Interrupt triggered but zero messages\n"); */
		return IRQ_NONE;
	} else if (count > data->max_reportid) {
		dev_err(dev, "T44 count %d exceeded max report id\n", count);
		count = data->max_reportid;
	}
	zeroIntCnt = 0;

	/* Process first message */
	ret = mxt_proc_message(data, data->msg_buf + 1);
	if (ret < 0) {
		dev_warn(dev, "Unexpected invalid message\n");
		return IRQ_NONE;
	}

	num_left = count - 1;

	/* Process remaining messages if necessary */
	if (num_left) {
		ret = mxt_read_and_process_messages(data, num_left);
		if (ret < 0)
			goto end;
		else if (ret != num_left)
			dev_warn(dev, "Unexpected invalid message\n");
	}

end:

	if (data->update_input) {
#ifndef PROC_COORD_EX
		mxt_input_sync(data->input_dev, data->num_touched);
#else
		InputSubsystem_update();
		if (mPlugin != NULL) {
			mPlugin->process(&mCurrentEvent, 1, 1);
		}

		/* when disable TP sometimes, don't report coordinate */
		if((TP_DISABLE == atomic_read(&data->nfc_enable_tp))
				|| (TP_DISABLE == atomic_read(&data->enable_tp)))
		mxt_reset_slots(data);

		else
			InputSubsystem_enqueue(data->input_dev, &mCurrentEvent);
#endif
		data->update_input = false;
	}

	return IRQ_HANDLED;
}

static int mxt_process_messages_until_invalid(struct mxt_data *data)
{
	struct device *dev = &data->client->dev;
	int num_handled, processed = 0;
	bool pin_high;

	ATMEL_TRACE();

	do {
		/* It appears host has to read "beyond" the message */
		/* with invalid id to ensure CHG line gets deasserted */
		num_handled = mxt_read_and_process_messages(data, 3);
		processed += num_handled;
	} while (num_handled > 0);

	if (processed)
		dev_dbg(dev, "processed %d messages\n", processed);

	pin_high = gpio_get_value(data->pdata->gpio_irq) == 1;
	if (!pin_high) {
		dev_err(dev, "CHG pin still asserted\n");
		return -1;
	} else {
		return 0;
	}
}

static irqreturn_t mxt_process_messages(struct mxt_data *data)
{
	int total_handled, num_handled;
	u8 count = data->last_message_count;

	ATMEL_TRACE();

	if (count < 1 || count > data->max_reportid)
		count = 1;

	/* include final invalid message */
	total_handled = mxt_read_and_process_messages(data, count + 1);
	if (total_handled < 0)
		return IRQ_NONE;
	/* if there were invalid messages, then we are done */
	else if (total_handled <= count)
		goto update_count;

	/* read two at a time until an invalid message or else we reach
	 * reportid limit */
	do {
		num_handled = mxt_read_and_process_messages(data, 2);
		if (num_handled < 0)
			return IRQ_NONE;

		total_handled += num_handled;

		if (num_handled < 2)
			break;
	} while (total_handled < data->num_touchids);

update_count:
	data->last_message_count = total_handled;

	if (data->enable_reporting && data->update_input) {
		mxt_input_sync(data->input_dev, data->num_touched);
		data->update_input = false;
	}

	return IRQ_HANDLED;
}

static irqreturn_t mxt_interrupt(int irq, void *dev_id)
{
	struct mxt_data *data = dev_id;

	ATMEL_TRACE();

	if (data->in_bootloader) {
		/* bootloader state transition completion */
		complete(&data->bl_completion);
	} else {
		if (data->enable_fw_update == 0) {
			mxt_queue_irq(data);
			flush_delayed_work(&data->irq_work);
		} else {
			/* if updating fw, process appmode interruption directry */
			if (data->T44_address)
				(void)mxt_process_messages_t44(data);
			else
				(void)mxt_process_messages(data);
		}
	}

	return IRQ_HANDLED;
}

static int mxt_t6_command(struct mxt_data *data, u16 cmd_offset,
			  u8 value, bool wait)
{
	u16 reg;
	u8 command_register;
	int timeout_counter = 0;
	int ret;

	ATMEL_TRACE();
	ATMEL_DBG("cmd=%d\n", cmd_offset);

	reg = data->T6_address + cmd_offset;

	ret = mxt_write_reg(data->client, reg, value);
	if (ret)
		return ret;

	if (!wait)
		return 0;

	do {
		msleep(20);
		ret = __mxt_read_reg(data->client, reg, 1, &command_register);
		if (ret)
			return ret;
	} while ((command_register != 0) && (timeout_counter++ <= 100));

	if (timeout_counter > 100) {
		ATMEL_ERR("Command failed!\n");
		return -EIO;
	}

	return 0;
}

static int mxt_t6_command_fast(struct mxt_data *data, u16 cmd_offset,
			       u8 value, bool wait)
{
	u16 reg;
	u8 command_register;
	int timeout_counter = 0;
	int ret;

	ATMEL_TRACE();
	ATMEL_DBG("cmd=%d\n", cmd_offset);

	reg = data->T6_address + cmd_offset;

	ret = mxt_write_reg(data->client, reg, value);
	if (ret)
		return ret;

	if (!wait)
		return 0;

	do {
		mdelay(1);
		ret = __mxt_read_reg(data->client, reg, 1, &command_register);
		if (ret)
			return ret;
	} while ((command_register != 0) && (timeout_counter++ <= 100));

	if (timeout_counter > 100) {
		ATMEL_ERR("Command failed!\n");
		return -EIO;
	}

	return 0;
}

static int mxt_read_revision_id(struct mxt_data *data, u8 *revision)
{
	struct mxt_object *object;
	int error = -EINVAL;

	ATMEL_TRACE();

	object = mxt_get_object(data, MXT_DEBUG_DIAGNOSTIC_T37);
	if (!object)
		goto err_exit;

	error = mxt_t6_command(data, MXT_COMMAND_DIAGNOSTIC,
				MXT_IDENT_VALUE, true);
	if (error)
		goto err_exit;

	error = __mxt_read_reg(data->client,
			       object->start_address + MXT_T37_REVISION,
			       sizeof(*revision), revision);
err_exit:
	return error;
}

static int mxt_soft_reset(struct mxt_data *data)
{
	int ret = 0;

	ATMEL_TRACE();

	ATMEL_MSG("Resetting chip\n");

	reinit_completion(&data->reset_completion);

	ret = mxt_t6_command(data, MXT_COMMAND_RESET, MXT_RESET_VALUE, false);
	if (ret)
		return ret;

	ret = mxt_wait_for_completion(data, &data->reset_completion,
				      MXT_RESET_TIMEOUT);
	if (ret)
		return ret;

	return 0;
}

static void mxt_update_crc(struct mxt_data *data, u8 cmd, u8 value)
{
	ATMEL_TRACE();

	/* on failure, CRC is set to 0 and config will always be downloaded */
	data->config_crc = 0;
	reinit_completion(&data->crc_completion);

	mxt_t6_command(data, cmd, value, true);

	/* Wait for crc message. On failure, CRC is set to 0 and config will
	 * always be downloaded */
	mxt_wait_for_completion(data, &data->crc_completion, MXT_CRC_TIMEOUT);
}

static void mxt_calc_crc24(u32 *crc, u8 firstbyte, u8 secondbyte)
{
	static const unsigned int crcpoly = 0x80001B;
	u32 result;
	u32 data_word;

	/* ATMEL_TRACE(); */

	data_word = (secondbyte << 8) | firstbyte;
	result = ((*crc << 1) ^ data_word);

	if (result & 0x1000000)
		result ^= crcpoly;

	*crc = result;
}

static u32 mxt_calculate_crc(u8 *base, off_t start_off, off_t end_off)
{
	u32 crc = 0;
	u8 *ptr = base + start_off;
	u8 *last_val = base + end_off - 1;

	ATMEL_TRACE();

	if (end_off < start_off)
		return -EINVAL;

	while (ptr < last_val) {
		mxt_calc_crc24(&crc, *ptr, *(ptr + 1));
		ptr += 2;
	}

	/* if len is odd, fill the last byte with 0 */
	if (ptr == last_val)
		mxt_calc_crc24(&crc, *ptr, 0);

	/* Mask to 24-bit */
	crc &= 0x00FFFFFF;

	return crc;
}

static int mxt_check_retrigen(struct mxt_data *data)
{
	struct i2c_client *client = data->client;
	int error;
	int val;

	ATMEL_TRACE();

	if (data->pdata->irqflags & IRQF_TRIGGER_LOW)
		return 0;

	if (data->T18_address) {
		error = __mxt_read_reg(client,
				       data->T18_address + MXT_COMMS_CTRL,
				       1, &val);
		if (error)
			return error;

		if (val & MXT_COMMS_RETRIGEN)
			return 0;
	}

	dev_warn(&client->dev, "Enabling RETRIGEN workaround\n");
	data->use_retrigen_workaround = true;
	return 0;
}

static int force_update = 0;
static int mxt_check_for_fw_update(struct mxt_data *data, const unsigned char *buf)
{
	int must_update = 0;
	unsigned char img[5];
	u16 dev_ver;
	long img_ver;
	int version_offset = 12;
	int ret;
	int error;
	u8 fw_update_flag = 0;

	memset(img, 0, sizeof(img));

	/*
	 * priority 1: device error
	 * if touch device can not communicate to host,
	 * fw must be flashed
	 */
	error = mxt_read_fw_version(data, &dev_ver);
	if (error != 0) {
		ATMEL_ERR("fw version read failed\n");
		must_update = 1;
		goto ver_check_end;
	}
	ATMEL_MSG("[DEV] %04X\n", dev_ver);

	/* read and parse version */
	memcpy(img, buf+version_offset, sizeof(img)-1);
	img[4] = 0;
	error = kstrtol(img, 16, &img_ver);
	if (error) {
		ATMEL_ERR("cannot parse version in fw image, v=%s\n", img);
		must_update = 0;
		goto ver_check_err;
	}
	ATMEL_MSG("[IMG] %04X\n", (u16)img_ver);

	/*
	 * priority 2: fw update flag in nvm
	 * if fw update flag in nvm == 0, fw must NOT be flashed.
	 */
#ifdef ENABLE_CFGDRV
	ret = cfgdrv_read(D_HCM_E_TP_FW_REFLASH_FLAG, sizeof(fw_update_flag),
			  &fw_update_flag);
#else
	ret = fw_update_flag = 0;
#endif
	if (ret == 0) {
		ATMEL_MSG("fw_update_flag=%d\n", fw_update_flag);
		if (fw_update_flag == 0) {
			must_update = 0;
			goto ver_check_end;
		}
	}

	/*
	 * priority 3: version value
	 * if base/config version of fw image is newer than the device's one,
	 * fw must be flashed
	 */
	if ((u16)img_ver > dev_ver) {
		ATMEL_MSG("img version is newer than the device's one\n");
		must_update = 1;
		goto ver_check_end;
	}

ver_check_end:
	if (force_update) {
		ATMEL_MSG("force_update=%d\n", force_update);
		must_update = 1;
	}
	if (must_update == 1) {
		/* if firmware will be updated, config should also be updated */
		force_update = 1;
	}
ver_check_err:
	ATMEL_MSG("must_update=%d\n", must_update);
	return must_update;
}

static int mxt_check_for_cfg_update(struct mxt_data *data, struct firmware *cfg)
{
	int must_update = 0;
	u8 userdata_img[6];
	u32 dev_ver;
	u32 img_ver;
	u16 dev_local_rev;
	u16 img_local_rev;
	char *version_offset;
	int i;
	int offset;
	int ret;
	int error;
	u8 fw_update_flag = 0;

	/*
	 * priority 1: device error
	 * if touch device can not communicate to host,
	 * fw must be flashed
	 */
	error = mxt_read_cfg_version(data, &dev_ver, &dev_local_rev);
	if (error != 0) {
		ATMEL_ERR("cfg version read failed\n");
		must_update = 1;
		goto ver_check_end;
	}
	ATMEL_MSG("[DEV] %08X.%04X\n", dev_ver, dev_local_rev);

	/* read and parse version */
	version_offset = strstr(cfg->data, "0026 0000");
	if (version_offset == NULL) {
		ATMEL_ERR("version data not found\n");
		must_update = 0;
		goto ver_check_end;
	}
	version_offset += 27;
	memset(userdata_img, 0, sizeof(userdata_img));
	for (i = 0; i < sizeof(userdata_img)/sizeof(userdata_img[0]); i++) {
		ret = sscanf(version_offset, "%hhx%n",
			     (unsigned char *)&userdata_img[i], &offset);
		if (ret != 1) {
			ATMEL_ERR("Bad format\n");
			must_update = 0;
			goto ver_check_end;
		}
		version_offset += offset;
	}
	img_ver = be32_to_cpu(*(unsigned int *)&userdata_img[0]);
	img_local_rev = be16_to_cpu(*(unsigned short *)&userdata_img[4]);

	ATMEL_MSG("[IMG] %08X.%04X\n", img_ver, img_local_rev);

	/*
	 * priority 2: local revision number (developing version)
	 * if local revision number of fw image is newer than the device's one,
	 * fw must be flashed
	 */
	if (img_local_rev > dev_local_rev) {
		ATMEL_MSG("img local revisiton is newer than the device's one\n");
		must_update = 1;
		goto ver_check_end;
	}

	/*
	 * priority 3: fw update flag in nvm
	 * if fw update flag in nvm == 0, RTM revision fw must NOT be flashed.
	 * RTM revision is 0x0000
	 */
#ifdef ENABLE_CFGDRV
	ret = cfgdrv_read(D_HCM_E_TP_FW_REFLASH_FLAG, sizeof(fw_update_flag),
			  &fw_update_flag);
#else
	ret = fw_update_flag = img_local_rev = 0;
#endif
	if (ret == 0) {
		ATMEL_MSG("fw_update_flag=%d\n", fw_update_flag);
		if ((fw_update_flag == 0) && (img_local_rev == 0)) {
			must_update = 0;
			goto ver_check_end;
		}
	}

	/*
	 * priority 4: base/config version
	 * if base/config version of fw image is newer than the device's one,
	 * fw must be flashed
	 */
	if ((img_local_rev == 0) && (img_ver > dev_ver)) {
		ATMEL_MSG("img base/config version is newer than the device's one\n");
		must_update = 1;
		goto ver_check_end;
	}

ver_check_end:
	if (force_update) {
		ATMEL_MSG("force_update=%d\n", force_update);
		must_update = 1;
	}

	ATMEL_MSG("must_update=%d\n", must_update);
	return must_update;
}

/*
 * mxt_check_reg_init - download configuration to chip
 *
 * Atmel Raw Config File Format
 *
 * The first four lines of the raw config file contain:
 *  1) Version
 *  2) Chip ID Information (first 7 bytes of device memory)
 *  3) Chip Information Block 24-bit CRC Checksum
 *  4) Chip Configuration 24-bit CRC Checksum
 *
 * The rest of the file consists of one line per object instance:
 *   <TYPE> <INSTANCE> <SIZE> <CONTENTS>
 *
 *   <TYPE> - 2-byte object type as hex
 *   <INSTANCE> - 2-byte object instance number as hex
 *   <SIZE> - 2-byte object size as hex
 *   <CONTENTS> - array of <SIZE> 1-byte hex values
 */
#define TYPEA_RAW   "/vendor/firmware/atmel_cfg_typeA.raw"
#define TYPEB_RAW   "/vendor/firmware/atmel_cfg_typeB.raw"
#define TYPEC_RAW   "/vendor/firmware/atmel_cfg_typeC.raw"
static int mxt_check_reg_init(struct mxt_data *data) /* close */
{
	struct device *dev = &data->client->dev;
	struct mxt_info cfg_info;
	struct mxt_object *object;
#if (USE_REQUEST_FIRMWARE == 1)
	const struct firmware *cfg = NULL;
#else
	struct firmware *cfg = NULL;
#endif
	int ret;
	int offset;
	int data_pos;
	int byte_offset;
	int i;
	int cfg_start_ofs;
	u32 info_crc, config_crc, calculated_crc;
	u8 *config_mem;
	size_t config_mem_size;
	unsigned int type, instance, size;
	u8 val;
	u16 reg;
	char panel_name[256] = {'\0'};
	char raw_path[128] = {'\0'};
	u8 tp_cfg;

	ATMEL_TRACE();
	if ((data->cfg_name == NULL)
		&& (data->T38_address != 0)) {
		mdss_get_panel_name(panel_name);
		ret = __mxt_read_reg(data->client,
					data->T38_address + MXT_T38_VENDOR_VER,
					sizeof(tp_cfg), &tp_cfg);
		if (ret) {
			ATMEL_ERR("Unable to read config id\n");
		}
		if (!strcmp(TYPEA,panel_name)) {
			strcpy(raw_path, TYPEA_RAW);
			pr_info("raw_path:%s.\n",raw_path);
		}else if (!strcmp(TYPEB,panel_name)) {
		    if(tp_cfg != 0x0A){
			    strcpy(raw_path, TYPEB_RAW);
			    pr_info("raw_path:%s.\n",raw_path);
			} else {
			    strcpy(raw_path, TYPEC_RAW);
			    pr_info("raw_path:%s.\n",raw_path);
			}
		}else {
			ATMEL_ERR("Failed to compatible LCD panel %s.\n",__func__);
		}
		mxt_update_file_name(dev, &data->cfg_name, raw_path, strlen(raw_path));
	}
	/* store default configuration */
	ATMEL_DBG("T38_address=0x%04X, cfg_name=%s\n",
		 data->T38_address, data->cfg_name);

	if (!data->cfg_name)
		return 0;

	if (data->enable_fw_update == 0) {
		ATMEL_MSG("skip update cfg because flag is off\n");
		return 0;
	}

#if (USE_REQUEST_FIRMWARE == 1)
	ret = request_firmware(&cfg, data->cfg_name, dev);
	if (ret < 0) {
		dev_err(dev, "Failure to request config file %s\n",
			data->cfg_name);
		return 0;
	}
#else
	{
		const int CFG_SIZE = 65536;
		struct file *filep;
		u8 *cfgbuf;
		int num_read = 0;
		int cfgsize;
		filep = filp_open(data->cfg_name, O_RDONLY|O_LARGEFILE, 0);
		if (IS_ERR(filep)) {
			ATMEL_ERR("can't open %s", data->cfg_name);
			return 0;
		}

		cfgsize = i_size_read(file_inode(filep));
		if ((cfgsize <= 0) || (cfgsize > CFG_SIZE)) {
			fput(filep);
			return -EINVAL;
		}

		cfgbuf = kzalloc(CFG_SIZE, GFP_KERNEL);
		if (cfgbuf == NULL) {
			fput(filep);
			return 0;
		}

		num_read = kernel_read(filep, 0, cfgbuf, cfgsize);
		fput(filep);
		kfree(data->cfg_name);
		data->cfg_name = NULL;

		cfg = kzalloc(sizeof(struct firmware), GFP_KERNEL);
		if (cfg == NULL) {
			ret = -ENOMEM;
			kfree(cfgbuf);
			return ret;
		}
		cfg->size = num_read;
		cfg->data = cfgbuf;

		ATMEL_DBG("cfg num_read=%d\n", num_read);

		if (mxt_check_for_cfg_update(data, cfg) == 0) {
			ret = 0;
			goto release;
		}
	}
#endif

	mxt_update_crc(data, MXT_COMMAND_REPORTALL, 1);

	if (strncmp(cfg->data, MXT_CFG_MAGIC, strlen(MXT_CFG_MAGIC))) {
		dev_err(dev, "Unrecognised config file\n");
		ret = -EINVAL;
		goto release;
	}

	data_pos = strlen(MXT_CFG_MAGIC);

	/* Load information block and check */
	for (i = 0; i < sizeof(struct mxt_info); i++) {
		ret = sscanf(cfg->data + data_pos, "%hhx%n",
			     (unsigned char *)&cfg_info + i,
			     &offset);
		if (ret != 1) {
			dev_err(dev, "Bad format\n");
			ret = -EINVAL;
			goto release;
		}

		data_pos += offset;
	}

	if (cfg_info.family_id != data->info->family_id) {
		dev_err(dev, "Family ID mismatch!\n");
		ret = -EINVAL;
		goto release;
	}

	if (cfg_info.variant_id != data->info->variant_id) {
		dev_err(dev, "Variant ID mismatch!\n");
		ret = -EINVAL;
		goto release;
	}

	/* Read CRCs */
	ret = sscanf(cfg->data + data_pos, "%x%n", &info_crc, &offset);
	if (ret != 1) {
		dev_err(dev, "Bad format: failed to parse Info CRC\n");
		ret = -EINVAL;
		goto release;
	}
	data_pos += offset;

	ret = sscanf(cfg->data + data_pos, "%x%n", &config_crc, &offset);
	if (ret != 1) {
		dev_err(dev, "Bad format: failed to parse Config CRC\n");
		ret = -EINVAL;
		goto release;
	}
	data_pos += offset;

#if 0
	/* The Info Block CRC is calculated over mxt_info and the object table
	 * If it does not match then we are trying to load the configuration
	 * from a different chip or firmware version, so the configuration CRC
	 * is invalid anyway. */
	if (info_crc == data->info_crc) {
		if (config_crc == 0 || data->config_crc == 0) {
			dev_info(dev, "CRC zero, attempting to apply config\n");
		} else if (config_crc == data->config_crc) {
			dev_info(dev, "Config CRC 0x%06X: OK\n",
				 data->config_crc);
			ret = 0;
			goto release;
		} else {
			dev_info(dev, "Config CRC 0x%06X: does not match file 0x%06X\n",
				 data->config_crc, config_crc);
		}
	} else {
		dev_warn(dev,
			 "Warning: Info CRC error - device=0x%06X file=0x%06X\n",
			data->info_crc, info_crc);
	}
#endif

	/* Malloc memory to store configuration */
	cfg_start_ofs = MXT_OBJECT_START
		+ data->info->object_num * sizeof(struct mxt_object)
		+ MXT_INFO_CHECKSUM_SIZE;
	config_mem_size = data->mem_size - cfg_start_ofs;
	config_mem = kzalloc(config_mem_size, GFP_KERNEL);
	if (!config_mem) {
		dev_err(dev, "Failed to allocate memory\n");
		ret = -ENOMEM;
		goto release;
	}

	while (data_pos < cfg->size) {
		/* Read type, instance, length */
		ret = sscanf(cfg->data + data_pos, "%x %x %x%n",
			     &type, &instance, &size, &offset);
		if (ret == 0) {
			/* EOF */
			break;
		} else if (ret != 3) {
			dev_err(dev, "Bad format: failed to parse object\n");
			ret = -EINVAL;
			goto release_mem;
		}
		data_pos += offset;

		object = mxt_get_object(data, type);
		if (!object) {
			/* Skip object */
			for (i = 0; i < size; i++) {
				ret = sscanf(cfg->data + data_pos, "%hhx%n",
					     &val,
					     &offset);
				data_pos += offset;
			}
			continue;
		}

		if (size > mxt_obj_size(object)) {
			/* Either we are in fallback mode due to wrong
			 * config or config from a later fw version,
			 * or the file is corrupt or hand-edited */
			dev_warn(dev, "Discarding %zu byte(s) in T%u\n",
				 size - mxt_obj_size(object), type);
		} else if (mxt_obj_size(object) > size) {
			/* If firmware is upgraded, new bytes may be added to
			 * end of objects. It is generally forward compatible
			 * to zero these bytes - previous behaviour will be
			 * retained. However this does invalidate the CRC and
			 * will force fallback mode until the configuration is
			 * updated. We warn here but do nothing else - the
			 * malloc has zeroed the entire configuration. */
			dev_warn(dev, "Zeroing %zu byte(s) in T%d\n",
				 mxt_obj_size(object) - size, type);
		}

		if (instance >= mxt_obj_instances(object)) {
			dev_err(dev, "Object instances exceeded!\n");
			ret = -EINVAL;
			goto release_mem;
		}

		reg = object->start_address + mxt_obj_size(object) * instance;

		for (i = 0; i < size; i++) {
			ret = sscanf(cfg->data + data_pos, "%hhx%n",
				     &val,
				     &offset);
			if (ret != 1) {
				dev_err(dev, "Bad format in T%d\n", type);
				ret = -EINVAL;
				goto release_mem;
			}
			data_pos += offset;

			if (i > mxt_obj_size(object))
				continue;

			byte_offset = reg + i - cfg_start_ofs;

			if ((byte_offset >= 0)
			    && (byte_offset < config_mem_size)) {
				*(config_mem + byte_offset) = val;
			} else {
				dev_err(dev, "Bad object: reg:%d, T%d, ofs=%d\n",
					reg, object->type, byte_offset);
				ret = -EINVAL;
				goto release_mem;
			}
		}
	}

	/* calculate crc of the received configs (not the raw config file) */
	if (data->T7_address < cfg_start_ofs) {
		dev_err(dev, "Bad T7 address, T7addr = %x, config offset %x\n",
			data->T7_address, cfg_start_ofs);
		ret = 0;
		goto release_mem;
	}

	calculated_crc = mxt_calculate_crc(config_mem,
					   data->T7_address - cfg_start_ofs,
					   config_mem_size);

	if (config_crc > 0 && (config_crc != calculated_crc)) {
		ATMEL_DBG("Config CRC error, calculated=%06X, file=%06X\n",
			  calculated_crc, config_crc);
	}

	/* Write configuration as blocks */
	byte_offset = 0;
	while (byte_offset < config_mem_size) {
		size = config_mem_size - byte_offset;

		if (size > MXT_MAX_BLOCK_WRITE)
			size = MXT_MAX_BLOCK_WRITE;

		ret = __mxt_write_reg(data->client,
				      cfg_start_ofs + byte_offset,
				      size, config_mem + byte_offset);
		if (ret != 0) {
			dev_err(dev, "Config write error, ret=%d\n", ret);
			goto release_mem;
		}

		byte_offset += size;
	}

	mxt_update_crc(data, MXT_COMMAND_BACKUPNV, MXT_BACKUP_VALUE);

	ret = mxt_check_retrigen(data);
	if (ret)
		goto release_mem;

	/* ret = mxt_soft_reset(data); */
	/* if (ret) */
	/* 	goto release_mem; */

	ATMEL_MSG("Config written\n");

	/* T7 config may have changed */
	mxt_init_t7_power_cfg(data);
	mxt_t6_command(data, MXT_COMMAND_CALIBRATE, 1, false);

	/* update new config id */
	if (data->T38_address) {
		data->config_id = mxt_update_config_id(data);
		mxt_update_vendor_ver_id(data);
	}

release_mem:
	kfree(config_mem);
release:
#if (USE_REQUEST_FIRMWARE == 1)
	release_firmware(cfg);
#else
	kfree(cfg->data);
	kfree(cfg);
#endif
	return ret;
}

static int mxt_set_t7_power_cfg(struct mxt_data *data, u8 sleep)
{
	int error;
	struct t7_config *new_config;
	struct t7_config deepsleep = { .active = 0, .idle = 0 };

	ATMEL_TRACE();

	if (sleep == MXT_POWER_CFG_DEEPSLEEP) {
		new_config = &deepsleep;
	} else if (sleep == MXT_POWER_CFG_RUN_FAST) {
		new_config = &touch_dev_cfg.t7_stylus_cfg;
	} else {
		new_config = &touch_dev_cfg.t7_finger_cfg;
	}

	error = __mxt_write_reg(data->client, data->T7_address,
				sizeof(*new_config),
				new_config);
	if (error)
		return error;

	ATMEL_MSG("Set T7 ACTV:%d IDLE:%d ACTV2IDLE:%d\n",
		  new_config->active, new_config->idle, new_config->actv2idle);

	return 0;
}

static int mxt_init_t7_power_cfg(struct mxt_data *data)
{
	ATMEL_TRACE();

	mxt_set_t7_power_cfg(data, MXT_POWER_CFG_RUN);

	return 0;
}

static bool mxt_t47_stylus_state(struct mxt_data *data)
{
#if 0
	struct mxt_object *object;
	int error;
	u8 control = 0;

	ATMEL_TRACE();

	object = mxt_get_object(data, MXT_PROCI_STYLUS_T47);
	if (!object)
		return false;

	error = __mxt_read_reg(data->client, object->start_address,
				sizeof(control), &control);
	if (error)
		dev_warn(&data->client->dev, "Unable to read T47\n");

	return control ? true : false;
#else
	return true;
#endif
}

static inline int mxt_acquire_irq(struct mxt_data *data)
{
	ATMEL_TRACE();

	return mxt_process_messages_until_invalid(data);
}

static void mxt_irq_enable(struct mxt_data *data, bool enable)
{
	ATMEL_TRACE();

	if (enable && !data->irq_enabled) {
		ATMEL_DBG("enable_irq\n");
		enable_irq(data->irq);
		data->irq_enabled = true;
	} else if (!enable && data->irq_enabled) {
		ATMEL_DBG("disable_irq\n");
		disable_irq_nosync(data->irq);
		data->irq_enabled = false;
	}
}

#if 0
static void mxt_apply_patchset(struct mxt_data *data, struct mxt_patch *patch)
{
	struct mxt_obj_patch *mo_patch;
	struct mxt_object *object;
	u16 reg2w;
	int error;

	ATMEL_TRACE();

	if (!patch || !patch->cfg_num) {
		pr_debug("patchset is empty!\n");
		return;
	}

	list_for_each_entry(mo_patch, &patch->cfg_head, link) {
		object = mxt_get_object(data, mo_patch->number);
		if (!object)
			continue;

		if (mo_patch->offset > object->size_minus_one)
			continue;

		reg2w = object->start_address +
			mxt_obj_size(object) * mo_patch->instance +
			mo_patch->offset;

		dev_dbg(&data->client->dev, "patching T%d-%d@%d,0x%x off=%d\n",
				mo_patch->number, mo_patch->instance,
				mo_patch->offset, mo_patch->value, reg2w);

		error = __mxt_write_reg(data->client, reg2w,
					1, &mo_patch->value);
		if (error)
			dev_warn(&data->client->dev, "patch failed\n");
	}
}
#endif

static void mxt_enable_wakeup_source(struct mxt_data *data, bool enable)
{
	int error;

	ATMEL_TRACE();

	error = irq_set_irq_wake(data->irq, (int)enable);
	pr_debug("%s wakeup; rc=%d\n", enable ? "enabled" : "disabled", error);
}

static void mxt_sensor_state_config(struct mxt_data *data, int state)
{
	ATMEL_TRACE();

	if (data->mode_is_wakeable)
		mxt_enable_wakeup_source(data, true);
#if 0
	mxt_apply_patchset(data, &data->current_mode->patch_data[state]);
	pr_debug("applying  %s in mode %s\n",
		state == ACTIVE_IDX ? "ACTIVE" : "SUSPEND",
		data->current_mode == data->default_mode ? "DEFAULT" : "OTHER");
#endif
}

static void mxt_wait_for_idle(struct mxt_data *data)
{
	int state, timeout_retries = 0;
	unsigned long start_wait_jiffies = jiffies;

	ATMEL_TRACE();

	/* Reset completion indicated by asserting CHG  */
	/* Wait for CHG asserted or timeout after 200ms */
	do {
		state = gpio_get_value(data->pdata->gpio_irq);
		if (!state)
			break;
		usleep_range(1000, 1000);
	} while (++timeout_retries < 100);

	if (!state) {
		ATMEL_DBG("reset took %ums\n",
			jiffies_to_msecs(jiffies - start_wait_jiffies));
	} else {
		ATMEL_ERR("timeout waiting for idle %ums\n",
			jiffies_to_msecs(jiffies - start_wait_jiffies));
	}
}

static inline void mxt_set_alternate_mode(struct mxt_data *data,
	struct mxt_patchset *mode, bool wakeable, bool persistent)
{
	ATMEL_TRACE();

	data->mode_is_wakeable = wakeable;
	data->mode_is_persistent = persistent;
	data->current_mode = mode;
	if (wakeable)
		mxt_enable_wakeup_source(data, true);
}

static inline void mxt_restore_default_mode(struct mxt_data *data)
{
	ATMEL_TRACE();

	if (data->mode_is_wakeable)
		mxt_enable_wakeup_source(data, false);
	data->mode_is_wakeable = false;
	data->mode_is_persistent = true;
	data->current_mode = data->default_mode;
}

static const char * const mxt_state_names[] = { "UNKNOWN", "ACTIVE", "SUSPEND",
	"UNUSED", "STANDBY", "BL", "INIT", "FLASH", "QUERY", "INVALID" };

static const char *mxt_state_name(int state)
{
	int index = state < 0 || state > STATE_INVALID ? STATE_INVALID : state;
	return mxt_state_names[index];
}

static int mxt_get_sensor_state(struct mxt_data *data)
{
	/* ATMEL_TRACE(); */

	dev_dbg(&data->client->dev,
                "state is %s\n", mxt_state_name(atomic_read(&data->state)));

	return atomic_read(&data->state);
}

static void mxt_set_sensor_state(struct mxt_data *data, int state)
{
	int current_state = mxt_get_sensor_state(data);

	/* ATMEL_TRACE(); */

	if (current_state == state)
		return;

	switch (state) {
	case STATE_QUERY:
		/* drop flag to allow object specific message handling */
		if (data->in_bootloader)
			data->in_bootloader = false;
	case STATE_UNKNOWN:
	case STATE_FLASH:
		/* no special handling for these states */
			break;

	case STATE_SUSPEND:
		if (!data->mode_is_wakeable)
			mxt_irq_enable(data, false);
		data->enable_reporting = false;
		if (!data->in_bootloader)
			mxt_sensor_state_config(data, SUSPEND_IDX);
			break;

	case STATE_ACTIVE:
		if (!data->in_bootloader)
			mxt_sensor_state_config(data, ACTIVE_IDX);
		data->enable_reporting = true;

		if (!data->mode_is_persistent) {
			mxt_restore_default_mode(data);
			pr_debug("Non-persistent mode; restoring default\n");
		}
			break;

	case STATE_STANDBY:
		mxt_irq_enable(data, false);
			break;

	case STATE_BL:
		if (!data->in_bootloader)
			data->in_bootloader = true;

		mxt_irq_enable(data, false);
			break;

	case STATE_INIT:
		/* set flag to avoid object specific message handling */
		if (!data->in_bootloader)
			data->in_bootloader = true;
			break;
	}

	ATMEL_DBG("state change %s -> %s\n",
                mxt_state_name(current_state), mxt_state_name(state));
	atomic_set(&data->state, state);

	if (state == STATE_ACTIVE)
		mxt_irq_enable(data, true);
}


static int mxt_set_glove(struct mxt_data *data, int enable)
{
	struct mxt_object *object;
	struct glove_config_t *cfg = NULL;
	int error = 0;

	ATMEL_TRACE();

	if (data->poweron == false) {
		ATMEL_MSG("device is power off, skip glove configuration\n");
		goto end;
	}

	if (enable != 0) {
		switch (data->glove_conf) {
		case TP_GLOVE_SENS_HIGH:
			cfg = &touch_dev_cfg.glove_on_cfg_high;
			break;
		case TP_GLOVE_SENS_LOW:
			cfg = &touch_dev_cfg.glove_on_cfg_low;
			break;
		case TP_GLOVE_SENS_MID:
		default:
			cfg = &touch_dev_cfg.glove_on_cfg_mid;
			break;
		}
	} else {
		cfg = &touch_dev_cfg.glove_off_cfg;
	}

	object = mxt_get_object(data, MXT_PROCI_GLOVEDETECTION_T78);
	if (object != NULL) {
		/* glove mode */
		error = __mxt_write_reg(data->client, object->start_address,
					sizeof(cfg->enable), &cfg->enable);
		if (error) {
			ATMEL_ERR("write T78 ctrl failed\n");
			goto end;
		}

		/* glove sensitivity */
		error = __mxt_write_reg(data->client, object->start_address + 9,
					sizeof(cfg->glove_self_thr),
					&cfg->glove_self_thr);
		if (error) {
			ATMEL_ERR("write T78 thr failed\n");
			goto end;
		}
		error = __mxt_write_reg(data->client, object->start_address + 10,
					sizeof(cfg->glove_hys), &cfg->glove_hys);
		if (error) {
			ATMEL_ERR("write T78 hys failed\n");
			goto end;
		}
		error = __mxt_write_reg(data->client, object->start_address + 8,
					sizeof(cfg->discr_thr), &cfg->discr_thr);
		if (error) {
			ATMEL_ERR("write T78 discr failed\n");
			goto end;
		}
	}

	object = mxt_get_object(data, MXT_GEN_ACQUIRE_T8);
	if (object != NULL) {
		u16 reg_addr = object->start_address + 12;
		u8 mut_self = 0x00;

		error = __mxt_read_reg(data->client, reg_addr,
				       sizeof(mut_self), &mut_self);
		if (error != 0) {
			ATMEL_ERR("T%d MUT_SELF read failed\n", object->type);
			goto end;
		}

		mut_self &= (~0x3);
		mut_self |= (cfg->active_mutual_self & 0x03);

		error = __mxt_write_reg(data->client, reg_addr,
					sizeof(mut_self), &mut_self);
		if (error) {
			ATMEL_ERR("write T%d MUT_SELF failed\n", object->type);
			goto end;
		}
	}

	object = mxt_get_object(data, MXT_TOUCH_MULTITOUCHSCREEN_T100);
	if (object != NULL) {
		u16 addr;
		u8 hys = 0x00;

		addr = object->start_address + MXT_T100_MOVHYSTN;
		error = __mxt_write_reg(data->client, addr,
					sizeof(cfg->next_move_filter),
					&cfg->next_move_filter);
		if (error != 0) {
			ATMEL_ERR("write T%d MOV_FILTER failed\n", object->type);
			goto end;
		}

		addr = object->start_address + MXT_T100_TCHTHR;
		error = __mxt_write_reg(data->client, addr,
					sizeof(cfg->z_touch_thr), &cfg->z_touch_thr);
		if (error) {
			ATMEL_ERR("write T100 z touch thr failed\n");
			goto end;
		}

		addr = object->start_address + MXT_T100_TCHTHRHYS;
		error = __mxt_read_reg(data->client, addr, sizeof(hys), &hys);
		if (error != 0) {
			ATMEL_ERR("T%d TCHTHRHYS read failed\n", object->type);
			goto end;
		}

		hys &= (~0x3F);
		hys |= (cfg->z_touch_thr_hys & 0x3F);

		error = __mxt_write_reg(data->client, addr, sizeof(hys), &hys);
		if (error) {
			ATMEL_ERR("write T100 z touch thr hys failed\n");
			goto end;
		}

		addr = object->start_address + MXT_T100_MOVHYSTI;
		error = __mxt_write_reg(data->client, addr,
					sizeof(cfg->init_move_filter), &cfg->init_move_filter);
		if (error) {
			ATMEL_ERR("write T100 init move filter failed\n");
			goto end;
		}
	}

	object = mxt_get_object(data, MXT_PROCI_STYLUS_T47);
	if (object != NULL) {
		error = __mxt_write_reg(data->client, object->start_address + 5,
					sizeof(cfg->max_touch_amp), &cfg->max_touch_amp);
		if (error) {
			ATMEL_ERR("write T47 max touch amp failed\n");
			goto end;
		}
	}

	object = mxt_get_object(data, MXT_PROCG_NOISESUPSELFCAT_T108);
	if (object != NULL) {
		u16 addr;

		addr = object->start_address + 1;
		error = __mxt_write_reg(data->client, addr,
					sizeof(cfg->noise_on_cal), &cfg->noise_on_cal);
		if (error != 0) {
			ATMEL_ERR("write T%d CALCFG1 failed\n", object->type);
			goto end;
		}

		addr = object->start_address + 2;
		error = __mxt_write_reg(data->client, addr,
					sizeof(cfg->noise_on), &cfg->noise_on);
		if (error != 0) {
			ATMEL_ERR("write T%d CFG1 failed\n", object->type);
			goto end;
		}
	}

	object = mxt_get_object(data, MXT_SPT_DYNAMICCONFIGURATIONCONTROLLER_T70);
	if (object != NULL) {
		u16 addr;

		addr = object->start_address + mxt_obj_size(object) * 1;
		error = __mxt_write_reg(data->client, addr,
					sizeof(cfg->ins1_dynamic_en), &cfg->ins1_dynamic_en);
		if (error) {
			ATMEL_ERR("write T%d(1) CTRL failed\n", object->type);
			return 0;
		}

		addr = object->start_address + mxt_obj_size(object) * 2;
		error = __mxt_write_reg(data->client, addr,
					sizeof(cfg->ins2_dynamic_en), &cfg->ins2_dynamic_en);
		if (error) {
			ATMEL_ERR("write T%d(2) CTRL failed\n", object->type);
			return 0;
		}

		addr = object->start_address + mxt_obj_size(object) * 5;
		error = __mxt_write_reg(data->client, addr,
					sizeof(cfg->ins5_dynamic_en), &cfg->ins5_dynamic_en);
		if (error) {
			ATMEL_ERR("write T%d(5) CTRL failed\n", object->type);
			return 0;
		}

		addr = object->start_address + mxt_obj_size(object) * 7;
		error = __mxt_write_reg(data->client, addr,
					sizeof(cfg->ins7_dynamic_en), &cfg->ins7_dynamic_en);
		if (error) {
			ATMEL_ERR("write T%d(7) CTRL failed\n", object->type);
			return 0;
		}

		addr = object->start_address + mxt_obj_size(object) * 8;
		error = __mxt_write_reg(data->client, addr,
					sizeof(cfg->ins8_dynamic_en), &cfg->ins8_dynamic_en);
		if (error) {
			ATMEL_ERR("write T%d(8) CTRL failed\n", object->type);
			return 0;
		}
	}

	object = mxt_get_object(data, MXT_SPT_CTECONFIG_T46);
	if (object != NULL) {
		u16 addr;
		u8 lcal = 0x00;

		addr = object->start_address;
		error = __mxt_read_reg(data->client, addr, sizeof(lcal), &lcal);
		if (error != 0) {
			ATMEL_ERR("T%d long cal en read failed\n", object->type);
			goto end;
		}

		lcal &= (~0x04);
		lcal |= (cfg->long_cal_en & 0x04);

		error = __mxt_write_reg(data->client, addr, sizeof(lcal), &lcal);
		if (error) {
			ATMEL_ERR("write T%d long cal en failed\n", object->type);
			goto end;
		}
	}

	object = mxt_get_object(data, MXT_SPT_SELFCAPCONFIG_T111);
	if (object != NULL) {
		u16 addr;

		addr = object->start_address + 20;
		error = __mxt_write_reg(data->client, addr,
					sizeof(cfg->dc_filter), &cfg->dc_filter);
		if (error) {
			ATMEL_ERR("write T%d dcfilter failed\n", object->type);
			goto end;
		}

		addr = object->start_address + 21;
		error = __mxt_write_reg(data->client, addr,
					sizeof(cfg->dc_cal_rec_str), &cfg->dc_cal_rec_str);
		if (error) {
			ATMEL_ERR("write T%d dccalrecstr failed\n", object->type);
			goto end;
		}
	}

end:
	data->glove_enabled = enable;

	return error;
}

static int mxt_set_hover(struct mxt_data *data, int enable)
{
	struct mxt_object *object;
	u8 control;
	u8 mask = 0x01;
	int error = 0;

	ATMEL_TRACE();

	if (data->poweron == false) {
		ATMEL_MSG("device is power off, skip hover configuration\n");
		goto end;
	}

	object = mxt_get_object(data, MXT_SPT_TOUCHSCREENHOVER_T101);
	if (object != NULL) {
		error = __mxt_read_reg(data->client, object->start_address,
				       sizeof(control), &control);
		if (error) {
			ATMEL_ERR("read T101 ctrl failed\n");
			goto end;
		}

		control &= (~mask);
		if (enable != 0) {
			control |= enable;
		}

		error = __mxt_write_reg(data->client, object->start_address,
					sizeof(control), &control);
		if (error) {
			ATMEL_ERR("write T101 ctrl failed\n");
			goto end;
		}
	}

end:
	data->hover_enabled = enable;

	return error;
}

static int mxt_set_passive_stylus(struct mxt_data *data, int enable)
{
	struct mxt_object *object;
	struct passive_stylus_config_t *cfg = NULL;
	int error = 0;

	ATMEL_TRACE();

	if (data->poweron == false) {
		ATMEL_MSG("device is power off, skip passive stylus configuration\n");
		goto end;
	}

	if (enable != 0) {
		cfg = &touch_dev_cfg.passive_stylus_on_cfg;
	} else {
		cfg = &touch_dev_cfg.passive_stylus_off_cfg;
	}

	object = mxt_get_object(data, MXT_PROCI_STYLUS_T47);
	if (object != NULL) {
		error = __mxt_write_reg(data->client, object->start_address,
					sizeof(cfg->enable), &cfg->enable);
		if (error) {
			ATMEL_ERR("write T47 ctrl failed\n");
			goto end;
		}
	}

	object = mxt_get_object(data, MXT_TOUCH_MULTITOUCHSCREEN_T100);
	if (object != NULL) {
		u16 addr;
		u8 debounce = 0x00;
		u8 mask = 0x3F;

		addr = object->start_address + 39;
		error = __mxt_read_reg(data->client, addr,
				       sizeof(debounce), &debounce);
		if (error) {
			ATMEL_ERR("read T100 touch debounce failed\n");
			goto end;
		}

		debounce &= (~mask);
		debounce |= (cfg->touch_debounce & mask);

		error = __mxt_write_reg(data->client, addr,
					sizeof(debounce), &debounce);
		if (error) {
			ATMEL_ERR("write T100 touch debounce failed\n");
			goto end;
		}

		addr = object->start_address + 40;
		error = __mxt_read_reg(data->client, addr,
				       sizeof(debounce), &debounce);
		if (error) {
			ATMEL_ERR("read T100 release debounce failed\n");
			goto end;
		}

		debounce &= (~mask);
		debounce |= (cfg->release_debounce & mask);

		error = __mxt_write_reg(data->client, addr,
					sizeof(debounce), &debounce);
		if (error) {
			ATMEL_ERR("write T100 release debounce failed\n");
			goto end;
		}

		addr = object->start_address + 45;
		error = __mxt_write_reg(data->client, addr,
					sizeof(cfg->smooth_level), &(cfg->smooth_level));
		if (error) {
			ATMEL_ERR("write T100 smooth level failed\n");
			goto end;
		}
	}

end:
	data->passive_stylus_enabled = enable;

	return error;
}

static int mxt_set_active_stylus(struct mxt_data *data, int enable)
{
	struct mxt_object *object;
	struct active_stylus_config_t *cfg = NULL;
	int error = 0;

	ATMEL_TRACE();

	if (data->poweron == false) {
		ATMEL_MSG("device is power off, skip active stylus configuration\n");
		goto end;
	}

	if (enable != 0) {
		switch (data->active_stylus_pressure) {
		case TP_PEN_ACTIVE_PRESSURE_LIGHT:
			cfg = &touch_dev_cfg.active_stylus_on_cfg_light;
			break;
		case TP_PEN_ACTIVE_PRESSURE_NORMAL:
		default:
			cfg = &touch_dev_cfg.active_stylus_on_cfg_normal;
			break;
		}
	} else {
		cfg = &touch_dev_cfg.active_stylus_off_cfg;
	}

	object = mxt_get_object(data, MXT_GEN_ACQUIRE_T8);
	if (object != NULL) {
		error = __mxt_write_reg(data->client,
					object->start_address + 10,
					sizeof(cfg->enable), &cfg->enable);
		if (error) {
			ATMEL_ERR("write T8 measure allow failed\n");
			goto end;
		}
	}

	object = mxt_get_object(data, MXT_PROCI_ACTIVESTYLUS_T107);
	if (object != NULL) {
		error = __mxt_write_reg(data->client,
					object->start_address + MXT_T107_CFG1,
					sizeof(cfg->detect_mode), &cfg->detect_mode);
		if (error) {
			ATMEL_ERR("write T107 cfg1 failed\n");
			goto end;
		}
	}

end:
	data->active_stylus_enabled = enable;

	return error;
}

static int mxt_set_waterproof(struct mxt_data *data, int enable)
{
	struct mxt_object *object;
	struct waterproof_config_t *cfg = NULL;
	int error = 0;

	ATMEL_TRACE();

	if (data->poweron == false) {
		ATMEL_MSG("device is power off, skip waterproof configuration\n");
		goto end;
	}

	if (enable != 0) {
		switch (data->waterproof_conf) {
		case TP_WATER_RESISTANCE_NORMAL:
			cfg = &touch_dev_cfg.waterproof_on_cfg_light;
			break;
		case TP_WATER_RESISTANCE_HIGH:
		default:
			cfg = &touch_dev_cfg.waterproof_on_cfg_heavy;
			break;
		}
	} else {
		cfg = &touch_dev_cfg.waterproof_off_cfg;
	}

	object = mxt_get_object(data, MXT_TOUCH_MULTITOUCHSCREEN_T100);
	if (object != NULL) {
		u16 addr = object->start_address + MXT_T100_CFG1;
		u8 cfg1 = 0x00;
		u8 mask = 0x01;

		error = __mxt_read_reg(data->client, addr, sizeof(cfg1), &cfg1);
		if (error) {
			ATMEL_ERR("read T100 cfg1 failed\n");
			goto end;
		}

		cfg1 &= (~mask);
		cfg1 |= (cfg->report_mode & mask);

		error = __mxt_write_reg(data->client, addr,
					sizeof(cfg1), &cfg1);
		if (error) {
			ATMEL_ERR("write T100 cfg1 failed\n");
			goto end;
		}

		addr = object->start_address + MXT_T100_GAIN;
		error = __mxt_write_reg(data->client, addr,
					sizeof(cfg->touch_gain), &cfg->touch_gain);
		if (error) {
			ATMEL_ERR("write T100 touch gain failed\n");
			goto end;
		}

		addr = object->start_address + MXT_T100_INTTHR;
		error = __mxt_write_reg(data->client, addr,
					sizeof(cfg->delta_touch_thr), &cfg->delta_touch_thr);
		if (error) {
			ATMEL_ERR("write T100 dalta touch thr failed\n");
			goto end;
		}
	}

	object = mxt_get_object(data, MXT_PROCI_RETRANSMISSIONCOMPENSATION_T80);
	if (object != NULL) {
		error = __mxt_write_reg(data->client, object->start_address,
					sizeof(cfg->anti_moisture), &cfg->anti_moisture);
		if (error) {
			ATMEL_ERR("write T80 anti-moisture failed\n");
			goto end;
		}
	}

end:
	data->waterproof_enabled = enable;

	return error;
}

static void mxt_update_t7_config(struct mxt_data *data)
{
	int cfg = MXT_POWER_CFG_RUN;

	ATMEL_TRACE();

	if (data->poweron == false) {
		ATMEL_MSG("device is power off, skip t7 config\n");
		return;
	}

	if ((data->active_stylus_enabled != 0)
	    || (data->passive_stylus_enabled != 0)) {
		cfg = MXT_POWER_CFG_RUN_FAST;
	}

	mxt_set_t7_power_cfg(data, cfg);

	/* mark 'active' flag into status area */
	if (data->T38_address) {
		struct i2c_client *client = data->client;
		u8 status = MXT_T38_STATUS_DATA;

		(void)__mxt_write_reg(client,
				      data->T38_address + MXT_T38_STATUS,
				      sizeof(status), &status);
	}
}

static void mxt_init_feature(struct mxt_data *data)
{
	ATMEL_TRACE();

	mxt_set_detect_mode(data);

	/* mark 'active' flag into status area*/
	if (data->T38_address) {
		struct i2c_client *client = data->client;
		u8 status = MXT_T38_STATUS_DATA;
		(void)__mxt_write_reg(client,
				      data->T38_address + MXT_T38_STATUS,
				      sizeof(status), &status);
	}

	mForceCalibCnt = touch_dev_cfg.calib_cfg.force_calib_cnt;

	return;
}

static void mxt_healthcheck(struct mxt_data *data)
{
	struct i2c_client *client = data->client;
	u8 status = 0;
	int error = 0;
	int dev_status_error = 0;

	ATMEL_TRACE();

	if (atmel_healthcheck_force_disable != 0) {
		ATMEL_MSG("healthcheck is disabled forcibly\n");
		return;
	}

	if (data->running_healthcheck == 0) {
		ATMEL_MSG("healthcheck has already stopped\n");
		return;
	}

	if (data->enable_fw_update != 0) {
		ATMEL_MSG("skip healthcheck under updating fw\n");
		return;
	}

	/*
	 * no need to disable/enable interruption because interrupt handler
	 * and healthcheck are processed sequentially in a single workqueue
	 */

	/* check 'active' flag in status area */
	error = __mxt_read_reg(client,
			       data->T38_address + MXT_T38_STATUS,
			       sizeof(status), &status);

	/* Save current status */
	if ((data->t6_status & MXT_T6_STATUS_DEVERR) != 0) {
		dev_status_error = data->t6_status;
	}

	if ((error != 0)
	    || (status != MXT_T38_STATUS_DATA)
	    || (dev_status_error != 0)) {

		ATMEL_ERR("device error detected, "
			  "i2c=%s, status=0x%02X, dev_status=0x%02X\n",
			  (error != 0) ? "fail" : "success",
			  status,
			  dev_status_error);

		error = mxt_hw_reset(data);
		if (error == 0) {
			/* hard reset succeeded --> initial setting */
			mxt_init_feature(data);
		} else {
			/* hard reset failed --> power off */
			ATMEL_ERR("hardware reset failed\n");
			data->running_healthcheck = 0;
			mxt_suspend(&client->dev, MXT_SUSPEND_ASYNC);
		}
	}

	/* calibration when detecting touch duration is too long */
	if (mForceCalibCnt > 0) { /* enable force calib */
		if (mCurrentEvent.posState != 0) {
			/* detect touch */
			mForceCalibCnt--;
			if (mForceCalibCnt <= 0) {
				/* run calib forcibly */
				ATMEL_DBG("force calibration\n");
				mxt_t6_command(data, MXT_COMMAND_CALIBRATE, 1, false);
				mForceCalibCnt = touch_dev_cfg.calib_cfg.force_calib_cnt;
			}
		} else {
			mForceCalibCnt = touch_dev_cfg.calib_cfg.force_calib_cnt;
		}
        }


	if (data->running_healthcheck == 1) {
		mxt_queue_healthcheck(data);
	}

	return;
}

static void mxt_free_input_device(struct mxt_data *data)
{
	ATMEL_TRACE();

	if (data->input_dev) {
		if (data->input_registered)
			input_unregister_device(data->input_dev);
		else
			input_free_device(data->input_dev);
		data->input_dev = NULL;
		data->input_registered = false;

		if (data->touch_state != NULL) {
			kfree(data->touch_state);
			data->touch_state = NULL;
		}
	}
}

static void mxt_free_object_table(struct mxt_data *data)
{
	ATMEL_TRACE();

	mxt_debug_msg_remove(data);

	kfree(data->raw_info_block);
	data->object_table = NULL;
	data->info = NULL;
	data->raw_info_block = NULL;
	kfree(data->msg_buf);
	data->msg_buf = NULL;
	kfree(data->T100_data);
	data->T100_data = NULL;

	mxt_free_input_device(data);

	data->enable_reporting = false;
	data->T5_address = 0;
	data->T5_msg_size = 0;
	data->T6_reportid = 0;
	data->T7_address = 0;
	data->T9_reportid_min = 0;
	data->T9_reportid_max = 0;
	data->T15_reportid_min = 0;
	data->T15_reportid_max = 0;
	data->T18_address = 0;
	data->T19_reportid = 0;
	data->T42_reportid_min = 0;
	data->T42_reportid_max = 0;
	data->T44_address = 0;
	data->T48_reportid = 0;
	data->T63_reportid_min = 0;
	data->T63_reportid_max = 0;
	data->T93_reportid = 0;
	data->T100_reportid_min = 0;
	data->T100_reportid_max = 0;
	data->max_reportid = 0;
	data->T38_address = 0;
	data->T25_reportid = 0;
}

static int mxt_parse_object_table(struct mxt_data *data)
{
	struct i2c_client *client = data->client;
	int i;
	u8 reportid;
	u16 end_address;

	ATMEL_TRACE();

	/* Valid Report IDs start counting from 1 */
	reportid = 1;
	data->mem_size = 0;
	for (i = 0; i < data->info->object_num; i++) {
		struct mxt_object *object = data->object_table + i;
		u8 min_id, max_id;

		le16_to_cpus(&object->start_address);

		if (object->num_report_ids) {
			min_id = reportid;
			reportid += object->num_report_ids *
					mxt_obj_instances(object);
			max_id = reportid - 1;
		} else {
			min_id = 0;
			max_id = 0;
		}

		ATMEL_MSG("T%-3u Start:%-4u Size:%-3zu Instances:%zu x %u Report IDs:%u-%u\n",
			object->type, object->start_address,
			mxt_obj_size(object), mxt_obj_instances(object),
                        object->num_report_ids,
			min_id, max_id);

		switch (object->type) {
		case MXT_GEN_MESSAGE_T5:
			if (data->info->family_id == 0x80) {
				/* On mXT224 read and discard unused CRC byte
				 * otherwise DMA reads are misaligned */
				data->T5_msg_size = mxt_obj_size(object);
			} else {
				/* CRC not enabled, so skip last byte */
				data->T5_msg_size = mxt_obj_size(object) - 1;
			}
			data->T5_address = object->start_address;
			break;
		case MXT_GEN_COMMAND_T6:
			data->T6_reportid = min_id;
			data->T6_address = object->start_address;
			break;
		case MXT_GEN_POWER_T7:
			data->T7_address = object->start_address;
			break;
		case MXT_SPT_USERDATA_T38:
			data->T38_address = object->start_address;
                        /* data->T38_address = 0; */
                        break;
		case MXT_TOUCH_MULTI_T9:
			/* Only handle messages from first T9 instance */
			data->T9_reportid_min = min_id;
			data->T9_reportid_max = min_id +
						object->num_report_ids - 1;
			data->num_touchids = object->num_report_ids;
			break;
		case MXT_TOUCH_KEYARRAY_T15:
			data->T15_reportid_min = min_id;
			data->T15_reportid_max = max_id;
			break;
		case MXT_SPT_COMMSCONFIG_T18:
			data->T18_address = object->start_address;
			break;
		case MXT_PROCI_TOUCHSUPPRESSION_T42:
			data->T42_reportid_min = min_id;
			data->T42_reportid_max = max_id;
			break;
		case MXT_SPT_MESSAGECOUNT_T44:
			data->T44_address = object->start_address;
			break;
		case MXT_SPT_GPIOPWM_T19:
			data->T19_reportid = min_id;
			break;
		case MXT_PROCG_NOISESUPPRESSION_T48:
			data->T48_reportid = min_id;
			break;
		case MXT_PROCI_ACTIVE_STYLUS_T63:
			/* Only handle messages from first T63 instance */
			data->T63_reportid_min = min_id;
			data->T63_reportid_max = min_id;
			data->num_stylusids = 1;
			break;
		case MXT_PROCI_TOUCHSEQUENCELOGGER_T93:
			data->T93_reportid = min_id;
			break;
		case MXT_TOUCH_MULTITOUCHSCREEN_T100:
			data->T100_reportid_min = min_id;
			data->T100_reportid_max = max_id;
			/* first two report IDs reserved */
			data->num_touchids = object->num_report_ids - 2;
			break;
		case  MXT_SPT_SELFTEST_T25:
			data->T25_reportid = min_id;
			break;
		}

		end_address = object->start_address
			+ mxt_obj_size(object) * mxt_obj_instances(object) - 1;

		if (end_address >= data->mem_size)
			data->mem_size = end_address + 1;
	}

	/* Store maximum reportid */
	data->max_reportid = reportid;

	/* If T44 exists, T5 position has to be directly after */
	if (data->T44_address && (data->T5_address != data->T44_address + 1)) {
		dev_err(&client->dev, "Invalid T44 position\n");
		return -EINVAL;
	}

	data->msg_buf = kcalloc(data->max_reportid,
				data->T5_msg_size, GFP_KERNEL);
	if (!data->msg_buf) {
		dev_err(&client->dev, "Failed to allocate message buffer\n");
		return -ENOMEM;
	}

	return 0;
}

static u32 mxt_update_config_id(struct mxt_data *data)
{
	struct i2c_client *client = data->client;
	u8 userdata[4] = {0};
	u32 config_id = 0;
	int error;

	ATMEL_TRACE();

	error = __mxt_read_reg(client, data->T38_address,
				sizeof(userdata), userdata);
	if (error)
		dev_warn(&client->dev, "Unable to read config id\n");
	else
		config_id = be32_to_cpu(*(unsigned int *)&userdata[0]);

	return config_id;
}

static void mxt_update_vendor_ver_id(struct mxt_data *data)
{
	struct i2c_client *client = data->client;
	u8 userdata[6] = {0};
	int error;

	ATMEL_TRACE();

	error = __mxt_read_reg(client,
			       data->T38_address + MXT_T38_VENDOR_VER,
			       sizeof(userdata), userdata);
	if (error) {
		ATMEL_ERR("Unable to read config id\n");
	} else {
		data->vendor_config_id = be32_to_cpu(*(unsigned int *)&userdata[0]);
		data->vendor_local_rev = be16_to_cpu(*(unsigned short *)&userdata[4]);
		ATMEL_DAT("vendor ver = %08X.%04X\n",
			  (unsigned int)data->vendor_config_id,
			  (unsigned int)data->vendor_local_rev);
	}

	return;
}

static int mxt_read_fw_version(struct mxt_data *data, u16 *ver)
{
	int ret = 0;
	struct i2c_client *client = data->client;
	struct mxt_info info_buf;
	u8 dev[2];
	int error = 0;

	ATMEL_TRACE();

	memset(&info_buf, 0, sizeof(info_buf));
	error = __mxt_read_reg(client, 0, sizeof(info_buf), &info_buf);
	if (error != 0) {
		ATMEL_ERR("I2C read failed\n");
		ret = error;
		goto ver_check_end;
	}
	dev[0] = info_buf.version;
	dev[1] = info_buf.build;
	*ver = be16_to_cpu(*(unsigned short *)&dev[0]);

	ATMEL_MSG("device fw version = %04X\n", (unsigned int)*ver);

ver_check_end:
	return ret;
}

static int mxt_read_cfg_version(struct mxt_data *data, u32 *ver, u16 *rev)
{
	int ret = 0;
	struct i2c_client *client = data->client;
	u8 userdata_dev[6];
	int error = 0;

	ATMEL_TRACE();

	memset(userdata_dev, 0, sizeof(userdata_dev));

	error = __mxt_read_reg(client, data->T38_address + MXT_T38_VENDOR_VER,
			       sizeof(userdata_dev), userdata_dev);
	if (error != 0) {
		ATMEL_ERR("I2C read failed\n");
		goto ver_check_end;
	}
	*ver = be32_to_cpu(*(unsigned int *)&userdata_dev[0]);
	*rev = be16_to_cpu(*(unsigned short *)&userdata_dev[4]);

	ATMEL_MSG("device cfg version = %08X.%04X\n",
		  (unsigned int)*ver, (unsigned int)*rev);

ver_check_end:
	return ret;
}

static int mxt_read_info_block(struct mxt_data *data)
{
	struct i2c_client *client = data->client;
	int error;
	size_t size;
	void *buf;
	uint8_t num_objects;
	u32 calculated_crc;
	u8 *crc_ptr;
	u8 revision = 0;

	ATMEL_TRACE();

	/* If info block already allocated, free it */
	if (data->raw_info_block != NULL)
		mxt_free_object_table(data);

	/* Read 7-byte ID information block starting at address 0 */
	size = sizeof(struct mxt_info);
	buf = kzalloc(size, GFP_KERNEL);
	if (!buf) {
		dev_err(&client->dev, "Failed to allocate memory\n");
		return -ENOMEM;
	}

	error = __mxt_read_reg(client, 0, size, buf);
	if (error) {
		dev_err(&client->dev, "error reading info block\n");
		goto err_free_mem;
	}

	/* Resize buffer to give space for rest of info block */
	num_objects = ((struct mxt_info *)buf)->object_num;
	size += (num_objects * sizeof(struct mxt_object))
		+ MXT_INFO_CHECKSUM_SIZE;

	buf = krealloc(buf, size, GFP_KERNEL);
	if (!buf) {
		dev_err(&client->dev, "Failed to allocate memory\n");
		error = -ENOMEM;
		goto err_free_mem;
	}

	/* Read rest of info block */
	error = __mxt_read_reg(client, MXT_OBJECT_START,
			       size - MXT_OBJECT_START,
			       buf + MXT_OBJECT_START);
	if (error) {
		dev_err(&client->dev, "error reading object table\n");
		goto err_free_mem;
	}

	/* Extract & calculate checksum */
	crc_ptr = buf + size - MXT_INFO_CHECKSUM_SIZE;
	data->info_crc = crc_ptr[0] | (crc_ptr[1] << 8) | (crc_ptr[2] << 16);

	calculated_crc = mxt_calculate_crc(buf, 0,
					   size - MXT_INFO_CHECKSUM_SIZE);

	/* dump information block */
	if (atmel_debug_flag & ATMEL_DEBUG_DBG) {
		int i;
		struct mxt_info *info_block = (struct mxt_info *)buf;
		struct mxt_object *object_table = (struct mxt_object *)(buf + MXT_OBJECT_START);
		ATMEL_DBG("----------Information block-----------\n");
		ATMEL_DBG("%02X %02X %02X %02X %02X %02X %02X\n", 
			  info_block->family_id,
			  info_block->variant_id,
			  info_block->version,
			  info_block->build,
			  info_block->matrix_xsize,
			  info_block->matrix_ysize,
			  info_block->object_num);
		for (i = 0; i < num_objects; i++) {
			struct mxt_object *object = object_table + i;
			ATMEL_DBG("%02X %02X %02X %02X %02X %02X\n",
				  object->type,
				  object->start_address & 0x00FF,
				  (object->start_address >> 8) & 0x00FF,
				  object->size_minus_one,
				  object->instances_minus_one,
				  object->num_report_ids);

		}
		ATMEL_DBG("%02X %02X %02X\n", crc_ptr[0], crc_ptr[1], crc_ptr[2]);
		ATMEL_DBG("--------------------------------------\n");
	}

	/* CRC mismatch can be caused by data corruption due to I2C comms
	 * issue or else device is not using Object Based Protocol */
	if ((data->info_crc == 0) || (data->info_crc != calculated_crc)) {
		dev_err(&client->dev,
			"Info Block CRC error calculated=0x%06X read=0x%06X\n",
			data->info_crc, calculated_crc);
		return -EIO;
	}

	/* Save pointers in device data structure */
	data->raw_info_block = buf;
	data->info = (struct mxt_info *)buf;
	data->object_table = (struct mxt_object *)(buf + MXT_OBJECT_START);

	dev_info(&client->dev,
		 "Family: 0x%X Variant: 0x%X Firmware V%u.%u.%02X Objects: %u\n",
		 data->info->family_id, data->info->variant_id,
		 data->info->version >> 4, data->info->version & 0xf,
		 data->info->build, data->info->object_num);

	/* Parse object table information */
	error = mxt_parse_object_table(data);
	if (error) {
		dev_err(&client->dev, "Error %d reading object table\n", error);
		mxt_free_object_table(data);
		return error;
	}

	if (data->T38_address) {
		data->config_id = mxt_update_config_id(data);
		mxt_update_vendor_ver_id(data);
	}

	error = mxt_read_revision_id(data, &revision);
	if (!error) {
		data->revision_id = revision;
		dev_info(&client->dev,
			"Revision ID: 0x%02x\n", data->revision_id);
	}

	return 0;

err_free_mem:
	kfree(buf);
	return error;
}

static int mxt_read_t9_resolution(struct mxt_data *data)
{
	struct i2c_client *client = data->client;
	int error;
	struct t9_range range;
	unsigned char orient;
	struct mxt_object *object;

	ATMEL_TRACE();

	object = mxt_get_object(data, MXT_TOUCH_MULTI_T9);
	if (!object)
		return -EINVAL;

	error = __mxt_read_reg(client,
			       object->start_address + MXT_T9_RANGE,
			       sizeof(range), &range);
	if (error)
		return error;

	le16_to_cpus(range.x);
	le16_to_cpus(range.y);

	error =  __mxt_read_reg(client,
				object->start_address + MXT_T9_ORIENT,
				1, &orient);
	if (error)
		return error;

	/* Handle default values */
	if (range.x == 0)
		range.x = 1023;

	if (range.y == 0)
		range.y = 1023;

	if (orient & MXT_T9_ORIENT_SWITCH) {
		data->max_x = range.y;
		data->max_y = range.x;
	} else {
		data->max_x = range.x;
		data->max_y = range.y;
	}

	dev_info(&client->dev,
		 "Touchscreen size X%uY%u\n", data->max_x, data->max_y);

	return 0;
}

static void mxt_regulator_enable(struct mxt_data *data)
{
	int error;

	ATMEL_TRACE();

	/* RESET = OUTPUT/LOW */
	if (gpio_is_valid(data->pdata->gpio_reset)) {
		gpio_direction_output(data->pdata->gpio_reset, 0);
		msleep(MXT_RESET_CTRL_TIME);
	}

	/* VDD/VDDIO --> XVDD */
	/* turn on VDD/VDDIO */
	if (gpio_is_valid(data->pdata->gpio_ldo_en)) {
		gpio_set_value_cansleep(data->pdata->gpio_ldo_en, 1);
		if (gpio_is_valid(data->pdata->gpio_io_ldo_en)) {
			gpio_set_value_cansleep(data->pdata->gpio_io_ldo_en, 1);
		}
	} else {
		if (gpio_is_valid(data->pdata->gpio_io_ldo_en)) {
			gpio_set_value_cansleep(data->pdata->gpio_io_ldo_en, 1);
		}

		if (data->reg_vdd) {
			error = regulator_enable(data->reg_vdd);
			if (error) {
				ATMEL_ERR("Error %d enabling vdd regulator\n", error);
				return;
			}
		}
	}

	/* turn on XVDD */
	gpio_set_value_cansleep(data->pdata->gpio_dcdc_en, 1);
	ATMEL_MSG("Regulator On\n");

	/* RESET = INPUT/NO-PULL */
	if (gpio_is_valid(data->pdata->gpio_reset)) {
		msleep(MXT_RESET_CTRL_TIME);
		gpio_direction_input(data->pdata->gpio_reset);
	}

	msleep(MXT_WAKEUP_TIME);

	mxt_wait_for_idle(data);
}

static void mxt_regulator_disable(struct mxt_data *data)
{
	ATMEL_TRACE();

	gpio_set_value_cansleep(data->pdata->gpio_dcdc_en, 0);
	if (gpio_is_valid(data->pdata->gpio_io_ldo_en)) {
		gpio_set_value_cansleep(data->pdata->gpio_io_ldo_en, 0);
	}
	if (gpio_is_valid(data->pdata->gpio_ldo_en)) {
		gpio_set_value_cansleep(data->pdata->gpio_ldo_en, 0);
	} else {
		if (data->reg_vdd) {
			regulator_disable(data->reg_vdd);
		}
	}

	ATMEL_MSG("Regulator Off\n");
}

static int mxt_gpio_configure(struct mxt_data *data)
{
	struct device *dev = &data->client->dev;
	int error = -EINVAL;

	ATMEL_TRACE();

	if (data->pdata->ts_pinctrl != NULL) {
		error = pinctrl_select_state(data->pdata->ts_pinctrl,
					     data->pdata->pinctrl_state_active);
		if (error < 0) {
			struct device *dev = &data->client->dev;
			dev_err(dev, "%s: Failed to select %s pinstate %d\n",
				__func__, PINCTRL_STATE_ACTIVE, error);
			goto fail;
		}
	}

	if (gpio_is_valid(data->pdata->gpio_dcdc_en)) {
		/* configure touchscreen XVDD control gpio */
		error = gpio_request(data->pdata->gpio_dcdc_en, "mxt_gpio_dcdc_en");
		if (error) {
			dev_err(dev, "unable to request gpio [%u]\n",
				data->pdata->gpio_dcdc_en);
			goto fail;
		}

		error = gpio_direction_output(data->pdata->gpio_dcdc_en, 0);
		if (error) {
			dev_err(dev, "unable to set direction for gpio [%u]\n",
				data->pdata->gpio_dcdc_en);
			goto fail_release_gpio_dcdc_en;
		}

                ATMEL_MSG("gpio_dcdc_en [%u]\n", data->pdata->gpio_dcdc_en);

	} else {
		dev_err(dev, "xvdd control gpio not provided\n");
		goto fail;
	}

	if (gpio_is_valid(data->pdata->gpio_io_ldo_en)) {
		/* configure touchscreen VDDIO control gpio */
		error = gpio_request(data->pdata->gpio_io_ldo_en, "mxt_gpio_io_ldo_en");
		if (error) {
			dev_err(dev, "unable to request gpio [%u]\n",
				data->pdata->gpio_io_ldo_en);
			goto fail_release_gpio_dcdc_en;
		}

		error = gpio_direction_output(data->pdata->gpio_io_ldo_en, 0);
		if (error) {
			dev_err(dev, "unable to set direction for gpio [%u]\n",
				data->pdata->gpio_io_ldo_en);
			goto fail_release_gpio_io_ldo_en;
		}

		ATMEL_MSG("gpio_io_ldo_en [%u]\n", data->pdata->gpio_io_ldo_en);

	} else {
		dev_info(dev, "vddio control gpio not provided\n");
	}

	if (gpio_is_valid(data->pdata->gpio_ldo_en)) {
		/* configure touchscreen VDD control gpio */
		error = gpio_request(data->pdata->gpio_ldo_en, "mxt_gpio_ldo_en");
		if (error) {
			dev_err(dev, "unable to request gpio [%u]\n",
				data->pdata->gpio_ldo_en);
			goto fail_release_gpio_io_ldo_en;
		}

		error = gpio_direction_output(data->pdata->gpio_ldo_en, 0);
		if (error) {
			dev_err(dev, "unable to set direction for gpio [%u]\n",
				data->pdata->gpio_ldo_en);
			goto fail_release_gpio_ldo_en;
		}

                ATMEL_MSG("gpio_ldo_en [%u]\n", data->pdata->gpio_ldo_en);

	} else {
		dev_info(dev, "vdd control gpio not provided\n");
	}

	if (gpio_is_valid(data->pdata->gpio_irq)) {
		/* configure touchscreen irq gpio */
		error = gpio_request(data->pdata->gpio_irq, "mxt_gpio_irq");
		if (error) {
			dev_err(dev, "unable to request gpio [%u]\n",
				data->pdata->gpio_irq);
			goto fail_release_gpio_ldo_en;
		}
		error = gpio_direction_input(data->pdata->gpio_irq);
		if (error) {
			dev_err(dev, "unable to set direction for gpio [%u]\n",
				data->pdata->gpio_irq);
			goto fail_release_gpios;
		}

                ATMEL_MSG("gpio_irq [%u]\n", data->pdata->gpio_irq);

	} else {
		dev_err(dev, "irq gpio not provided\n");
		goto fail_release_gpio_ldo_en;
	}

	/* FIXME: reset gpio */

	return 0;

fail_release_gpios:
	gpio_free(data->pdata->gpio_irq);
fail_release_gpio_ldo_en:
	if (gpio_is_valid(data->pdata->gpio_ldo_en)) {
		gpio_free(data->pdata->gpio_ldo_en);
	}
fail_release_gpio_io_ldo_en:
	if (gpio_is_valid(data->pdata->gpio_io_ldo_en)) {
		gpio_free(data->pdata->gpio_io_ldo_en);
	}
fail_release_gpio_dcdc_en:
	gpio_free(data->pdata->gpio_dcdc_en);
fail:
	return error;
}

static void mxt_gpio_free(struct mxt_data *data)
{
	ATMEL_TRACE();

	gpio_free(data->pdata->gpio_dcdc_en);
	if (gpio_is_valid(data->pdata->gpio_io_ldo_en)) {
		gpio_free(data->pdata->gpio_io_ldo_en);
	}
	if (gpio_is_valid(data->pdata->gpio_ldo_en)) {
		gpio_free(data->pdata->gpio_ldo_en);
	}
	gpio_free(data->pdata->gpio_irq);

	if (data->pdata->ts_pinctrl != NULL) {
		int retval = pinctrl_select_state(data->pdata->ts_pinctrl,
						  data->pdata->pinctrl_state_suspend);
		if (retval < 0) {
			struct device *dev = &data->client->dev;
			dev_err(dev, "%s: Failed to select %s pinstate %d\n",
				__func__, PINCTRL_STATE_SUSPEND, retval);
		}

		devm_pinctrl_put(data->pdata->ts_pinctrl);
		data->pdata->ts_pinctrl = NULL;
		data->pdata->pinctrl_state_active = NULL;
		data->pdata->pinctrl_state_suspend = NULL;
	}
}

static int mxt_power_init(struct mxt_data *data)
{
	struct device *dev = &data->client->dev;
	int error = 0;

	ATMEL_TRACE();

	if (gpio_is_valid(data->pdata->gpio_ldo_en)) {
		data->reg_vdd = NULL;
		data->reg_avdd = NULL;
		return 0;
	}

	data->reg_vdd = regulator_get(dev, "vdd");
	if (IS_ERR(data->reg_vdd)) {
		error = PTR_ERR(data->reg_vdd);
		dev_err(dev, "Error %d getting vdd regulator\n", error);
		goto end;
	}

        error = regulator_set_voltage(data->reg_vdd, 3300000, 3300000);
        if (error) {
            dev_err(dev, "%s set_voltage fail, error=%d\n", __func__, error);
            goto set_regulator_conf_fail;
        }
        ATMEL_DBG("vdd get voltage=%d\n", regulator_get_voltage(data->reg_vdd));

        regulator_set_load(data->reg_vdd, 150000);

	data->reg_avdd = NULL;

	return 0;

set_regulator_conf_fail:
        regulator_put(data->reg_vdd);
end:
	data->reg_vdd = NULL;
	data->reg_avdd = NULL;
	return error;
}

static void mxt_power_deinit(struct mxt_data *data)
{
	ATMEL_TRACE();

	if (gpio_is_valid(data->pdata->gpio_ldo_en)) {
		goto end;
	}

	if (data->poweron) {
            regulator_disable(data->reg_vdd);
        }

	if (data->reg_vdd) regulator_put(data->reg_vdd);
	if (data->reg_avdd) regulator_put(data->reg_avdd);

end:
	data->reg_vdd = NULL;
	data->reg_avdd = NULL;

	return;
}

static int mxt_read_t100_config(struct mxt_data *data)
{
	struct i2c_client *client = data->client;
	int error;
	struct mxt_object *object;
	u16 range_x, range_y;
	u8 cfg, tchaux;
	u8 aux;

	ATMEL_TRACE();

	object = mxt_get_object(data, MXT_TOUCH_MULTITOUCHSCREEN_T100);
	if (!object)
		return -EINVAL;

	error = __mxt_read_reg(client,
			       object->start_address + MXT_T100_XRANGE,
			       sizeof(range_x), &range_x);
	if (error)
		return error;

	le16_to_cpus(range_x);

	error = __mxt_read_reg(client,
			       object->start_address + MXT_T100_YRANGE,
			       sizeof(range_y), &range_y);
	if (error)
		return error;

	le16_to_cpus(range_y);

	error =  __mxt_read_reg(client,
				object->start_address + MXT_T100_CFG1,
				1, &cfg);
	if (error)
		return error;

	error =  __mxt_read_reg(client,
				object->start_address + MXT_T100_TCHAUX,
				1, &tchaux);
	if (error)
		return error;

	/* Handle default values */
	if (range_x == 0)
		range_x = 1023;

	if (range_y == 0)
		range_y = 1023;

	if (cfg & MXT_T100_CFG_SWITCHXY) {
		data->max_x = range_y;
		data->max_y = range_x;
	} else {
		data->max_x = range_x;
		data->max_y = range_y;
	}

	/* allocate aux bytes */
	aux = 6;

	if (tchaux & MXT_T100_TCHAUX_VECT)
		data->t100_aux_vect = aux++;

	if (tchaux & MXT_T100_TCHAUX_AMPL)
		data->t100_aux_ampl = aux++;

	if (tchaux & MXT_T100_TCHAUX_AREA)
		data->t100_aux_area = aux++;

	ATMEL_MSG("T100 Touchscreen size X%uY%u\n", data->max_x, data->max_y);

	/* T100 object size might change, thus free allocated buffer */
	kfree(data->T100_data);

	/* allocate memory to keep a copy of T100 */
	data->T100_data = kzalloc(mxt_obj_size(object), GFP_KERNEL);
	if (!data->T100_data)
		dev_warn(&client->dev, "Cannot allocate T100 data buffer\n");
	else {
		error = __mxt_read_reg(client, object->start_address,
				mxt_obj_size(object), data->T100_data);
		if (error) {
			dev_warn(&client->dev, "T100: failed to store data\n");
			kfree(data->T100_data);
			data->T100_data = NULL;
		} else
			dev_dbg(&client->dev, "T100: stored %zu bytes\n",
				mxt_obj_size(object));
	}

	/* read panel design and configure parameter */
	{
		u8 num_line_y = 0;
		u8 pitch_y = 0;
		u32 panel_width;

		/* panel lines and pitch to determin coordinate */
		error = __mxt_read_reg(client,
				       object->start_address + MXT_T100_YSIZE,
				       sizeof(num_line_y), &num_line_y);
		if (error != 0) {
			num_line_y = 14; /* use default value when error occurred */
		}
		error = __mxt_read_reg(client,
				       object->start_address + MXT_T100_YPITCH,
				       sizeof(pitch_y), &pitch_y);
		if (error != 0) {
			pitch_y = 42; /* use default value when error occurred */
		}

		/* calculate mm to dot convert factor */
		panel_width = num_line_y * pitch_y; /* unit 0.1mm */
		tp_mm2dot_factor = (((range_y + 1) * 10 / panel_width) << TP_MM2DOT_SHIFT);

		ATMEL_DBG("factor=%u, line=%u, pitch=%u\n",
			  tp_mm2dot_factor, num_line_y, pitch_y);
	}

	return 0;
}

static int mxt_initialize_t100_input_device(struct mxt_data *data)
{
	struct device *dev = &data->client->dev;
	struct input_dev *input_dev;
	int i, error;

	ATMEL_TRACE();

	error = mxt_read_t100_config(data);
	if (error)
		dev_warn(dev, "Failed to initialize T100 resolution\n");

	input_dev = data->input_dev;
#ifdef MXT_DISTINCT_INPUT_NAME
	input_dev->name = "Atmel maXTouch T100 Touchscreen";
#endif
	set_bit(INPUT_PROP_DIRECT, input_dev->propbit);

	/* For multi touch */
#ifdef INPUT_MT_POINTER
	/* new version of input_mt */
	input_mt_init_slots(input_dev, data->num_touchids, 0);
#else
	/* old version of input_mt */
	input_mt_init_slots(input_dev, data->num_touchids);
#endif
	data->num_touched = 0;
	data->touch_state = kcalloc(data->num_touchids, sizeof(u8), GFP_KERNEL);

	if (mxt_t47_stylus_state(data))
		input_set_abs_params(input_dev, ABS_MT_TOOL_TYPE,
				     0, MT_TOOL_MAX, 0, 0);

	input_set_abs_params(input_dev, ABS_MT_POSITION_X,
			     0, data->max_x, 0, 0);
	input_set_abs_params(input_dev, ABS_MT_POSITION_Y,
			     0, data->max_y, 0, 0);

	/* if (data->t100_aux_area) */
	/* 	input_set_abs_params(input_dev, ABS_MT_TOUCH_MAJOR, */
	/* 			     0, MXT_MAX_AREA, 0, 0); */

	/* if (data->t100_aux_ampl) */
	/* 	input_set_abs_params(input_dev, ABS_MT_PRESSURE, */
	/* 			     0, MXT_MAX_PRESSURE, 0, 0); */

	/* if (data->t100_aux_vect) */
	/* 	input_set_abs_params(input_dev, ABS_MT_ORIENTATION, */
	/* 			     0, 255, 0, 0); */

	input_set_abs_params(input_dev, ABS_MT_TOUCH_MAJOR,
			     0, TP_MM2DOT(MXT_MAX_AREA), 0, 0);

	/* input_set_abs_params(input_dev, ABS_MT_TOUCH_MINOR, */
	/* 		     0, MXT_MAX_AREA, 0, 0); */

	input_set_abs_params(input_dev, ABS_MT_PRESSURE,
			     0, MXT_MAX_PRESSURE, 0, 0);

	/* For T15 key array */
	if (data->T15_reportid_min) {
		data->t15_keystatus = 0;

		for (i = 0; i < data->pdata->t15_num_keys; i++)
			if (data->pdata->t15_keymap[i])
				input_set_capability(input_dev, EV_KEY,
					     data->pdata->t15_keymap[i]);
	}

	input_set_capability(input_dev, EV_KEY, KEY_POWER);
	input_set_capability(input_dev, EV_KEY, BTN_STYLUS);
	input_set_capability(input_dev, EV_KEY, BTN_STYLUS2);

	return 0;
}

static int mxt_initialize_t9_input_device(struct mxt_data *data);
static int mxt_configure_objects(struct mxt_data *data);
static int mxt_apply_tdat_tsett(struct mxt_data *data);
static int mxt_alloc_input_device(struct mxt_data *data);

static void mxt_request_irq(struct mxt_data *data, unsigned long flags)
{
	int error;

	ATMEL_TRACE();

	dev_dbg(&data->client->dev, "requesting IRQ, flags: %lu\n", flags);
	error = request_threaded_irq(data->irq, NULL, mxt_interrupt,
				     flags, data->client->name, data);
	/* no need to stay alive, touch is not functional */
	BUG_ON(error);
	data->irq_enabled = true;
}

static int mxt_initialize(struct mxt_data *data)
{
	struct i2c_client *client = data->client;
	int error, current_state = mxt_get_sensor_state(data);
	bool alt_bootloader_addr = false;
	bool retry = false;

	ATMEL_TRACE();

retry_info:
	error = mxt_read_info_block(data);
        dev_err(&client->dev, "mxt_read_info_block() ret=%d\n", error);

	queue_delayed_work(data->main_wq, &data->parse_mode_conf_work,0);

	if (error) {
		return error;
	}
	error = 0;
        data->in_bootloader = false;
	data->mode_is_persistent = true;
	if (error) {
retry_bootloader:
		error = mxt_probe_bootloader(data, alt_bootloader_addr);
		if (error) {
			if (alt_bootloader_addr) {
				/* Chip is not in appmode or bootloader mode */
				return error;
			}

			dev_info(&client->dev, "Trying alternate bootloader address\n");
			alt_bootloader_addr = true;
			goto retry_bootloader;
		} else {
			if (retry) {
				dev_err(&client->dev,
						"Could not recover device from "
						"bootloader mode\n");
				/* this is not an error state, we can reflash
				 * from here */
				mxt_set_sensor_state(data, STATE_BL);
				return 0;
			}

			/* Attempt to exit bootloader into app mode */
			mxt_send_bootloader_cmd(data, false);
			msleep(MXT_FW_RESET_TIME);
			retry = true;
			goto retry_info;
		}
	}

	error = mxt_check_retrigen(data);
	if (error)
		return error;

	(void)mxt_acquire_irq(data);

	/* Touch IC is in UI mode, re-register to level */
	/* triggered IRQ mode per Atmel specification   */
	/* free_irq(data->irq, data); */
	mxt_request_irq(data, data->pdata->irqflags);

	error = mxt_configure_objects(data);
	if (error)
		return error;

	mxt_set_sensor_state(data, (current_state == STATE_UNKNOWN) ?
					STATE_STANDBY : STATE_ACTIVE);
	return 0;
}

static int mxt_configure_objects(struct mxt_data *data)
{
	struct i2c_client *client = data->client;
	int error;

	ATMEL_TRACE();

	if (!data->input_dev) {
		error = mxt_alloc_input_device(data);
		if (error) {
			dev_err(&client->dev,
				"Failed to allocate input device\n");
			return error;
		}
	}

	mxt_debug_msg_remove(data);
	error = mxt_debug_msg_init(data);
	if (error)
		return error;

	error = mxt_init_t7_power_cfg(data);
	if (error) {
		dev_err(&client->dev, "Failed to initialize power cfg\n");
		return error;
	}
	mxt_t6_command(data, MXT_COMMAND_CALIBRATE, 1, false);

	/* Check register init values */
	if (data->tdat)
		error = mxt_apply_tdat_tsett(data);
	else
		error = mxt_check_reg_init(data);
	if (error) {
		dev_err(&client->dev, "Error %d initializing configuration\n",
			error);
		return error;
	}

	if (data->T9_reportid_min) {
		error = mxt_initialize_t9_input_device(data);
		if (error)
			return error;
	} else if (data->T100_reportid_min) {
		error = mxt_initialize_t100_input_device(data);
		if (error)
			return error;
	} else
		dev_warn(&client->dev, "No touch object detected\n");

	if (!data->in_bootloader && !data->input_registered) {
		error = input_register_device(data->input_dev);
		if (error) {
			dev_err(&client->dev,
				"Error %d registering input device\n", error);
			return error;
		}
		data->input_registered = true;
	}

	mxt_update_t7_config(data);

	return 0;
}


/* Firmware Version is returned as Major.Minor.Build */
static ssize_t mxt_fw_version_show(struct device *dev,
				   struct device_attribute *attr, char *buf)
{
	struct mxt_data *data = dev_get_drvdata(dev);
	return scnprintf(buf, PAGE_SIZE, "%u.%u.%02X, crc=%06X\n",
			 data->info->version >> 4, data->info->version & 0xf,
			 data->info->build, data->info_crc);
}

/* Hardware Version is returned as FamilyID.VariantID */
static ssize_t mxt_hw_version_show(struct device *dev,
				   struct device_attribute *attr, char *buf)
{
	struct mxt_data *data = dev_get_drvdata(dev);
	return scnprintf(buf, PAGE_SIZE, "%u.%u\n",
			data->info->family_id, data->info->variant_id);
}

/* Build Id is returned as FamilyID-ConfigID */
static ssize_t mxt_buildid_show(struct device *dev,
				   struct device_attribute *attr, char *buf)
{
	struct mxt_data *data = dev_get_drvdata(dev);
	struct mxt_info *info = data->info ? data->info : &data->pdata->dt_info;
	return scnprintf(buf, PAGE_SIZE, "%02x%02x%02x-%08x\n",
			info->version >> 4,
			info->version & 0xf,
			info->build,
			data->config_id);
}

/* Product Id returned as FamilyID */
static ssize_t mxt_productinfo_show(struct device *dev,
				   struct device_attribute *attr, char *buf)
{
	struct mxt_data *data = dev_get_drvdata(dev);
	struct mxt_info *info = data->info ? data->info : &data->pdata->dt_info;
	return scnprintf(buf, PAGE_SIZE, "%02x%02x\n",
				info->family_id, data->revision_id);
}

static ssize_t mxt_poweron_store(struct device *dev, struct device_attribute *attr,
				 const char *buf, size_t count)
{
	struct mxt_data *data = dev_get_drvdata(dev);
	unsigned long value = 0;
	int err = 0;

	err = kstrtoul(buf, 10, &value);
	if (err < 0) {
		dev_err(dev, "%s: Failed to convert value.\n", __func__);
		return -EINVAL;
	}

	switch (value) {
	case 0: /* power off */
		mxt_suspend(dev, MXT_SUSPEND_SYNC);
		break;
	case 1: /* power on */
		mxt_resume(dev);
		break;
	case 2: /* forcibly reset by power off->on */
		mxt_irq_enable(data, false);
		mxt_regulator_disable(data);
		msleep(MXT_REBOOT_TIME);
		mxt_regulator_enable(data);
		mxt_irq_enable(data, true);
		break;
	default:
		dev_err(dev, "%s: Invalid value\n", __func__);
		return -EINVAL;
	}

	return count;
}

static ssize_t mxt_poweron_show(struct device *dev,
		struct device_attribute *attr, char *buf)
{
	struct mxt_data *data = dev_get_drvdata(dev);
	return scnprintf(buf, PAGE_SIZE, "%d\n", data->poweron);
}

static ssize_t mxt_ic_ver_show(struct device *dev,
				   struct device_attribute *attr, char *buf)
{
	struct mxt_data *data = dev_get_drvdata(dev);
	struct mxt_info *info = data->info ? data->info : &data->pdata->dt_info;
	struct i2c_client *client = data->client;
	u8 userdata[10] = {0};
	u32 config_id = 0;
	u32 vendor_config_id = 0;
	u32 vendor_local_rev = 0;
	int error;

	error = __mxt_read_reg(client, data->T38_address,
			       sizeof(userdata), userdata);
	if (error) {
		dev_err(&client->dev, "Unable to read userdata\n");
		return scnprintf(buf, PAGE_SIZE, "cannot read version info\n");
	}

	config_id = be32_to_cpu(*(unsigned int *)&userdata[0]);
	vendor_config_id = be32_to_cpu(*(unsigned int *)&userdata[4]);
	vendor_local_rev = be16_to_cpu(*(unsigned short *)&userdata[8]);

	return scnprintf(buf, PAGE_SIZE,
			 "%s%02x%02x(%02x)\n%s%02x%02x%02x\n%s%08x\n%s%08X.%04X\n",
			 "Product ID: ",
			 info->family_id,
			 data->revision_id,
			 info->variant_id,
			 "Build ID: ",
			 info->version >> 4,
			 info->version & 0xf,
			 info->build,
			 "Config ID: ",
			 config_id,
			 "Vendor config ID: ",
			 vendor_config_id,
			 vendor_local_rev);
}

static ssize_t mxt_tp_cfg_show(struct device *dev,
				   struct device_attribute *attr, char *buf)
{
	struct mxt_data *data = dev_get_drvdata(dev);
	struct i2c_client *client = data->client;
	u8 tp_cfg;
	int ret;

	ret = __mxt_read_reg(client, data->T38_address + MXT_T38_VENDOR_VER,
			       sizeof(tp_cfg), &tp_cfg);
	if (ret) {
		dev_err(&client->dev, "Unable to read userdata\n");
		return scnprintf(buf, PAGE_SIZE, "cannot touchpanel version info\n");
	}

	if (tp_cfg != 0x0A){
		tp_cfg = 0x01;
	} else {
		tp_cfg = 0x02;
	}

	return scnprintf(buf, PAGE_SIZE, "0x%02X\n",tp_cfg);
}

static ssize_t mxt_drv_irq_show(struct device *dev,
		struct device_attribute *attr, char *buf)
{
	struct mxt_data *data = dev_get_drvdata(dev);
	return scnprintf(buf, PAGE_SIZE, "%s\n",
			data->irq_enabled ? "ENABLED" : "DISABLED");
}

static ssize_t mxt_drv_irq_store(struct device *dev,
		struct device_attribute *attr, const char *buf, size_t count)
{
	struct mxt_data *data = dev_get_drvdata(dev);
	unsigned long value = 0;
	int err = 0;

	err = kstrtoul(buf, 10, &value);
	if (err < 0) {
		dev_err(dev, "%s: Failed to convert value.\n", __func__);
		return -EINVAL;
	}

	switch (value) {
	case 0: /* Disable irq */
		mxt_irq_enable(data, false);
		data->enable_reporting = false;
		break;
	case 1: /* Enable irq */
		mxt_irq_enable(data, true);
		data->enable_reporting = true;
		break;
	default:
		dev_err(dev, "%s: Invalid value\n", __func__);
		return -EINVAL;
	}
	return count;
}

static ssize_t mxt_hw_irqstat_show(struct device *dev,
		struct device_attribute *attr, char *buf)
{
	struct mxt_data *data = dev_get_drvdata(dev);
	switch (gpio_get_value(data->pdata->gpio_irq)) {
	case 0:
		return scnprintf(buf, PAGE_SIZE, "Low\n");
	case 1:
		return scnprintf(buf, PAGE_SIZE, "High\n");
	default:
		dev_err(dev, "%s: Failed to get GPIO for irq %d.\n",
				__func__,
				data->irq);
		return scnprintf(buf, PAGE_SIZE, "Unknown\n");
	}
}

static int mxt_hw_reset(struct mxt_data *data)
{
	mxt_regulator_disable(data);
	msleep(MXT_REBOOT_TIME);
	mxt_regulator_enable(data);

	mxt_wait_for_idle(data);
	return mxt_acquire_irq(data);
}

static ssize_t mxt_reset_store(struct device *dev,
		struct device_attribute *attr, const char *buf, size_t count)
{
	unsigned int reset;
	struct mxt_data *data = dev_get_drvdata(dev);

	if (sscanf(buf, "%u", &reset) != 1)
		return -EINVAL;

	if (reset != 1)
		return -EINVAL;

	if (atomic_read(&data->suspended))
		mxt_resume(&data->client->dev);
	else {
		data->enable_reporting = false;
		(void)mxt_hw_reset(data);
		data->enable_reporting = true;
	}

	return count;
}

static ssize_t mxt_show_instance(char *buf, int count,
				 struct mxt_object *object, int instance,
				 const u8 *val)
{
	int i;

	if (mxt_obj_instances(object) > 1)
		count += scnprintf(buf + count, PAGE_SIZE - count,
				   "Instance %u\n", instance);

	for (i = 0; i < mxt_obj_size(object); i++)
		count += scnprintf(buf + count, PAGE_SIZE - count,
				"\t[%2u]: %02x (%d)\n", i, val[i], val[i]);
	count += scnprintf(buf + count, PAGE_SIZE - count, "\n");

	return count;
}

static ssize_t mxt_object_show(struct device *dev,
				    struct device_attribute *attr, char *buf)
{
	struct mxt_data *data = dev_get_drvdata(dev);
	struct mxt_object *object;
	int count = 0;
	int i, j;
	int error;
	u8 *obuf;

	/* Pre-allocate buffer large enough to hold max sized object. */
	obuf = kmalloc(256, GFP_KERNEL);
	if (!obuf)
		return -ENOMEM;

	error = 0;
	for (i = 0; i < data->info->object_num; i++) {
		object = data->object_table + i;

		if (!mxt_object_readable(object->type))
			continue;

		count += scnprintf(buf + count, PAGE_SIZE - count,
				"T%u:\n", object->type);

		for (j = 0; j < mxt_obj_instances(object); j++) {
			u16 size = mxt_obj_size(object);
			u16 addr = object->start_address + j * size;

			error = __mxt_read_reg(data->client, addr, size, obuf);
			if (error)
				goto done;

			count = mxt_show_instance(buf, count, object, j, obuf);
		}
	}

done:
	kfree(obuf);
	return error ?: count;
}

static int mxt_check_firmware_format(struct device *dev,
				     const struct firmware *fw)
{
	unsigned int pos = 0;
	char c;

	ATMEL_TRACE();

	while (pos < fw->size) {
		c = *(fw->data + pos);

		if (c < '0' || (c > '9' && c < 'A') || c > 'F')
			return 0;

		pos++;
	}

	/* To convert file try
	 * xxd -r -p mXTXXX__APP_VX-X-XX.enc > maxtouch.fw */
	dev_err(dev, "Aborting: firmware file must be in binary format\n");

	return -EINVAL;
}

static int mxt_load_fw(struct device *dev)
{
	struct mxt_data *data = dev_get_drvdata(dev);
#if (USE_REQUEST_FIRMWARE == 1)
	const struct firmware *fw = NULL;
#else
	struct firmware *fw = NULL;
	u8 *fwbuf;
#endif
	unsigned int frame_size;
	unsigned int pos = 0;
	unsigned int retry = 0;
	unsigned int frame = 0;
	int ret;

	ATMEL_TRACE();

#if (USE_REQUEST_FIRMWARE == 1)
	ret = request_firmware(&fw, data->fw_name, dev);
	if (ret) {
		dev_err(dev, "Unable to open firmware %s\n", data->fw_name);
		return ret;
	}
#else
	{
		const int FW_SIZE = 524288; /* 512KB */
		const int FW_HEADER_SIZE = 22; /* 22B */
		struct file *filep;
		int num_read = 0;
		int fwsize;

		filep = filp_open(data->fw_name, O_RDONLY|O_LARGEFILE, 0);
		if (filep == NULL) {
			ATMEL_ERR("can't open %s", data->fw_name);
			return 0;
		}

		fwsize = i_size_read(file_inode(filep));
		if ((fwsize <= 0) || (fwsize > FW_SIZE)) {
			fput(filep);
			return -EINVAL;
		}

		fwbuf = kzalloc(FW_SIZE, GFP_KERNEL);
		if (fwbuf == NULL) {
			fput(filep);
			return 0;
		}

		num_read = kernel_read(filep, 0, fwbuf, fwsize);
		fput(filep);
		kfree(data->fw_name);
		data->fw_name = NULL;

		fw = kzalloc(sizeof(struct firmware), GFP_KERNEL);
		if (fw == NULL) {
			ret = -ENOMEM;
			kfree(fwbuf);
			return ret;
		}
		fw->size = num_read - FW_HEADER_SIZE;
		fw->data = fwbuf + FW_HEADER_SIZE;

		ATMEL_DBG("fw num_read=%d\n", num_read);

		if (mxt_check_for_fw_update(data, fwbuf) == 0) {
			ret = 0;
			goto release_firmware;
		}
	}
#endif

	/* Check for incorrect enc file */
	ret = mxt_check_firmware_format(dev, fw);
	if (ret)
		goto release_firmware;

	reinit_completion(&data->bl_completion);

	if (mxt_get_object(data, MXT_TOUCH_MULTITOUCHSCREEN_T100) == NULL) {
		(void)mxt_lookup_bootloader_address(data, false);
		data->in_bootloader = true;
		mxt_irq_enable(data, false);
		free_irq(data->irq, data);
		mxt_request_irq(data, IRQF_TRIGGER_FALLING | IRQF_ONESHOT);
	}

	if (!data->in_bootloader) {
		mxt_irq_enable(data, false);

		/* Change to the bootloader mode */
		ret = mxt_t6_command(data, MXT_COMMAND_RESET,
				     MXT_BOOT_VALUE, false);
		if (ret)
			goto release_firmware;

		msleep(MXT_RESET_TIME);

		/* At this stage, do not need to scan since we know
		 * family ID */
		ret = mxt_lookup_bootloader_address(data, 0);
		if (ret)
			goto release_firmware;

		/* Level triggered IRQ causes WD reset due to soft IRQ   */
		/* lockup, thus change to edge for the duration of flash */
		free_irq(data->irq, data);
		mxt_request_irq(data, IRQF_TRIGGER_FALLING | IRQF_ONESHOT);
	}

	mxt_set_sensor_state(data, STATE_INIT);
	mxt_free_object_table(data);

	ATMEL_DBG("confirming bootloader mode\n");
	ret = mxt_check_bootloader(data, MXT_WAITING_BOOTLOAD_CMD, false);
	if (ret) {
		/* Bootloader may still be unlocked from previous update
		 * attempt */
		ret = mxt_check_bootloader(data, MXT_WAITING_FRAME_DATA, false);
		if (ret)
			goto release_firmware;
	} else {
		dev_info(dev, "Unlocking bootloader\n");
		/* Unlock bootloader */
		ret = mxt_send_bootloader_cmd(data, true);
		if (ret)
			goto release_firmware;
	}

	ATMEL_DBG("start flashing frames\n");
	mxt_set_sensor_state(data, STATE_FLASH);
	while (pos < fw->size) {
		ret = mxt_check_bootloader(data, MXT_WAITING_FRAME_DATA, true);
		if (ret)
			goto release_firmware;

		frame_size = ((*(fw->data + pos) << 8) | *(fw->data + pos + 1));

		/* Take account of CRC bytes */
		frame_size += 2;

		/* Write one frame to device */
		ret = mxt_bootloader_write(data, fw->data + pos, frame_size);
		if (ret)
			goto release_firmware;

		ret = mxt_check_bootloader(data, MXT_FRAME_CRC_PASS, true);
		if (ret) {
			retry++;

			/* Back off by 20ms per retry */
			msleep(retry * 20);

			if (retry > 20) {
				dev_err(dev, "Retry count exceeded\n");
				goto release_firmware;
			}
		} else {
			retry = 0;
			pos += frame_size;
			frame++;
		}

		if (frame % 50 == 0)
			ATMEL_DBG("Sent %d frames, %d/%zd bytes\n",
				  frame, pos, fw->size);
	}

	ATMEL_DBG("Sent %d frames, %u bytes\n", frame, pos);
	reinit_completion(&data->bl_completion);
	/* Wait for flash. */
	ret = mxt_wait_for_completion(data, &data->bl_completion,
				MXT_FW_RESET_TIME);
	if (ret)
		goto release_firmware;

	ATMEL_MSG("Firmware update succeeded\n");

	/* /\* Wait for device to reset. Some bootloader versions do not assert */
	/*  * the CHG line after bootloading has finished, so ignore error *\/ */
	/* mxt_wait_for_completion(data, &data->bl_completion, */
	/* 			MXT_FW_RESET_TIME); */

release_firmware:
#if (USE_REQUEST_FIRMWARE == 1)
	release_firmware(fw);
#else
	kfree(fwbuf);
	kfree(fw);
#endif
	return ret;
}

static int mxt_update_file_name(struct device *dev, char **file_name,
				const char *buf, size_t count)
{
	char *file_name_tmp;

	ATMEL_TRACE();

	/* Simple sanity check */
	if (count > TPIF_FW_FILE_NAME_SIZE) {
		dev_warn(dev, "File name too long\n");
		return -EINVAL;
	}

	file_name_tmp = krealloc(*file_name, count + 1, GFP_KERNEL);
	if (!file_name_tmp) {
		dev_warn(dev, "no memory\n");
		return -ENOMEM;
	}

	*file_name = file_name_tmp;
	memcpy(*file_name, buf, count);

	/* Echo into the sysfs entry may append newline at the end of buf */
	if (buf[count - 1] == '\n')
		(*file_name)[count - 1] = '\0';
	else
		(*file_name)[count] = '\0';

	return 0;
}

static ssize_t mxt_update_fw_store(struct device *dev,
					struct device_attribute *attr,
					const char *buf, size_t count)
{
	struct mxt_data *data = dev_get_drvdata(dev);
	int error;

	ATMEL_TRACE();

	error = mxt_update_file_name(dev, &data->fw_name, buf, count);
	if (error)
		return error;

	mxt_stop_healthcheck(data);
	mxt_stop_nfc_sensing(data);

	error = mxt_load_fw(dev);
	if (error) {
		ATMEL_ERR("The firmware update failed(%d)\n", error);
		count = error;
	} else {
		ATMEL_MSG("The firmware update succeeded\n");

		/* interrupt handler is released to switch irq in mxt_initialize() */
		mxt_irq_enable(data, false);
		free_irq(data->irq, data);
		mxt_free_input_device(data);

		mxt_set_sensor_state(data, STATE_QUERY);

		error = mxt_initialize(data);
		if (error) {
			data->init_completed = 0;
			return error;
		} else {
			data->init_completed = 1;
		}

		if (atomic_read(&data->suspended))
			mxt_resume(&data->client->dev);
	}

	return count;
}

static ssize_t mxt_update_cfg_store(struct device *dev,
		struct device_attribute *attr,
		const char *buf, size_t count)
{
	struct mxt_data *data = dev_get_drvdata(dev);
	int ret;

	ATMEL_TRACE();

	if (data->in_bootloader) {
		dev_err(dev, "Not in appmode\n");
		return -EINVAL;
	}

	ret = mxt_update_file_name(dev, &data->cfg_name, buf, count);
	if (ret)
		return ret;

	mxt_stop_healthcheck(data);
	mxt_stop_nfc_sensing(data);

	data->enable_reporting = false;
	mxt_free_input_device(data);

	if (atomic_read(&data->suspended))
		mxt_resume(&data->client->dev);

	force_update = 1;
	data->enable_fw_update = 1;

	ret = mxt_configure_objects(data);
	if (ret)
		goto out;

	ret = count;
out:
	data->enable_fw_update = 0;
	data->enable_reporting = true;
	force_update = 0;

	mxt_queue_healthcheck(data);

	return ret;
}

static ssize_t mxt_debug_enable_show(struct device *dev,
	struct device_attribute *attr, char *buf)
{
	struct mxt_data *data = dev_get_drvdata(dev);
	char c;

	c = data->debug_enabled ? '1' : '0';
	return scnprintf(buf, PAGE_SIZE, "%c\n", c);
}

static ssize_t mxt_debug_notify_show(struct device *dev,
	struct device_attribute *attr, char *buf)
{
	return scnprintf(buf, PAGE_SIZE, "0\n");
}

static ssize_t mxt_debug_v2_enable_store(struct device *dev,
	struct device_attribute *attr, const char *buf, size_t count)
{
	struct mxt_data *data = dev_get_drvdata(dev);
	int i;

	if (sscanf(buf, "%u", &i) == 1 && i < 2) {
		if (i == 1)
			mxt_debug_msg_enable(data);
		else
			mxt_debug_msg_disable(data);

		return count;
	} else {
		dev_dbg(dev, "debug_enabled write error\n");
		return -EINVAL;
	}
}

static ssize_t mxt_debug_enable_store(struct device *dev,
	struct device_attribute *attr, const char *buf, size_t count)
{
	struct mxt_data *data = dev_get_drvdata(dev);
	int i;

	if (sscanf(buf, "%u", &i) == 1 && i < 2) {
		data->debug_enabled = (i == 1);

		dev_dbg(dev, "%s\n", i ? "debug enabled" : "debug disabled");
		return count;
	} else {
		dev_dbg(dev, "debug_enabled write error\n");
		return -EINVAL;
	}
}

static int mxt_check_mem_access_params(struct mxt_data *data, loff_t off,
				       size_t *count)
{
	ATMEL_TRACE();

	if (off >= data->mem_size)
		return -EIO;

	if (off + *count > data->mem_size)
		*count = data->mem_size - off;

	if (*count > MXT_MAX_BLOCK_WRITE)
		*count = MXT_MAX_BLOCK_WRITE;

	return 0;
}

static ssize_t mxt_mem_access_read(struct file *filp, struct kobject *kobj,
	struct bin_attribute *bin_attr, char *buf, loff_t off, size_t count)
{
	struct device *dev = container_of(kobj, struct device, kobj);
	struct mxt_data *data = dev_get_drvdata(dev);
	int ret = 0;

	ATMEL_TRACE();

	ret = mxt_check_mem_access_params(data, off, &count);
	if (ret < 0)
		return ret;

	if (count > 0)
		ret = __mxt_read_reg(data->client, off, count, buf);

	return ret == 0 ? count : ret;
}

static ssize_t mxt_mem_access_write(struct file *filp, struct kobject *kobj,
	struct bin_attribute *bin_attr, char *buf, loff_t off,
	size_t count)
{
	struct device *dev = container_of(kobj, struct device, kobj);
	struct mxt_data *data = dev_get_drvdata(dev);
	int ret = 0;

	ATMEL_TRACE();

	ret = mxt_check_mem_access_params(data, off, &count);
	if (ret < 0)
		return ret;

	if (count > 0)
		ret = __mxt_write_reg(data->client, off, count, buf);

	return ret == 0 ? count : 0;
}

static int mxt_apply_tdat_tsett(struct mxt_data *data)
{
	struct device *dev = &data->client->dev;
	struct mxt_object *object;
	int ret = 0;
	int offset = 0;
	int byte_offset;
	int cfg_start_ofs;
	u8 *config_mem;
	size_t config_mem_size;
	const unsigned char *cfg_data, *obj_data;
	unsigned int type, instance, size;
	u16 reg;

	pr_info("Applying TDAT settings\n");

	ATMEL_TRACE();

	/* Alloc memory to store configuration */
	cfg_start_ofs = MXT_OBJECT_START
			+ data->info->object_num * sizeof(struct mxt_object)
			+ MXT_INFO_CHECKSUM_SIZE;
	config_mem_size = data->mem_size - cfg_start_ofs;
	config_mem = kzalloc(config_mem_size, GFP_KERNEL);
	if (!config_mem) {
		dev_err(dev, "Failed to allocate memory\n");
		ret = -ENOMEM;
		goto just_leave;
	}
	dev_dbg(dev, "Config memory: size %zd, start_ofs %u\n",
					config_mem_size, cfg_start_ofs);
	cfg_data = data->tsett.data;
	while (offset < data->tsett.size) {
		/* Read type, instance, length */
		type = (cfg_data[offset+1] << 8) | cfg_data[offset];
		instance = cfg_data[offset+2];
		size = (cfg_data[offset+4] << 8) | cfg_data[offset+3];
		obj_data = &cfg_data[offset+5];

		dev_dbg(dev, "T%u, instance: %u, size: %u, Tsett-offset: %d\n",
			type, instance, size, offset);
		offset += size + 5;

		object = mxt_get_object(data, type);
		if (!object)
			continue;

		if (size > mxt_obj_size(object)) {
			/* Either we are in fallback mode due to wrong
			 * config or config from a later fw version,
			 * or the file is corrupt or hand-edited */
			dev_warn(dev, "Discarding %zd byte(s) in T%u\n",
				 size - mxt_obj_size(object), type);
		} else if (mxt_obj_size(object) > size) {
			/* If firmware is upgraded, new bytes may be added to
			 * end of objects. It is generally forward compatible
			 * to zero these bytes - previous behaviour will be
			 * retained. However this does invalidate the CRC and
			 * will force fallback mode until the configuration is
			 * updated. We warn here but do nothing else - the
			 * malloc has zeroed the entire configuration. */
			dev_warn(dev, "Zeroing %zd byte(s) in T%d\n",
				 mxt_obj_size(object) - size, type);
		}

		if (instance >= mxt_obj_instances(object)) {
			dev_err(dev, "Object instances exceeded!\n");
			ret = -EINVAL;
			goto release_mem;
		}

		/* config data address in registers memory map */
		reg = object->start_address + mxt_obj_size(object) * instance;
		dev_dbg(dev, "T%u address: %u, instance[%u] data offset %zd\n",
				type, object->start_address,
				instance, mxt_obj_size(object) * instance);

		/* copy new config data to config memory buffer */
		memcpy(config_mem + reg - cfg_start_ofs, obj_data, size);
		dev_dbg(dev, "Copied %u bytes at offset %u\n",
			size, (unsigned)(reg - cfg_start_ofs));
	}

	/* calculate crc of the received configs (not the raw config file) */
	if (data->T7_address < cfg_start_ofs) {
		dev_err(dev, "Bad T7 address, T7addr = %x, config offset %x\n",
			data->T7_address, cfg_start_ofs);
		goto release_mem;
	}

	/* Write configuration as blocks */
	byte_offset = 0;
	while (byte_offset < config_mem_size) {
		size = config_mem_size - byte_offset;

		if (size > MXT_MAX_BLOCK_WRITE)
			size = MXT_MAX_BLOCK_WRITE;

		dev_dbg(dev, "Writing config chunk size %u, at address %u\n",
			size, cfg_start_ofs + byte_offset);

		ret = __mxt_write_reg(data->client,
				      cfg_start_ofs + byte_offset,
				      size, config_mem + byte_offset);
		if (ret != 0) {
			dev_err(dev, "Config write error, ret=%d\n", ret);
			goto release_mem;
		}

		byte_offset += size;
	}

	mxt_update_crc(data, MXT_COMMAND_BACKUPNV, MXT_BACKUP_VALUE);

	ret = mxt_check_retrigen(data);
	if (ret)
		goto release_mem;

	/* mxt_soft_reset(data); */
	dev_info(dev, "Config written\n");

	/* T7 config may have changed */
	mxt_init_t7_power_cfg(data);
	mxt_t6_command(data, MXT_COMMAND_CALIBRATE, 1, false);

	/* T38 may have changed */
	if (data->T38_address) {
		data->config_id = mxt_update_config_id(data);
		mxt_update_vendor_ver_id(data);
	}

release_mem:
	kfree(config_mem);
just_leave:
	return ret;
}

static bool forcereflash;
static ssize_t mxt_forcereflash_store(struct device *dev,
		struct device_attribute *attr,
		const char *buf, size_t count)
{
	unsigned int input;

	ATMEL_TRACE();

	if (sscanf(buf, "%u", &input) != 1)
		return -EINVAL;

	if (input != 1)
		return -EINVAL;

	forcereflash = true;

	return count;
}

static bool mxt_check_tdat_format(const unsigned char *image)
{
	ATMEL_TRACE();

	return image[0] == 0x31;
}

static int mxt_get_tdat_config_id(const unsigned char *data, size_t size,
	u32 *config_id, u8 *revision)
{
	unsigned short id;
	int error = -EINVAL;
	size_t length, offset;

	ATMEL_TRACE();

	for (offset = 0; offset < size; offset += length+5) {
		id = (data[offset+1] << 8) | data[offset];
		length = (data[offset+4] << 8) | data[offset+3];
		if (id == MXT_SPT_USERDATA_T38) {
			*config_id = be32_to_cpu(
					*(unsigned int *)&data[offset+5]);
			*revision = data[offset+9];
			error = 0;
			break;
		}
	}
	return error;
}

static int mxt_parse_tdat_image(struct mxt_data *data)
{
	struct device *dev = &data->client->dev;
	u32 firmware_id = 0;
	u8 *section, id;
	size_t ii, length, offset, header_sz, raw_size = data->tdat->size;
	const u8 *raw_image = data->tdat->data;
	int error;
	u32 config_id = 0;
	u8 revision_id = 0;

	pr_info("Start TDAT image processing\n");

	ATMEL_TRACE();

	for (ii = 0, offset = 1; offset < raw_size; offset += length+4) {
		length = (raw_image[offset+3] << 16) |
			 (raw_image[offset+2] << 8) | raw_image[offset+1];

		dev_dbg(dev, "Record[%zd]: length %zd, offset %zd\n",
				ii++, length, offset);

		if ((offset+length+4) > raw_size) {
			dev_err(dev, "Data overflow at offset %zd (%u)\n",
				offset, raw_image[offset]);
			return -EINVAL;
		}
	}

	if (offset != raw_size) {
		dev_err(dev, "Data is misaligned\n");
		return -EINVAL;
	}

	for (offset = 1; offset < raw_size; offset += length+4) {
		/* byte 0	- section id     */
		/* byte 1-3	- section length */
		/* byte 4-...	- section data   */
		id = raw_image[offset];
		length = (raw_image[offset+3] << 16) |
			 (raw_image[offset+2] << 8) | raw_image[offset+1];
		section = (u8 *)&raw_image[offset+4];
		dev_dbg(dev, "Section [%u], size %zd\n", id, length);

		switch (id) {
		case 1: /* config */
			error = mxt_get_tdat_config_id(section, length,
						&config_id, &revision_id);
			if (error)
				return -EINVAL;

			pr_info("TDAT: Config ID %x, revision 0x%02x\n",
						config_id, revision_id);
			data->tsett.data = section;
			data->tsett.size = length;
			break;

		case 2: /* firmware */
			firmware_id = be32_to_cpu(
						*(unsigned int *)(&section[1]));
			pr_info("TDAT: FW build ID %x\n", firmware_id);

			/* skip section header */
			header_sz = section[0] + 1;
			data->fw.data = section + header_sz;
			data->fw.size = length - header_sz;
			break;
		default:
			break;
		}
	}

	if (revision_id != data->revision_id) {
		dev_err(dev, "Incorrect firmware (revision id %x <-> %x\n",
				revision_id, data->revision_id);
		return -EINVAL;
	}

	if ((config_id <= data->config_id) && !forcereflash) {
		dev_err(dev, "Firmware upgrade is not required\n");
		return -EINVAL;
	}

	data->config_id = config_id;
	data->firmware_id = firmware_id;
	data->revision_id = revision_id;

	return 0;
}

static ssize_t mxt_doreflash_store(struct device *dev,
		struct device_attribute *attr,
		const char *buf, size_t count)
{
	struct mxt_data *data = dev_get_drvdata(dev);
	unsigned int frame_size;
	unsigned int pos = 0;
	unsigned int retry = 0;
	unsigned int frame = 0;
	int error;

	ATMEL_TRACE();

	error = mxt_update_file_name(dev, &data->fw_name, buf, count);
	if (error)
		return error;

	pr_info("TDAT file name: %s\n", data->fw_name);

	error = request_firmware(&data->tdat, data->fw_name, dev);
	if (error) {
		dev_err(dev, "Unable to load firmware %s\n", data->fw_name);
		return error;
	}

	if (mxt_check_tdat_format(data->tdat->data)) {
		error = mxt_parse_tdat_image(data);
		if (error)
			goto release_firmware;
		forcereflash = false;
	} else {
		dev_err(dev, "invalid TDAT format\n");
		goto release_firmware;
	}

	mxt_lock(&data->crit_section_lock);

	mxt_irq_enable(data, false);

	if (!data->in_bootloader) {
		dev_dbg(dev, "Reboot to bootloader\n\n");
		/* Change to the bootloader mode */
		error = mxt_t6_command(data, MXT_COMMAND_RESET,
				     MXT_BOOT_VALUE, false);
		if (error)
			goto release_firmware;

		msleep(MXT_RESET_TIME);

		/* At this stage, no need to scan since family ID is known */
		error = mxt_lookup_bootloader_address(data, 0);
		if (error)
			goto release_firmware;

		/* Level triggered IRQ causes WD reset due to soft IRQ   */
		/* lockup, thus change to edge for the duration of flash */
		free_irq(data->irq, data);
		mxt_request_irq(data, IRQF_TRIGGER_FALLING | IRQF_ONESHOT);
	} else
		mxt_irq_enable(data, true);

	mxt_set_sensor_state(data, STATE_INIT);

	mxt_free_object_table(data);

	dev_dbg(dev, "confirming bootloader mode\n");
	error = mxt_check_bootloader(data, MXT_WAITING_BOOTLOAD_CMD, false);
	if (error) {
		error = mxt_check_bootloader(data,
					MXT_WAITING_FRAME_DATA, false);
		if (error)
			goto flash_error;
	} else {
		dev_info(dev, "Unlocking bootloader\n");
		/* Unlock bootloader */
		error = mxt_send_bootloader_cmd(data, true);
		if (error)
			goto flash_error;
	}

	dev_dbg(dev, "start flashing frames\n");
	mxt_set_sensor_state(data, STATE_FLASH);
	while (pos < data->fw.size) {
		error = mxt_check_bootloader(data,
					MXT_WAITING_FRAME_DATA, true);
		if (error)
			goto flash_error;

		frame_size = ((*(data->fw.data + pos) << 8) |
				*(data->fw.data + pos + 1));
		/* Take account of CRC bytes */
		frame_size += 2;

		/* Write one frame to device */
		error = mxt_bootloader_write(data,
					data->fw.data + pos, frame_size);
		if (error)
			goto flash_error;

		error = mxt_check_bootloader(data,
					MXT_FRAME_CRC_PASS, true);
		if (error) {
			retry++;
			msleep(retry * 20);
			if (retry > 20) {
				dev_err(dev, "Retry count exceeded\n");
				goto flash_error;
			}
		} else {
			retry = 0;
			pos += frame_size;
			frame++;
		}

		if (frame % 50 == 0)
			dev_info(dev, "Sent %d frames, %u/%zd bytes\n",
				frame, pos, data->fw.size);
	}

	dev_info(dev, "FW: sent %d frames, %u bytes\n", frame, pos);
	reinit_completion(&data->bl_completion);
	error = mxt_wait_for_completion(data, &data->bl_completion,
				MXT_FW_RESET_TIME);
	if (error)
		goto flash_error;

	dev_info(dev, "The firmware update succeeded\n");
	goto initialize;

flash_error:
	mxt_soft_reset(data);

initialize:
	mxt_set_sensor_state(data, STATE_QUERY);
	error = mxt_initialize(data);
	if (error) {
		mxt_set_sensor_state(data, STATE_BL);
		dev_info(dev, "Init failed after firmware upgrade\n");
	}

	mxt_unlock(&data->crit_section_lock);

	memset(&data->fw, 0, sizeof(data->fw));
	memset(&data->tsett, 0, sizeof(data->tsett));

release_firmware:
	release_firmware(data->tdat);
	data->tdat = NULL;
	return count;
}

static ssize_t mxt_flashprog_show(struct device *dev,
				   struct device_attribute *attr, char *buf)
{
	struct mxt_data *data = dev_get_drvdata(dev);

	ATMEL_TRACE();

	return scnprintf(buf, PAGE_SIZE, "%d", data->in_bootloader);
}

enum wakeup_modes {
	WAKEUP_OFF = 0,
	WAKEUP_ON,
	WAKEUP_AUTO
};

static u8 tsp_mode;
static char *wakeup_mode_names[] = { "OFF", "ON", "AUTO" };

static char const *mxt_wakeup_mode_name(int mode)
{
	int index = mode < 0 || mode > WAKEUP_AUTO ? WAKEUP_OFF : mode;

	ATMEL_TRACE();

	return wakeup_mode_names[index];
}

static ssize_t mxt_tsp_store(struct device *dev,
		struct device_attribute *attr,
		const char *buf, size_t count)
{
	struct mxt_data *mxt_dev_data = dev_get_drvdata(dev);
	int state, mode;

	ATMEL_TRACE();

	state = mxt_get_sensor_state(mxt_dev_data);

	pr_debug("state: %s(%d), suspend flag: %d, BL flag: %d\n",
			mxt_state_name(state), state,
			atomic_read(&mxt_dev_data->suspended),
			mxt_dev_data->in_bootloader);

	if (!strncmp(buf, "on", 2) || !strncmp(buf, "ON", 2))
		mode = WAKEUP_ON;
	else if (!strncmp(buf, "auto", 4) || !strncmp(buf, "AUTO", 4))
		mode = WAKEUP_AUTO;
	else if (!strncmp(buf, "off", 2) || !strncmp(buf, "OFF", 2))
		mode = WAKEUP_OFF;
	else
		return -EINVAL;

	if (mode == tsp_mode)
		return count;

	tsp_mode = mode;

	switch (mode) {
	case WAKEUP_AUTO:
		mxt_set_alternate_mode(mxt_dev_data,
			mxt_dev_data->alternate_mode, true, true);
			break;
	case WAKEUP_ON:
		mxt_set_alternate_mode(mxt_dev_data,
			mxt_dev_data->alternate_mode, true, false);
			break;
	case WAKEUP_OFF:
		mxt_restore_default_mode(mxt_dev_data);
			break;
	}

	dev_dbg(&mxt_dev_data->client->dev, "TAP: %s\n",
			mxt_wakeup_mode_name(tsp_mode));
	return count;
}

static ssize_t mxt_tsp_show(struct device *dev,
				   struct device_attribute *attr, char *buf)
{
	return scnprintf(buf, PAGE_SIZE, "%s\n",
			mxt_wakeup_mode_name(tsp_mode));
}

/* noise evaluation function */
#define EVAL_NOISE_INTERVAL 1 /* [jiffies] */

static int eval_noise_running = 0;
static char *eval_noise_msg_buf = NULL;

static void mxt_queued_eval_noise(struct work_struct *w)
{
	struct mxt_data *data =
			container_of(w, struct mxt_data, eval_noise_work);

	ATMEL_TRACE();

	if (eval_noise_running != 0) {
		struct input_dev *input_dev = data->input_dev;
		int ret;
		static int move = 2;
		int x = data->max_x - move - 1, y = data->max_y - move - 1;

		/* dummy read */
		ret = __mxt_read_reg(data->client, data->T5_address,
				     data->T5_msg_size, eval_noise_msg_buf);

		/* send dummy position */
		move *= -1;
		x += move;
		y += move;
		input_mt_slot(input_dev, 0);
		input_mt_report_slot_state(input_dev, MT_TOOL_FINGER, 1);
		input_report_abs(input_dev, ABS_MT_POSITION_X, x);
		input_report_abs(input_dev, ABS_MT_POSITION_Y, y);
		input_report_abs(input_dev, ABS_MT_PRESSURE, 1);
		mxt_input_sync(input_dev, 1);

		ATMEL_DBG("eval_noise send dummy position (%d,%d)\n", x, y);

		/* repeat timer */
		mod_timer(&data->eval_noise_timer, jiffies + EVAL_NOISE_INTERVAL);
	} else {
		ATMEL_DBG("eval_noise already stopped\n");
	}

}


static void mxt_eval_noise_timer_func(unsigned long handle)
{
	struct mxt_data *data = (struct mxt_data *)handle;
	queue_work(system_wq, &data->eval_noise_work);

	return;
}


static void mxt_eval_noise(struct mxt_data *data, int is_start)
{
	static int is_initialized = 0;

	ATMEL_TRACE();

	if (is_start != 0) {
		/* start => disbale touch detection */
		mxt_irq_enable(data, false);

		eval_noise_running = 1;

		if (is_initialized == 0) {
			setup_timer(&data->eval_noise_timer,
				    mxt_eval_noise_timer_func,
				    (unsigned long)data);
			INIT_WORK(&data->eval_noise_work, mxt_queued_eval_noise);
			is_initialized = 1;
		}

		/* allocate temporal message buffer if needed */
		if (eval_noise_msg_buf == NULL) {
			eval_noise_msg_buf = kcalloc(data->max_reportid,
						     data->T5_msg_size, GFP_KERNEL);
		}

		mod_timer(&data->eval_noise_timer, jiffies + EVAL_NOISE_INTERVAL);
	} else {
		struct input_dev *input_dev = data->input_dev;

		eval_noise_running = 0;
		del_timer_sync(&data->eval_noise_timer);
		cancel_work_sync(&data->eval_noise_work);

		/* send dummy position (release) */
		input_mt_slot(input_dev, 0);
		input_mt_report_slot_state(input_dev, MT_TOOL_FINGER, 0);
		mxt_input_sync(input_dev, 0);

		/* release resource */
		if (eval_noise_msg_buf != NULL) {
			kfree(eval_noise_msg_buf);
			eval_noise_msg_buf = NULL;
		}

		/* stop => restart touch detection */
		mxt_irq_enable(data, true);
	}

	return;
}

static ssize_t mxt_eval_noise_store(struct device *dev,
		struct device_attribute *attr,
		const char *buf, size_t count)
{
	struct mxt_data *mxt_dev_data = dev_get_drvdata(dev);
	int cmd_start = 0;

	ATMEL_TRACE();

	if (!strncmp(buf, "1", 1))
		cmd_start = 1;
	else if (!strncmp(buf, "0", 1))
		cmd_start = 0;
	else
		return -EINVAL;

	mxt_eval_noise(mxt_dev_data, cmd_start);

	ATMEL_DBG("eval noise: %s\n",
		  eval_noise_running == 1 ? "start" : " stop");

	return count;
}

static ssize_t mxt_eval_noise_show(struct device *dev,
				   struct device_attribute *attr, char *buf)
{
	return scnprintf(buf, PAGE_SIZE, "eval noise is %s running\n",
			 eval_noise_running == 1 ? "" : "not");
}

static ssize_t mxt_glove_cfg_store(struct device *dev,
				   struct device_attribute *attr,
				   const char *buf, size_t count)
{
	unsigned int data[21];
	int len;
	struct glove_config_t *cfg = NULL;

	ATMEL_TRACE();

	if (buf == NULL) {
		goto glove_cfg_error;
	}

	memset(data, 0, sizeof(data));
	len = sscanf(buf, "%x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x",
		     &data[0], &data[1], &data[2], &data[3], &data[4], &data[5],
		     &data[6], &data[7], &data[8], &data[9], &data[10], &data[11],
		     &data[12], &data[13], &data[14], &data[15], &data[16], &data[17],
		     &data[18], &data[19], &data[20]);
	if (len != ARRAY_SIZE(data)) {
		ATMEL_ERR("input data is invalid\n");
		goto glove_cfg_error;
	}

	switch (data[0]) {
	case 0:
		cfg = &touch_dev_cfg.glove_off_cfg;
		break;
	case 1:
		cfg = &touch_dev_cfg.glove_on_cfg_low;
		break;
	case 2:
		cfg = &touch_dev_cfg.glove_on_cfg_mid;
		break;
	case 3:
		cfg = &touch_dev_cfg.glove_on_cfg_high;
		break;
	default:
		ATMEL_ERR("id is invalid\n");
		cfg = NULL;
		break;
	}

	if (cfg != NULL) {
		cfg->enable		= (u8)data[1];
		cfg->glove_self_thr	= (u8)data[2];
		cfg->glove_hys		= (u8)data[3];
		cfg->discr_thr		= (u8)data[4];
		cfg->active_mutual_self	= (u8)data[5];
		cfg->next_move_filter	= (u8)data[6];
		cfg->max_touch_amp	= (u8)data[7];
		cfg->z_touch_thr	= (u8)data[8];
		cfg->z_touch_thr_hys	= (u8)data[9];
		cfg->init_move_filter	= (u8)data[10];
		cfg->noise_on_cal	= (u8)data[11];
		cfg->noise_on		= (u8)data[12];
		cfg->ins1_dynamic_en	= (u8)data[13];
		cfg->ins2_dynamic_en	= (u8)data[14];
		cfg->ins5_dynamic_en	= (u8)data[15];
		cfg->ins7_dynamic_en	= (u8)data[16];
		cfg->ins8_dynamic_en	= (u8)data[17];
		cfg->long_cal_en	= (u8)data[18];
		cfg->dc_filter		= (u8)data[19];
		cfg->dc_cal_rec_str	= (u8)data[20];
	}

glove_cfg_error:
	return count;
}

static int mxt_print_glove_cfg(char *buf, int buf_size, const char *tag,
			       struct glove_config_t *cfg)
{
	int len = 0;

	ATMEL_TRACE();

	if ((buf != NULL) && (cfg != NULL)) {
		len += scnprintf(buf + len, buf_size - len, "[%s]\n", tag);
		len += scnprintf(buf + len, buf_size - len,
				 "- enable             : 0x%02X\n", cfg->enable);
		len += scnprintf(buf + len, buf_size - len,
				 "- glove_self_thr     : 0x%02X\n", cfg->glove_self_thr);
		len += scnprintf(buf + len, buf_size - len,
				 "- glove_hys          : 0x%02X\n", cfg->glove_hys);
		len += scnprintf(buf + len, buf_size - len,
				 "- discr_thr          : 0x%02X\n", cfg->discr_thr);
		len += scnprintf(buf + len, buf_size - len,
				 "- active_mutual_self : 0x%02X\n", cfg->active_mutual_self);
		len += scnprintf(buf + len, buf_size - len,
				 "- next_move_filter   : 0x%02X\n", cfg->next_move_filter);
		len += scnprintf(buf + len, buf_size - len,
				 "- max_touch_amp      : 0x%02X\n", cfg->max_touch_amp);
		len += scnprintf(buf + len, buf_size - len,
				 "- z_touch_thr        : 0x%02X\n", cfg->z_touch_thr);
		len += scnprintf(buf + len, buf_size - len,
				 "- z_touch_thr_hys    : 0x%02X\n", cfg->z_touch_thr_hys);
		len += scnprintf(buf + len, buf_size - len,
				 "- init_move_filter   : 0x%02X\n", cfg->init_move_filter);
		len += scnprintf(buf + len, buf_size - len,
				 "- noise_on_cal       : 0x%02X\n", cfg->noise_on_cal);
		len += scnprintf(buf + len, buf_size - len,
				 "- noise_on           : 0x%02X\n", cfg->noise_on);
		len += scnprintf(buf + len, buf_size - len,
				 "- ins1_dynamic_en    : 0x%02X\n", cfg->ins1_dynamic_en);
		len += scnprintf(buf + len, buf_size - len,
				 "- ins2_dynamic_en    : 0x%02X\n", cfg->ins2_dynamic_en);
		len += scnprintf(buf + len, buf_size - len,
				 "- ins5_dynamic_en    : 0x%02X\n", cfg->ins5_dynamic_en);
		len += scnprintf(buf + len, buf_size - len,
				 "- ins7_dynamic_en    : 0x%02X\n", cfg->ins7_dynamic_en);
		len += scnprintf(buf + len, buf_size - len,
				 "- ins8_dynamic_en    : 0x%02X\n", cfg->ins8_dynamic_en);
		len += scnprintf(buf + len, buf_size - len,
				 "- long_cal_en        : 0x%02X\n", cfg->long_cal_en);
		len += scnprintf(buf + len, buf_size - len,
				 "- dc_filter          : 0x%02X\n", cfg->dc_filter);
		len += scnprintf(buf + len, buf_size - len,
				 "- dc_cal_rec_str     : 0x%02X\n", cfg->dc_cal_rec_str);
	}

	return len;
}

static ssize_t mxt_glove_cfg_show(struct device *dev,
				  struct device_attribute *attr,
				  char *buf)
{
	int len = 0;

	ATMEL_TRACE();

	if (buf != NULL) {
		len += mxt_print_glove_cfg(buf + len, PAGE_SIZE - len,
					   "off",
					   &touch_dev_cfg.glove_off_cfg);
		len += mxt_print_glove_cfg(buf + len, PAGE_SIZE - len,
					   "sensitivity-low",
					   &touch_dev_cfg.glove_on_cfg_low);
		len += mxt_print_glove_cfg(buf + len, PAGE_SIZE - len,
					   "sensitivity-mid",
					   &touch_dev_cfg.glove_on_cfg_mid);
		len += mxt_print_glove_cfg(buf + len, PAGE_SIZE - len,
					   "sensitivity-high",
					   &touch_dev_cfg.glove_on_cfg_high);
	}

	return len;
}

static ssize_t mxt_glove_store(struct device *dev,
			       struct device_attribute *attr,
			       const char *buf, size_t count)
{
	struct mxt_data *data = dev_get_drvdata(dev);
	int cmd_enable = 0;
	int error = 0;

	ATMEL_TRACE();

	if (!strncmp(buf, "0", 1)) {
		cmd_enable = 0;
	} else {
		cmd_enable = 1;
	}

	error = mxt_set_glove(data, cmd_enable);
	if (error != 0) {
		goto glove_error;
	}

	mxt_update_t7_config(data);

glove_error:
	return count;
}

static ssize_t mxt_glove_show(struct device *dev,
			      struct device_attribute *attr, char *buf)
{
	struct mxt_data *data = dev_get_drvdata(dev);
	struct mxt_object *object;
	int len = 0;
	struct glove_config_t cfg;
	int error;

	memset(&cfg, 0, sizeof(cfg));

	object = mxt_get_object(data, MXT_PROCI_GLOVEDETECTION_T78);
	if (object == NULL) {
		ATMEL_ERR("T%d object not found\n", MXT_PROCI_GLOVEDETECTION_T78);
		return 0;
	}

	error = __mxt_read_reg(data->client, object->start_address,
			       sizeof(cfg.enable), &cfg.enable);
	if (error != 0) {
		ATMEL_ERR("T%d CTRL read failed\n", object->type);
		return 0;
	}

	error = __mxt_read_reg(data->client, object->start_address + 9,
			       sizeof(cfg.glove_self_thr), &cfg.glove_self_thr);
	if (error != 0) {
		ATMEL_ERR("T%d THR read failed\n", object->type);
		return 0;
	}

	error = __mxt_read_reg(data->client, object->start_address + 10,
			       sizeof(cfg.glove_hys), &cfg.glove_hys);
	if (error != 0) {
		ATMEL_ERR("T%d HYS read failed\n", object->type);
		return 0;
	}

	error = __mxt_read_reg(data->client, object->start_address + 8,
			       sizeof(cfg.discr_thr), &cfg.discr_thr);
	if (error) {
		ATMEL_ERR("T78 discr read failed\n");
		return 0;
	}

	object = mxt_get_object(data, MXT_GEN_ACQUIRE_T8);
	if (object == NULL) {
		ATMEL_ERR("T%d object not found\n", MXT_GEN_ACQUIRE_T8);
		return 0;
	}

	error = __mxt_read_reg(data->client, object->start_address + 12,
			       sizeof(cfg.active_mutual_self), &cfg.active_mutual_self);
	if (error != 0) {
		ATMEL_ERR("T%d MUT_SELF read failed\n", object->type);
		return 0;
	}
	cfg.active_mutual_self &= 0x3;

	object = mxt_get_object(data, MXT_TOUCH_MULTITOUCHSCREEN_T100);
	if (object == NULL) {
		ATMEL_ERR("T%d object not found\n", MXT_TOUCH_MULTITOUCHSCREEN_T100);
		return 0;
	}

	error = __mxt_read_reg(data->client, object->start_address + MXT_T100_MOVHYSTN,
			       sizeof(cfg.next_move_filter), &cfg.next_move_filter);
	if (error != 0) {
		ATMEL_ERR("T%d MOV_FILTER read failed\n", object->type);
		return 0;
	}

	error = __mxt_read_reg(data->client, object->start_address + MXT_T100_TCHTHR,
			       sizeof(cfg.z_touch_thr), &cfg.z_touch_thr);
	if (error != 0) {
		ATMEL_ERR("T%d TCHTHR read failed\n", object->type);
		return 0;
	}

	error = __mxt_read_reg(data->client, object->start_address + MXT_T100_TCHTHRHYS,
			       sizeof(cfg.z_touch_thr_hys), &cfg.z_touch_thr_hys);
	if (error != 0) {
		ATMEL_ERR("T%d TCHTHR read failed\n", object->type);
		return 0;
	}
	cfg.z_touch_thr_hys &= 0x3F;

	error = __mxt_read_reg(data->client, object->start_address + MXT_T100_MOVHYSTI,
			       sizeof(cfg.init_move_filter), &cfg.init_move_filter);
	if (error != 0) {
		ATMEL_ERR("T%d MOVHYSTI read failed\n", object->type);
		return 0;
	}

	object = mxt_get_object(data, MXT_PROCG_NOISESUPSELFCAT_T108);
	if (object == NULL) {
		ATMEL_ERR("T%d object not found\n", MXT_PROCG_NOISESUPSELFCAT_T108);
		return 0;
	}

	error = __mxt_read_reg(data->client, object->start_address + 1,
			       sizeof(cfg.noise_on_cal), &cfg.noise_on_cal);
	if (error) {
		ATMEL_ERR("T108 CALCFG1 read failed\n");
		return 0;
	}

	error = __mxt_read_reg(data->client, object->start_address + 2,
			       sizeof(cfg.noise_on), &cfg.noise_on);
	if (error) {
		ATMEL_ERR("T108 CFG1 read failed\n");
		return 0;
	}

	object = mxt_get_object(data, MXT_SPT_DYNAMICCONFIGURATIONCONTROLLER_T70);
	if (object == NULL) {
		ATMEL_ERR("T%d object not found\n", MXT_SPT_DYNAMICCONFIGURATIONCONTROLLER_T70);
		return 0;
	}

	error = __mxt_read_reg(data->client, object->start_address + mxt_obj_size(object) * 1,
			       sizeof(cfg.ins1_dynamic_en), &cfg.ins1_dynamic_en);
	if (error) {
		ATMEL_ERR("T70(1) CTRL read failed\n");
		return 0;
	}

	error = __mxt_read_reg(data->client, object->start_address + mxt_obj_size(object) * 2,
			       sizeof(cfg.ins2_dynamic_en), &cfg.ins2_dynamic_en);
	if (error) {
		ATMEL_ERR("T70(2) CTRL read failed\n");
		return 0;
	}

	error = __mxt_read_reg(data->client, object->start_address + mxt_obj_size(object) * 5,
			       sizeof(cfg.ins5_dynamic_en), &cfg.ins5_dynamic_en);
	if (error) {
		ATMEL_ERR("T70(5) CTRL read failed\n");
		return 0;
	}

	error = __mxt_read_reg(data->client, object->start_address + mxt_obj_size(object) * 7,
			       sizeof(cfg.ins7_dynamic_en), &cfg.ins7_dynamic_en);
	if (error) {
		ATMEL_ERR("T70(7) CTRL read failed\n");
		return 0;
	}

	error = __mxt_read_reg(data->client, object->start_address + mxt_obj_size(object) * 8,
			       sizeof(cfg.ins8_dynamic_en), &cfg.ins8_dynamic_en);
	if (error) {
		ATMEL_ERR("T70(8) CTRL read failed\n");
		return 0;
	}

	object = mxt_get_object(data, MXT_SPT_CTECONFIG_T46);
	if (object == NULL) {
		ATMEL_ERR("T%d object not found\n", MXT_SPT_CTECONFIG_T46);
		return 0;
	}

	error = __mxt_read_reg(data->client, object->start_address,
			       sizeof(cfg.long_cal_en), &cfg.long_cal_en);
	if (error) {
		ATMEL_ERR("T46 CTRL read failed\n");
		return 0;
	}
	cfg.long_cal_en &= 0x04;

	object = mxt_get_object(data, MXT_SPT_SELFCAPCONFIG_T111);
	if (object == NULL) {
		ATMEL_ERR("T%d object not found\n", MXT_SPT_SELFCAPCONFIG_T111);
		return 0;
	}

	error = __mxt_read_reg(data->client, object->start_address + 20,
			       sizeof(cfg.dc_filter), &cfg.dc_filter);
	if (error) {
		ATMEL_ERR("T111 DCFILTER read failed\n");
		return 0;
	}

	error = __mxt_read_reg(data->client, object->start_address + 21,
			       sizeof(cfg.dc_cal_rec_str), &cfg.dc_cal_rec_str);
	if (error) {
		ATMEL_ERR("T111 DCCALRECSTR read failed\n");
		return 0;
	}

	object = mxt_get_object(data, MXT_PROCI_STYLUS_T47);
	if (object == NULL) {
		ATMEL_ERR("T%d object not found\n", MXT_PROCI_STYLUS_T47);
		return 0;
	}

	error = __mxt_read_reg(data->client, object->start_address + 5,
			       sizeof(cfg.max_touch_amp), &cfg.max_touch_amp);
	if (error) {
		ATMEL_ERR("T47 max touch amp read failed\n");
		return 0;
	}

	len += mxt_print_glove_cfg(buf + len, PAGE_SIZE - len, "current", &cfg);

	return len;
}

static ssize_t mxt_hover_store(struct device *dev,
			       struct device_attribute *attr,
			       const char *buf, size_t count)
{
	struct mxt_data *data = dev_get_drvdata(dev);
	int cmd_enable = 0;

	ATMEL_TRACE();

	if (!strncmp(buf, "0", 1)) {
		cmd_enable = 0;
	} else {
		cmd_enable = 1;
	}

	mxt_set_hover(data, cmd_enable);

	mxt_update_t7_config(data);

	return count;
}

static ssize_t mxt_hover_show(struct device *dev,
			      struct device_attribute *attr, char *buf)
{
	struct mxt_data *data = dev_get_drvdata(dev);
	struct mxt_object *object;
	u8 control = 0xFF;
	int error;

	object = mxt_get_object(data, MXT_SPT_TOUCHSCREENHOVER_T101);
	if (object == NULL) {
		ATMEL_ERR("T%d object not found\n", MXT_SPT_TOUCHSCREENHOVER_T101);
		return 0;
	}

	error = __mxt_read_reg(data->client, object->start_address,
			       sizeof(control), &control);
	if (error != 0) {
		ATMEL_ERR("T%d CTRL read failed\n", object->type);
		return 0;
	}

	return scnprintf(buf, PAGE_SIZE, "0x%02X\n", control);
}

static ssize_t mxt_passive_stylus_cfg_store(struct device *dev,
					    struct device_attribute *attr,
					    const char *buf, size_t count)
{
	unsigned int data[5];
	int len;
	struct passive_stylus_config_t *cfg = NULL;

	ATMEL_TRACE();

	if (buf == NULL) {
		goto passive_stylus_cfg_error;
	}

	memset(data, 0, sizeof(data));
	len = sscanf(buf, "%x %x %x %x %x", &data[0], &data[1], &data[2], &data[3], &data[4]);
	if (len != ARRAY_SIZE(data)) {
		ATMEL_ERR("input data is invalid\n");
		goto passive_stylus_cfg_error;
	}

	switch (data[0]) {
	case 0:
		cfg = &touch_dev_cfg.passive_stylus_off_cfg;
		break;
	case 1:
		cfg = &touch_dev_cfg.passive_stylus_on_cfg;
		break;
	default:
		ATMEL_ERR("id is invalid\n");
		cfg = NULL;
		break;
	}

	if (cfg != NULL) {
		cfg->enable		= (u8)data[1];
		cfg->touch_debounce	= (u8)data[2];
		cfg->release_debounce	= (u8)data[3];
		cfg->smooth_level	= (u8)data[4];
	}

passive_stylus_cfg_error:
	return count;
}

static int mxt_print_passive_stylus_cfg(char *buf, int buf_size,
					const char *tag,
					struct passive_stylus_config_t *cfg)
{
	int len = 0;

	ATMEL_TRACE();

	if ((buf != NULL) && (cfg != NULL)) {
		len += scnprintf(buf + len, buf_size - len, "[%s]\n", tag);
		len += scnprintf(buf + len, buf_size - len,
				 "- enable           : 0x%02X\n", cfg->enable);
		len += scnprintf(buf + len, buf_size - len,
				 "- touch_debounce   : 0x%02X\n", cfg->touch_debounce);
		len += scnprintf(buf + len, buf_size - len,
				 "- release_debounce : 0x%02X\n", cfg->release_debounce);
		len += scnprintf(buf + len, buf_size - len,
				 "- smooth_level     : 0x%02X\n", cfg->smooth_level);
	}

	return len;
}

static ssize_t mxt_passive_stylus_cfg_show(struct device *dev,
					   struct device_attribute *attr,
					   char *buf)
{
	int len = 0;

	ATMEL_TRACE();

	if (buf != NULL) {
		len += mxt_print_passive_stylus_cfg(buf + len, PAGE_SIZE - len,
						    "off",
						    &touch_dev_cfg.passive_stylus_off_cfg);
		len += mxt_print_passive_stylus_cfg(buf + len, PAGE_SIZE - len,
						    "on",
						    &touch_dev_cfg.passive_stylus_on_cfg);
	}

	return len;
}

static ssize_t mxt_passive_stylus_store(struct device *dev,
					struct device_attribute *attr,
					const char *buf, size_t count)
{
	struct mxt_data *data = dev_get_drvdata(dev);
	int cmd_enable = 0;
	int error;

	ATMEL_TRACE();

	if (!strncmp(buf, "0", 1)) {
		cmd_enable = 0;
	} else {
		cmd_enable = 1;
	}

	error = mxt_set_passive_stylus(data, cmd_enable);
	if (error != 0) {
		goto passive_stylus_error;
	}

	mxt_update_t7_config(data);

passive_stylus_error:
	return count;
}

static ssize_t mxt_passive_stylus_show(struct device *dev,
			      struct device_attribute *attr, char *buf)
{
	struct mxt_data *data = dev_get_drvdata(dev);
	struct mxt_object *object;
	struct passive_stylus_config_t cfg;
	u16 addr = 0x00;
	int error;
	int len = 0;

	memset(&cfg, 0, sizeof(cfg));

	object = mxt_get_object(data, MXT_PROCI_STYLUS_T47);
	if (object == NULL) {
		ATMEL_ERR("T%d object not found\n", MXT_PROCI_STYLUS_T47);
		return 0;
	}

	error = __mxt_read_reg(data->client, object->start_address,
			       sizeof(cfg.enable), &cfg.enable);
	if (error != 0) {
		ATMEL_ERR("T%d CTRL read failed\n", object->type);
		return 0;
	}

	object = mxt_get_object(data, MXT_TOUCH_MULTITOUCHSCREEN_T100);
	if (object == NULL) {
		ATMEL_ERR("T%d object not found\n", MXT_TOUCH_MULTITOUCHSCREEN_T100);
		return 0;
	}

	addr = object->start_address + 39;
	error = __mxt_read_reg(data->client, addr,
			       sizeof(cfg.touch_debounce), &cfg.touch_debounce);
	if (error != 0) {
		ATMEL_ERR("T%d touch debounce read failed\n", object->type);
		return 0;
	}
	cfg.touch_debounce &= 0x3F;

	addr = object->start_address + 40;
	error = __mxt_read_reg(data->client, addr,
			       sizeof(cfg.release_debounce), &cfg.release_debounce);
	if (error != 0) {
		ATMEL_ERR("T%d release debounce read failed\n", object->type);
		return 0;
	}
	cfg.release_debounce &= 0x3F;

	addr = object->start_address + 45;
	error = __mxt_read_reg(data->client, addr,
			       sizeof(cfg.smooth_level), &cfg.smooth_level);
	if (error != 0) {
		ATMEL_ERR("T%d smooth level read failed\n", object->type);
		return 0;
	}

	len += mxt_print_passive_stylus_cfg(buf + len, PAGE_SIZE - len, "current", &cfg);

	return len;
}

static ssize_t mxt_active_stylus_cfg_store(struct device *dev,
					   struct device_attribute *attr,
					   const char *buf, size_t count)
{
	unsigned int data[3];
	int len;
	struct active_stylus_config_t *cfg = NULL;

	ATMEL_TRACE();

	if (buf == NULL) {
		goto active_stylus_cfg_error;
	}

	memset(data, 0, sizeof(data));
	len = sscanf(buf, "%x %x %x", &data[0], &data[1], &data[2]);
	if (len != ARRAY_SIZE(data)) {
		ATMEL_ERR("input data is invalid\n");
		goto active_stylus_cfg_error;
	}

	switch (data[0]) {
	case 0:
		cfg = &touch_dev_cfg.active_stylus_off_cfg;
		break;
	case 1:
		cfg = &touch_dev_cfg.active_stylus_on_cfg_normal;
		break;
	case 2:
		cfg = &touch_dev_cfg.active_stylus_on_cfg_light;
		break;
	default:
		ATMEL_ERR("id is invalid\n");
		cfg = NULL;
		break;
	}

	if (cfg != NULL) {
		cfg->enable		= (u8)data[1];
		cfg->detect_mode	= (u8)data[2];
	}

active_stylus_cfg_error:
	return count;
}

static int mxt_print_active_stylus_cfg(char *buf, int buf_size,
				       const char *tag,
				       struct active_stylus_config_t *cfg)
{
	int len = 0;

	ATMEL_TRACE();

	if ((buf != NULL) && (cfg != NULL)) {
		len += scnprintf(buf + len, buf_size - len, "[%s]\n", tag);
		len += scnprintf(buf + len, buf_size - len,
				 "- enable      : 0x%02X\n", cfg->enable);
		len += scnprintf(buf + len, buf_size - len,
				 "- detect_mode : 0x%02X\n", cfg->detect_mode);
	}

	return len;
}

static ssize_t mxt_active_stylus_cfg_show(struct device *dev,
					  struct device_attribute *attr,
					  char *buf)
{
	int len = 0;

	ATMEL_TRACE();

	if (buf != NULL) {
		len += mxt_print_active_stylus_cfg(buf + len, PAGE_SIZE - len,
						   "off",
						   &touch_dev_cfg.active_stylus_off_cfg);
		len += mxt_print_active_stylus_cfg(buf + len, PAGE_SIZE - len,
						   "normal",
						   &touch_dev_cfg.active_stylus_on_cfg_normal);
		len += mxt_print_active_stylus_cfg(buf + len, PAGE_SIZE - len,
						   "light",
						   &touch_dev_cfg.active_stylus_on_cfg_light);
	}

	return len;
}

static ssize_t mxt_active_stylus_store(struct device *dev,
				       struct device_attribute *attr,
				       const char *buf, size_t count)
{
	struct mxt_data *data = dev_get_drvdata(dev);
	int cmd_enable = 0;
	int error;

	ATMEL_TRACE();

	if (!strncmp(buf, "0", 1)) {
		cmd_enable = 0;
	} else {
		cmd_enable = 1;
	}

	error = mxt_set_active_stylus(data, cmd_enable);
	if (error != 0) {
		goto active_stylus_error;
	}

	mxt_update_t7_config(data);

active_stylus_error:
	return count;
}

static ssize_t mxt_active_stylus_show(struct device *dev,
				      struct device_attribute *attr,
				      char *buf)
{
	struct mxt_data *data = dev_get_drvdata(dev);
	struct mxt_object *object;
	u8 measallow = 0xFF;
	u8 detect_mode = 0xFF;
	int len = 0;
	int error;

	object = mxt_get_object(data, MXT_GEN_ACQUIRE_T8);
	if (object == NULL) {
		ATMEL_ERR("T%d object not found\n", MXT_GEN_ACQUIRE_T8);
		return 0;
	}

	error = __mxt_read_reg(data->client, object->start_address + 10,
			       sizeof(measallow), &measallow);
	if (error != 0) {
		ATMEL_ERR("T%d MEASALLOW read failed\n", object->type);
		return 0;
	}

	object = mxt_get_object(data, MXT_PROCI_ACTIVESTYLUS_T107);
	if (object == NULL) {
		ATMEL_ERR("T%d object not found\n", MXT_PROCI_ACTIVESTYLUS_T107);
		return 0;
	}

	error = __mxt_read_reg(data->client, object->start_address + 1,
			       sizeof(detect_mode), &detect_mode);
	if (error != 0) {
		ATMEL_ERR("T%d CFG1 read failed\n", object->type);
		return 0;
	}

	len += scnprintf(buf + len, PAGE_SIZE - len, "- measallow   = 0x%02X\n", measallow);
	len += scnprintf(buf + len, PAGE_SIZE - len, "- detect_mode = 0x%02X\n", detect_mode);

	return len;
}

static ssize_t mxt_waterproof_cfg_store(struct device *dev,
					struct device_attribute *attr,
					const char *buf, size_t count)
{
	unsigned int data[5];
	int len;
	struct waterproof_config_t *cfg = NULL;

	ATMEL_TRACE();

	if (buf == NULL) {
		goto waterproof_cfg_error;
	}

	memset(data, 0, sizeof(data));
	len = sscanf(buf, "%x %x %x %x %x",
		     &data[0], &data[1], &data[2], &data[3], &data[4]);
	if (len != ARRAY_SIZE(data)) {
		ATMEL_ERR("input data is invalid\n");
		goto waterproof_cfg_error;
	}

	switch (data[0]) {
	case 0:
		cfg = &touch_dev_cfg.waterproof_off_cfg;
		break;
	case 1:
		cfg = &touch_dev_cfg.waterproof_on_cfg_light;
		break;
	case 2:
		cfg = &touch_dev_cfg.waterproof_on_cfg_heavy;
		break;
	default:
		ATMEL_ERR("id is invalid\n");
		cfg = NULL;
		break;
	}

	if (cfg != NULL) {
		cfg->report_mode	= (u8)data[1];
		cfg->touch_gain		= (u8)data[2];
		cfg->delta_touch_thr	= (u8)data[3];
		cfg->anti_moisture	= (u8)data[4];
	}

waterproof_cfg_error:
	return count;
}

static int mxt_print_waterproof_cfg(char *buf, int buf_size, const char *tag,
				    struct waterproof_config_t *cfg)
{
	int len = 0;

	ATMEL_TRACE();

	if ((buf != NULL) && (cfg != NULL)) {
		len += scnprintf(buf + len, buf_size - len, "[%s]\n", tag);
		len += scnprintf(buf + len, buf_size - len,
				 "- report_mode     : 0x%02X\n", cfg->report_mode);
		len += scnprintf(buf + len, buf_size - len,
				 "- touch_gain      : 0x%02X\n", cfg->touch_gain);
		len += scnprintf(buf + len, buf_size - len,
				 "- delta_touch_thr : 0x%02X\n", cfg->delta_touch_thr);
		len += scnprintf(buf + len, buf_size - len,
				 "- anti_moisture   : 0x%02X\n", cfg->anti_moisture);
	}

	return len;
}

static ssize_t mxt_waterproof_cfg_show(struct device *dev,
				       struct device_attribute *attr,
				       char *buf)
{
	int len = 0;

	ATMEL_TRACE();

	if (buf != NULL) {
		len += mxt_print_waterproof_cfg(buf + len, PAGE_SIZE - len,
						"off",
						&touch_dev_cfg.waterproof_off_cfg);
		len += mxt_print_waterproof_cfg(buf + len, PAGE_SIZE - len,
						"condition-light",
						&touch_dev_cfg.waterproof_on_cfg_light);
		len += mxt_print_waterproof_cfg(buf + len, PAGE_SIZE - len,
						"condition-heavy",
						&touch_dev_cfg.waterproof_on_cfg_heavy);
	}

	return len;
}

static ssize_t mxt_waterproof_store(struct device *dev,
				    struct device_attribute *attr,
				    const char *buf, size_t count)
{
	struct mxt_data *data = dev_get_drvdata(dev);
	int cmd_enable = 0;
	int error;

	ATMEL_TRACE();

	if (!strncmp(buf, "0", 1)) {
		cmd_enable = 0;
	} else {
		cmd_enable = 1;
	}

	error = mxt_set_waterproof(data, cmd_enable);
	if (error != 0) {
		goto waterproof_error;
	}

waterproof_error:
	return count;
}

static ssize_t mxt_waterproof_show(struct device *dev,
				   struct device_attribute *attr,
				   char *buf)
{
	struct mxt_data *data = dev_get_drvdata(dev);
	struct mxt_object *object;
	int len = 0;
	struct waterproof_config_t cfg;
	int error;

	memset(&cfg, 0, sizeof(cfg));

	object = mxt_get_object(data, MXT_TOUCH_MULTITOUCHSCREEN_T100);
	if (object == NULL) {
		ATMEL_ERR("T%d object not found\n", MXT_TOUCH_MULTITOUCHSCREEN_T100);
		return 0;
	}

	error = __mxt_read_reg(data->client, object->start_address + MXT_T100_CFG1,
			       sizeof(cfg.report_mode), &cfg.report_mode);
	if (error != 0) {
		ATMEL_ERR("T%d CTRL read failed\n", object->type);
		return 0;
	}
	cfg.report_mode &= 0x01;

	error = __mxt_read_reg(data->client, object->start_address + MXT_T100_GAIN,
			       sizeof(cfg.touch_gain), &cfg.touch_gain);
	if (error != 0) {
		ATMEL_ERR("T%d touch_gain read failed\n", object->type);
		return 0;
	}

	error = __mxt_read_reg(data->client, object->start_address + MXT_T100_INTTHR,
			       sizeof(cfg.delta_touch_thr), &cfg.delta_touch_thr);
	if (error != 0) {
		ATMEL_ERR("T%d delta_touch_thr read failed\n", object->type);
		return 0;
	}

	object = mxt_get_object(data, MXT_PROCI_RETRANSMISSIONCOMPENSATION_T80);
	if (object == NULL) {
		ATMEL_ERR("T%d object not found\n", MXT_PROCI_RETRANSMISSIONCOMPENSATION_T80);
		return 0;
	}

	error = __mxt_read_reg(data->client, object->start_address,
			       sizeof(cfg.anti_moisture), &cfg.anti_moisture);
	if (error != 0) {
		ATMEL_ERR("T%d anti_moisture read failed\n", object->type);
		return 0;
	}

	len += mxt_print_waterproof_cfg(buf + len, PAGE_SIZE - len, "current", &cfg);

	return len;
}

static ssize_t mxt_calib_cfg_store(struct device *dev,
				   struct device_attribute *attr,
				   const char *buf, size_t count)
{
	unsigned int data[1];
	int len;
	struct calib_config_t *cfg = &touch_dev_cfg.calib_cfg;

	ATMEL_TRACE();

	if (buf == NULL) {
		goto calib_cfg_error;
	}

	memset(data, 0, sizeof(data));
	len = sscanf(buf, "%x", &data[0]);
	if (len != ARRAY_SIZE(data)) {
		ATMEL_ERR("input data is invalid\n");
		goto calib_cfg_error;
	}

	if (cfg != NULL) {
		cfg->force_calib_cnt = (((int)data[0] * 1000) / MXT_HEALTHCHK_INTERVAL);
	}

calib_cfg_error:
	return count;
}

static int mxt_print_calib_cfg(char *buf, int buf_size, const char *tag,
			       struct calib_config_t *cfg)
{
	int len = 0;

	ATMEL_TRACE();

	if ((buf != NULL) && (cfg != NULL)) {
		len += scnprintf(buf + len, buf_size - len, "[%s]\n", tag);
		len += scnprintf(buf + len, buf_size - len,
				 "- force calib time: %d[sec]\n",
				 (cfg->force_calib_cnt * MXT_HEALTHCHK_INTERVAL) / 1000);
	}

	return len;
}

static ssize_t mxt_calib_cfg_show(struct device *dev,
				  struct device_attribute *attr,
				  char *buf)
{
	int len = 0;

	ATMEL_TRACE();

	if (buf != NULL) {
		len += mxt_print_calib_cfg(buf + len, PAGE_SIZE - len,
					   "calib",
					   &touch_dev_cfg.calib_cfg);
	}

	return len;
}


#define MXT_NUM_LINE_X				(25)
#define MXT_NUM_LINE_Y				(14)
#define MXT_T37_CMD_NONE			(0x00)
#define MXT_T37_CMD_PAGE_UP			(0x01)
#define MXT_T37_CMD_PAGE_DOWN			(0x02)
#define MXT_T37_CMD_MUTUAL_CAP_DELTA		(0x10)
#define MXT_T37_CMD_MUTUAL_CAP_REF		(0x11)
#define MXT_T37_CMD_CALIB_RECOVERY_TUNING_DATA	(0x33)
#define MXT_T37_CMD_TS				(0xF4)
#define MXT_T37_CMD_SELF_CAP_DELTA		(0xF7)
#define MXT_T37_CMD_SELF_CAP_REF		(0xF8)
#define MXT_T37_CMD_ACT_STYLUS_DELTA		(0xFB)
#define MXT_T37_CMD_ACT_STYLUS_REF		(0xFC)
#define MXT_T37_ADDR_MODE			(0)
#define MXT_T37_ADDR_PAGE			(1)
#define MXT_T37_ADDR_DATA			(2)
#define MXT_T37_PAGE_SIZE			(128) /* byte */
#define MXT_T37_AD_DATA_UNIT			(2)   /* byte */
#define MXT_MAX_CAP_LINE_Y			(72)
#define MAX_T37_AD_NODE_PER_PAGE		(MXT_T37_PAGE_SIZE/MXT_T37_AD_DATA_UNIT)
#define MXT_T37_AD_NODE_DELTA			(MXT_MAX_CAP_LINE_Y - MAX_T37_AD_NODE_PER_PAGE)
#define MXT_T37_MAX_MUTUAL_AD_PAGES		(32)

struct mxt_t37_data_map_t {
	int offset; /* unit: 2byte */
	int size;   /* unit: 2byte */
	int has_start;
	int has_end;
};

static int t37_mutual_map_size;
static struct mxt_t37_data_map_t *t37_mutual_map;
static struct mxt_t37_data_map_t t37_self_map[] = {
	{ 0, MXT_NUM_LINE_Y, 1, 1}, /* touch y */
	{ 8, MXT_NUM_LINE_X, 1, 1}, /* touch x */
	{ 0,              0, 0, 0},
	{ 0, MXT_NUM_LINE_Y, 1, 1}, /* hover y */
	{ 8, MXT_NUM_LINE_X, 1, 1}, /* hover x */
	{ 0,              0, 0, 0},
	{ 0, MXT_NUM_LINE_Y, 1, 1}, /* single y */
};
static struct mxt_t37_data_map_t t37_active_stylus_map[] = {
	{ 0, MXT_NUM_LINE_Y, 1, 1}, /* y scan0 */
	{ 8, MXT_NUM_LINE_Y, 1, 1}, /* y scan1 */
	{16, MXT_NUM_LINE_X, 1, 1}, /* x scan0 */
	{24, MXT_NUM_LINE_X, 1, 1}, /* x scan1 */
};

struct mxt_t37_cmd_t {
	u8 cmd;
	u8 page_data[MXT_T37_PAGE_SIZE];
	struct mxt_t37_data_map_t *map;
	int num_line_x;
	int num_line_y;
	int num_page;
	int num_node;
	int num_size_per_page;
	int current_page;
	int current_node;
	int seq_cnt;
};
static struct mxt_t37_cmd_t t37_cmd = {
	.cmd = MXT_T37_CMD_NONE,
	.num_line_x = MXT_NUM_LINE_Y,
	.num_line_y = MXT_NUM_LINE_Y,
	.num_page = 0,
	.num_node = 0,
	.num_size_per_page = 0,
	.current_page = 0,
	.current_node = 0,
	.seq_cnt = 0,
};

static void mxt_create_t37_data_map(struct mxt_data *data)
{
	struct mxt_object *object;
	struct i2c_client *client = data->client;
	u8 num_line_x = 25;
	u8 num_line_y = 14;
	size_t map_size;
	int max_nodes;
	int offset;
	int i;
	int ret;

	object = mxt_get_object(data, MXT_TOUCH_MULTITOUCHSCREEN_T100);
	if (!object) {
		goto err_exit;
	}
	ret = __mxt_read_reg(client, object->start_address + MXT_T100_XSIZE,
			     sizeof(num_line_x), &num_line_x);
	if (ret != 0) {
		goto err_exit;
	}
	ret = __mxt_read_reg(client, object->start_address + MXT_T100_YSIZE,
			     sizeof(num_line_y), &num_line_y);
	if (ret != 0) {
		goto err_exit;
	}

	t37_cmd.num_line_x = num_line_x;
	t37_cmd.num_line_y = num_line_y;

	ATMEL_MSG("num_line_x=%d, num_line_y=%d\n",
		  t37_cmd.num_line_x, t37_cmd.num_line_y);

	/* create map for mutual */
	map_size = sizeof(struct mxt_t37_data_map_t) * MXT_T37_MAX_MUTUAL_AD_PAGES;
	t37_mutual_map = kzalloc(map_size, GFP_KERNEL);
	if (t37_mutual_map == NULL) {
		goto err_exit;
	}

	/* NOTE: this logic does not supports for 
	   large y-line number (lines > nodes per page) */
	memset(t37_mutual_map, 0, map_size);
	max_nodes = MXT_MAX_CAP_LINE_Y * t37_cmd.num_line_x;
	t37_mutual_map_size = max_nodes / MAX_T37_AD_NODE_PER_PAGE;
	offset = 0;
	for (i = 0; i < t37_mutual_map_size; i++) {
		t37_mutual_map[i].offset = offset;
		t37_mutual_map[i].size = t37_cmd.num_line_y;
		t37_mutual_map[i].has_start = 1;
		t37_mutual_map[i].has_end = 1;

		if ((offset + num_line_y) > MAX_T37_AD_NODE_PER_PAGE) {
			int remain = offset + t37_cmd.num_line_y - MAX_T37_AD_NODE_PER_PAGE;
			t37_mutual_map[i].size -= remain;
			t37_mutual_map[i].has_end = 0;

			i++;
			t37_mutual_map[i].offset = 0;
			t37_mutual_map[i].size = remain;
			t37_mutual_map[i].has_start = 0;
			t37_mutual_map[i].has_end = 1;
		}

		offset = ((offset + MXT_T37_AD_NODE_DELTA) % MAX_T37_AD_NODE_PER_PAGE);
	}

	/* create map for self */
	t37_self_map[0].size = t37_cmd.num_line_y;
	t37_self_map[1].size = t37_cmd.num_line_x;
	t37_self_map[3].size = t37_cmd.num_line_y;
	t37_self_map[4].size = t37_cmd.num_line_x;
	t37_self_map[6].size = t37_cmd.num_line_y;

	/* create map for stylus */
	t37_active_stylus_map[0].size = t37_cmd.num_line_y;
	t37_active_stylus_map[1].size = t37_cmd.num_line_y;
	t37_active_stylus_map[2].size = t37_cmd.num_line_x;
	t37_active_stylus_map[3].size = t37_cmd.num_line_x;

err_exit:
	return;
}

static ssize_t mxt_diag_store(struct device *dev,
			      struct device_attribute *attr,
			      const char *buf, size_t count)
{
	struct mxt_data *data = dev_get_drvdata(dev);
	struct mxt_object *object;

	ATMEL_TRACE();

	if (t37_mutual_map == NULL) {
		mxt_create_t37_data_map(data);
	}

	object = mxt_get_object(data, MXT_DEBUG_DIAGNOSTIC_T37);
	if (!object)
		goto err_exit;

	t37_cmd.cmd = (u8)simple_strtol(buf, NULL, 16);
	t37_cmd.seq_cnt = 0;

	ATMEL_MSG("input cmd=0x%02X\n", t37_cmd.cmd);

	switch (t37_cmd.cmd) {
	case MXT_T37_CMD_MUTUAL_CAP_DELTA:
	case MXT_T37_CMD_MUTUAL_CAP_REF:
		t37_cmd.map = t37_mutual_map;
		t37_cmd.num_node = t37_cmd.num_line_x * t37_cmd.num_line_y;
		t37_cmd.num_page = t37_mutual_map_size;
		break;
	case MXT_T37_CMD_SELF_CAP_DELTA:
	case MXT_T37_CMD_SELF_CAP_REF:
		t37_cmd.map = t37_self_map;
		t37_cmd.num_node = t37_cmd.num_line_y + t37_cmd.num_line_x;
		t37_cmd.num_page = sizeof(t37_self_map) / sizeof(t37_self_map[0]);
		break;
	case MXT_T37_CMD_ACT_STYLUS_DELTA:
	case MXT_T37_CMD_ACT_STYLUS_REF:
		t37_cmd.map = t37_active_stylus_map;
		t37_cmd.num_node = t37_cmd.num_line_y + t37_cmd.num_line_x;
		t37_cmd.num_page = sizeof(t37_active_stylus_map) / sizeof(t37_active_stylus_map[0]);
		break;
	case MXT_T37_CMD_CALIB_RECOVERY_TUNING_DATA:
	case MXT_T37_CMD_TS:
	default:
		t37_cmd.cmd = MXT_T37_CMD_NONE;
		if (t37_mutual_map != NULL) {
			kfree(t37_mutual_map);
			t37_mutual_map = NULL;
		}
		break;
	}

err_exit:
	return count;
}

static ssize_t mxt_diag_show(struct device *dev,
			     struct device_attribute *attr, char *buf)
{
	struct mxt_data *data = dev_get_drvdata(dev);
	struct mxt_object *object;
	int error;
	int i;
	ssize_t len = 0;

	if (t37_cmd.cmd == MXT_T37_CMD_NONE) {
		goto err_exit;
	}

	if (t37_cmd.map == NULL) {
		goto err_exit;
	}

	object = mxt_get_object(data, MXT_DEBUG_DIAGNOSTIC_T37);
	if (object == NULL) {
		goto err_exit;
	}

	/* run diag command */
	error = mxt_t6_command(data, MXT_COMMAND_DIAGNOSTIC,
			       t37_cmd.cmd, true);
	if (error) {
		goto err_exit;
	}

	t37_cmd.current_node = 0;

	/* read and print result */
	for (i = 0; i < t37_cmd.num_page; i++) {
		int j;

		t37_cmd.current_page = i;

		t37_cmd.num_size_per_page = t37_cmd.map[t37_cmd.current_page].size * sizeof(short);
		if (t37_cmd.num_size_per_page > 0) {
			short val = 0;

			/* sequence number */
			if (t37_cmd.map[t37_cmd.current_page].has_start == 1) {
				len += scnprintf(buf + len, PAGE_SIZE, "%d,", t37_cmd.seq_cnt);
			}

			/* read page data */
			memset(t37_cmd.page_data, 0, sizeof(t37_cmd.num_size_per_page));
			error = __mxt_read_reg(data->client,
					       object->start_address + MXT_T37_ADDR_DATA
					       + t37_cmd.map[t37_cmd.current_page].offset * sizeof(short),
					       t37_cmd.num_size_per_page, t37_cmd.page_data);
			if (error) {
				goto err_exit;
			}

			/* parse data */
			for (j = 0; j < t37_cmd.num_size_per_page; j += 2) {
				/* get 1 node value */
				memcpy(&val, &t37_cmd.page_data[j], sizeof(val));
				if ((t37_cmd.cmd == MXT_T37_CMD_MUTUAL_CAP_DELTA)
				    || (t37_cmd.cmd == MXT_T37_CMD_MUTUAL_CAP_REF)) {
					len += scnprintf(buf + len, PAGE_SIZE, "%5d,", val);
				} else {
					len += scnprintf(buf + len, PAGE_SIZE, "%d,", val);
				}

				/* goto next node */
				t37_cmd.current_node++;
			}

			/* end of this line */
			if (t37_mutual_map[t37_cmd.current_page].has_end == 1) {
				len += scnprintf(buf + len, PAGE_SIZE, "\n");
			}
		}

		/* goto next page */
		error = mxt_t6_command_fast(data, MXT_COMMAND_DIAGNOSTIC,
					    MXT_T37_CMD_PAGE_UP, true);
		if (error) {
			goto err_exit;
		}
	}

	t37_cmd.seq_cnt++;

	return len;

err_exit:
	return -EINVAL;
}

static ssize_t mxt_power_cfg_store(struct device *dev,
				   struct device_attribute *attr,
				   const char *buf, size_t count)
{
	struct mxt_data *data = dev_get_drvdata(dev);
	uint8_t fin_idle, fin_active, fin_actv2idle;
	uint8_t pen_idle, pen_active, pen_actv2idle;
	uint8_t *values[] = {&fin_idle, &fin_active, &fin_actv2idle, 
			     &pen_idle, &pen_active, &pen_actv2idle};
	const int num_args = sizeof(values) / sizeof(values[0]);
	char tmp_buf[32];
	char *head;
	int i;

	ATMEL_TRACE();

	/* extract values */
	memset(tmp_buf, 0, sizeof(tmp_buf));
	memcpy(tmp_buf, buf, sizeof(tmp_buf) - 1);
	head = tmp_buf;
	for (i = 0; i < num_args; i++)
	{
		char *ptr = strchr(head, ' ');
		if (ptr != NULL) {
			*ptr = '\0';
		}
		*values[i] = (uint8_t)simple_strtol(head, NULL, 16);
		if (ptr == NULL) {
			/* parsing completed */
			break;
		}
		head = ptr + 1;
	}

	if (i == (num_args - 1)) {
		touch_dev_cfg.t7_finger_cfg.idle = fin_idle;
		touch_dev_cfg.t7_finger_cfg.active = fin_active;
		touch_dev_cfg.t7_finger_cfg.actv2idle = fin_actv2idle;
		touch_dev_cfg.t7_stylus_cfg.idle = pen_idle;
		touch_dev_cfg.t7_stylus_cfg.active = pen_active;
		touch_dev_cfg.t7_stylus_cfg.actv2idle = pen_actv2idle;
		mxt_update_t7_config(data);
	}

	return count;
}

static ssize_t mxt_power_cfg_show(struct device *dev,
				  struct device_attribute *attr, char *buf)
{
	struct mxt_data *data = dev_get_drvdata(dev);
	int len = 0;
	struct t7_config t7cfg;
	int ret;

	ATMEL_TRACE();

	memset(&t7cfg, 0, sizeof(t7cfg));

	ret = __mxt_read_reg(data->client, data->T7_address,
			     sizeof(t7cfg), &t7cfg);
	if (ret != 0) {
		return -EIO;
	}

	len += scnprintf(buf + len, PAGE_SIZE - len,
			 "finger: idle=0x%02X,active=0x%02X,act2idle=0x%02X\n",
			 touch_dev_cfg.t7_finger_cfg.idle,
			 touch_dev_cfg.t7_finger_cfg.active,
			 touch_dev_cfg.t7_finger_cfg.actv2idle);
	len += scnprintf(buf + len, PAGE_SIZE - len,
			 "stylus: idle=0x%02X,active=0x%02X,act2idle=0x%02X\n",
			 touch_dev_cfg.t7_stylus_cfg.idle,
			 touch_dev_cfg.t7_stylus_cfg.active,
			 touch_dev_cfg.t7_stylus_cfg.actv2idle);
	len += scnprintf(buf + len, PAGE_SIZE - len,
			 "current: idle=0x%02X,active=0x%02X,act2idle=0x%02X\n",
			 t7cfg.idle,
			 t7cfg.active,
			 t7cfg.actv2idle);

	return len;
}

static ssize_t mxt_dev_cfg_store(struct device *dev,
				 struct device_attribute *attr,
				 const char *buf, size_t count)
{
	char tmp_buf[32];
	char *head, *ptr;
	int cmd = -1;
	ssize_t ret = count;

	ATMEL_TRACE();

	/* extract values */
	memset(tmp_buf, 0, sizeof(tmp_buf));
	memcpy(tmp_buf, buf, sizeof(tmp_buf) - 1);
	head = tmp_buf;
	ptr = strchr(head, ' ');
	if (ptr != NULL) {
		ssize_t cmd_len = ptr - head + 1;
		ssize_t cmd_buf_len = count - cmd_len;
		const char *cmd_buf = buf + cmd_len;
		*ptr = '\0';
		cmd = (int)simple_strtol(head, NULL, 16);

		switch (cmd) {
		case TP_CFG_CMD_POWER:
			ret = mxt_power_cfg_store(dev, attr, cmd_buf, cmd_buf_len);
			break;
		case TP_CFG_CMD_GLOVE:
			ret = mxt_glove_cfg_store(dev, attr, cmd_buf, cmd_buf_len);
			break;
		case TP_CFG_CMD_WATERPROOF:
			ret = mxt_waterproof_cfg_store(dev, attr, cmd_buf, cmd_buf_len);
			break;
		case TP_CFG_CMD_ACTIVE_STYLUS:
			ret = mxt_active_stylus_cfg_store(dev, attr, cmd_buf, cmd_buf_len);
			break;
		case TP_CFG_CMD_PASSIVE_STYLUS:
			ret = mxt_passive_stylus_cfg_store(dev, attr, cmd_buf, cmd_buf_len);
			break;
		case TP_CFG_CMD_CALIB:
			ret = mxt_calib_cfg_store(dev, attr, cmd_buf, cmd_buf_len);
			break;
		default:
			ATMEL_ERR("unknown command %d\n", cmd);
			ret = -EINVAL;
			break;
		}
		if (ret > 0) {
			ret = count;
		}
	} else {
		ret = -EINVAL;;
	}

	return ret;
}

static ssize_t mxt_nfc_sening_store(struct device *dev,
			      struct device_attribute *attr,
			      const char *buf, size_t count)
{
	struct mxt_data *data = dev_get_drvdata(dev);
	char tmp_buf[32];
	char *head, *ptr;

	ATMEL_TRACE();

	/* extract values */
	memset(tmp_buf, 0, sizeof(tmp_buf));
	memcpy(tmp_buf, buf, sizeof(tmp_buf) - 1);
	head = tmp_buf;
	ptr = strchr(head, ' ');
	data->force_restart_time = (uint32_t)simple_strtol((ptr--)+1,NULL,10) *1000;
	data->restart_wait_time = (uint32_t)simple_strtol(head,&ptr,10);

	ATMEL_DBG("received restart_wait_time: %u force_restart_time: %u.\n",\
			data->restart_wait_time,data->force_restart_time);

	return count;
}

static ssize_t mxt_nfc_sening_show(struct device *dev,
			     struct device_attribute *attr, char *buf)
{
	struct mxt_data *data = dev_get_drvdata(dev);
	ATMEL_TRACE();
	return scnprintf(buf, PAGE_SIZE, "restart_wait_time: %u\nforce_restart_time: %u\n", \
			data->restart_wait_time,data->force_restart_time/1000);
}

static ssize_t mxt_tp_enable_store(struct device *dev,
			      struct device_attribute *attr,
			      const char *buf, size_t count)
{
	struct mxt_data *data = dev_get_drvdata(dev);
	ATMEL_TRACE();
	if (!strncmp(buf, "0", 1)) {
		atomic_set(&data->enable_tp,TP_DISABLE);
	}
	else
		atomic_set(&data->enable_tp,TP_ENABLE);

	return count;
}

static ssize_t mxt_tp_enable_show(struct device *dev,
			     struct device_attribute *attr, char *buf)
{
	struct mxt_data *data = dev_get_drvdata(dev);
	ATMEL_TRACE();
	return scnprintf(buf, PAGE_SIZE, "%d\n",atomic_read(&data->enable_tp));
}

static ssize_t mxt_tp_settings_store(struct device *dev,
				struct device_attribute *attr,
				const char *buf, size_t count)
{
	ATMEL_TRACE();
	if (!strncmp(buf, "0", 1)) {
		istpenable = 0;
	}
	else
		istpenable = 1;

	return count;
}

static ssize_t mxt_tp_settings_show(struct device *dev,
				struct device_attribute *attr, char *buf)
{
	ATMEL_TRACE();
	return scnprintf(buf, PAGE_SIZE, "%d\n",istpenable);
}

static ssize_t mxt_tp_mode_store(struct device *dev,
			      struct device_attribute *attr,
			      const char *buf, size_t count)
{
	struct mxt_data *data = dev_get_drvdata(dev);
	uint16_t tmp_mode = data->mxt_tp_mode;
	int error ;

	ATMEL_TRACE();

	error = sscanf(buf,"%hx",&(data->mxt_tp_mode));
	if(-1 == error)
		data->mxt_tp_mode = tmp_mode;
	else
	mxt_set_detect_mode(data);

	return count;
}

static ssize_t mxt_tp_mode_show(struct device *dev,
			     struct device_attribute *attr, char *buf)
{
	struct mxt_data *data = dev_get_drvdata(dev);
	ATMEL_TRACE();

	return scnprintf(buf, PAGE_SIZE, "0x%X\n",data->mxt_tp_mode);
}

static DEVICE_ULONG_ATTR(debug_flag, S_IWUSR|S_IWGRP|S_IRUSR|S_IRGRP,
			 atmel_debug_flag);
static DEVICE_ULONG_ATTR(healthcheck_disable, S_IWUSR|S_IWGRP|S_IRUSR|S_IRGRP,
			 atmel_healthcheck_force_disable);
static DEVICE_ATTR(fw_version, S_IRUGO, mxt_fw_version_show, NULL);
static DEVICE_ATTR(hw_version, S_IRUGO, mxt_hw_version_show, NULL);
static DEVICE_ATTR(buildid, S_IRUGO, mxt_buildid_show, NULL);
static DEVICE_ATTR(productinfo, S_IRUGO, mxt_productinfo_show, NULL);
static DEVICE_ATTR(object, S_IRUGO, mxt_object_show, NULL);
static DEVICE_ATTR(poweron, S_IRUGO | S_IWUSR | S_IWGRP,
		   mxt_poweron_show, mxt_poweron_store);
static DEVICE_ATTR(ic_ver, S_IRUGO, mxt_ic_ver_show, NULL);
static DEVICE_ATTR(tp_cfg, S_IRUGO, mxt_tp_cfg_show, NULL);
static DEVICE_ATTR(flashprog, S_IRUGO, mxt_flashprog_show, NULL);
static DEVICE_ATTR(update_fw, S_IWUSR, NULL, mxt_update_fw_store);
static DEVICE_ATTR(update_cfg, S_IWUSR, NULL, mxt_update_cfg_store);
static DEVICE_ATTR(doreflash, S_IWUSR, NULL, mxt_doreflash_store);
static DEVICE_ATTR(forcereflash, S_IWUSR, NULL, mxt_forcereflash_store);
static DEVICE_ATTR(hw_irqstat, S_IRUGO, mxt_hw_irqstat_show, NULL);
static DEVICE_ATTR(reset, S_IWUSR | S_IWGRP, NULL, mxt_reset_store);
static DEVICE_ATTR(drv_irq, S_IRUGO | S_IWUSR | S_IWGRP,
				mxt_drv_irq_show, mxt_drv_irq_store);
static DEVICE_ATTR(debug_v2_enable, S_IWUSR | S_IRUSR, NULL,
				mxt_debug_v2_enable_store);
static DEVICE_ATTR(debug_notify, S_IRUGO, mxt_debug_notify_show, NULL);
static DEVICE_ATTR(debug_enable, S_IWUSR | S_IRUSR, mxt_debug_enable_show,
				mxt_debug_enable_store);
static DEVICE_ATTR(tsi, S_IRUGO, mxt_ud_show, NULL);
static DEVICE_ATTR(tsp, S_IWUSR | S_IWGRP | S_IRUGO,
				mxt_tsp_show, mxt_tsp_store);
static DEVICE_ATTR(eval_noise, S_IWUSR | S_IWGRP | S_IRUGO,
		   mxt_eval_noise_show, mxt_eval_noise_store);
static DEVICE_ATTR(glove_cfg, S_IWUSR | S_IWGRP | S_IRUGO,
		   mxt_glove_cfg_show, mxt_glove_cfg_store);
static DEVICE_ATTR(glove, S_IWUSR | S_IWGRP | S_IRUGO,
		   mxt_glove_show, mxt_glove_store);
static DEVICE_ATTR(hover, S_IWUSR | S_IWGRP | S_IRUGO,
		   mxt_hover_show, mxt_hover_store);
static DEVICE_ATTR(passive_stylus_cfg, S_IWUSR | S_IWGRP | S_IRUGO,
		   mxt_passive_stylus_cfg_show, mxt_passive_stylus_cfg_store);
static DEVICE_ATTR(passive_stylus, S_IWUSR | S_IWGRP | S_IRUGO,
		   mxt_passive_stylus_show, mxt_passive_stylus_store);
static DEVICE_ATTR(active_stylus_cfg, S_IWUSR | S_IWGRP | S_IRUGO,
		   mxt_active_stylus_cfg_show, mxt_active_stylus_cfg_store);
static DEVICE_ATTR(active_stylus, S_IWUSR | S_IWGRP | S_IRUGO,
		   mxt_active_stylus_show, mxt_active_stylus_store);
static DEVICE_ATTR(waterproof_cfg, S_IWUSR | S_IWGRP | S_IRUGO,
		   mxt_waterproof_cfg_show, mxt_waterproof_cfg_store);
static DEVICE_ATTR(waterproof, S_IWUSR | S_IWGRP | S_IRUGO,
		   mxt_waterproof_show, mxt_waterproof_store);
static DEVICE_ATTR(calib_cfg, S_IWUSR | S_IWGRP | S_IRUGO,
		   mxt_calib_cfg_show, mxt_calib_cfg_store);
static DEVICE_ATTR(diag, S_IWUSR | S_IWGRP | S_IRUGO,
		   mxt_diag_show, mxt_diag_store);
static DEVICE_ATTR(power_cfg, S_IWUSR | S_IWGRP | S_IRUGO,
		   mxt_power_cfg_show, mxt_power_cfg_store);
static DEVICE_ATTR(dev_cfg, S_IWUSR | S_IWGRP, NULL, mxt_dev_cfg_store);
static DEVICE_ATTR(nfc_sensing, S_IWUSR | S_IWGRP | S_IRUGO,
		   mxt_nfc_sening_show, mxt_nfc_sening_store);
static DEVICE_ATTR(tp_enable, S_IWUSR | S_IWGRP | S_IRUGO,
		   mxt_tp_enable_show, mxt_tp_enable_store);
static DEVICE_ATTR(tp_mode, S_IWUSR | S_IWGRP | S_IRUGO,
		   mxt_tp_mode_show, mxt_tp_mode_store);

static DEVICE_ATTR(tp_settings, S_IWUSR | S_IWGRP | S_IRUGO,
		   mxt_tp_settings_show, mxt_tp_settings_store);


static struct attribute *mxt_attrs[] = {
	&dev_attr_debug_flag.attr.attr,
	&dev_attr_healthcheck_disable.attr.attr,
	&dev_attr_fw_version.attr,
	&dev_attr_hw_version.attr,
	&dev_attr_object.attr,
	&dev_attr_update_fw.attr,
	&dev_attr_update_cfg.attr,
	&dev_attr_ic_ver.attr,
	&dev_attr_tp_cfg.attr,
	&dev_attr_doreflash.attr,
	&dev_attr_flashprog.attr,
	&dev_attr_buildid.attr,
	&dev_attr_productinfo.attr,
	&dev_attr_poweron.attr,
	&dev_attr_forcereflash.attr,
	&dev_attr_drv_irq.attr,
	&dev_attr_hw_irqstat.attr,
	&dev_attr_reset.attr,
	&dev_attr_debug_enable.attr,
	&dev_attr_debug_v2_enable.attr,
	&dev_attr_debug_notify.attr,
	&dev_attr_tsi.attr,
	&dev_attr_tsp.attr,
	&dev_attr_eval_noise.attr,
	&dev_attr_glove_cfg.attr,
	&dev_attr_glove.attr,
	&dev_attr_hover.attr,
	&dev_attr_passive_stylus_cfg.attr,
	&dev_attr_passive_stylus.attr,
	&dev_attr_active_stylus_cfg.attr,
	&dev_attr_active_stylus.attr,
	&dev_attr_waterproof_cfg.attr,
	&dev_attr_waterproof.attr,
	&dev_attr_calib_cfg.attr,
	&dev_attr_diag.attr,
	&dev_attr_power_cfg.attr,
	&dev_attr_dev_cfg.attr,
	&dev_attr_nfc_sensing.attr,
	&dev_attr_tp_enable.attr,
	&dev_attr_tp_mode.attr,
	&dev_attr_tp_settings.attr,
	NULL
};

static const struct attribute_group mxt_attr_group = {
	.attrs = mxt_attrs,
};

static void mxt_reset_slots(struct mxt_data *data)
{
	struct input_dev *input_dev = data->input_dev;
	unsigned int num_mt_slots = TP_MAX_POS;
	int id;

	ATMEL_TRACE();

	if (data->buttons_enabled)
		data->t15_keystatus = 0;

	if (!input_dev)
		return;

	/* force activate all slots by untouched data
	   to synchronize slot state between kernel and frameworks */
	for (id = 0; id < num_mt_slots; id++) {
		input_mt_slot(input_dev, id);
		input_mt_report_slot_state(input_dev, MT_TOOL_FINGER, 1);
		input_report_abs(input_dev, ABS_MT_POSITION_X, 0);
		input_report_abs(input_dev, ABS_MT_POSITION_Y, 0);
		input_report_abs(input_dev, ABS_MT_PRESSURE, 0);
		input_report_abs(input_dev, ABS_MT_TOUCH_MAJOR, 0);
	}
	mxt_input_sync(input_dev, 0);

	/* reset all slots actually */
	for (id = 0; id < num_mt_slots; id++) {
		input_mt_slot(input_dev, id);
		input_mt_report_slot_state(input_dev, MT_TOOL_FINGER, 0);

		data->touch_state[id] = 0;
	}

	data->num_touched = 0;

	memset(&mCurrentEvent, 0, sizeof(mCurrentEvent));

	mxt_input_sync(input_dev, data->num_touched);
}

#ifndef CONFIG_FB
static int mxt_input_open(struct input_dev *dev)
{
	struct mxt_data *data = input_get_drvdata(dev);

	ATMEL_TRACE();

	mxt_irq_enable(data, true);

	if (data->use_regulator) {
		mxt_regulator_enable(data);
		(void)mxt_acquire_irq(data);
	} else if (!data->in_bootloader)
		(void)mxt_hw_reset(hw);

	mxt_unlock(&data->crit_section_lock);

	return 0;
}

static void mxt_input_close(struct input_dev *dev)
{
	struct mxt_data *data = input_get_drvdata(dev);

	ATMEL_TRACE();

	mxt_reset_slots(data);
	mxt_irq_enable(data, false);

	if (data->use_regulator)
		mxt_regulator_disable(data);

	mxt_lock(&data->crit_section_lock);
}
#endif

#ifdef CONFIG_OF
int mxt_dt_parse_state(struct mxt_data *data, struct device_node *np_config,
		struct mxt_patch *state)
{
	const char *patch_data;
	struct device_node *np_state;
	int err;

	ATMEL_TRACE();

	np_state = of_node_get(np_config);
	err = of_property_read_string(np_config, "patch-data",
				(const char **)&patch_data);
	if (err < 0) {
		pr_err("unable to read patch-data\n");
		return err;
	}

	pr_debug("processing state: %s\n", patch_data);
	mxt_parse_setup_string(data, patch_data, state);

	return 0;
}

int mxt_dt_parse_mode(struct mxt_data *data, const char *mode_name,
		struct mxt_patchset *mode)
{
	struct device *dev = &data->client->dev;
	struct device_node *np = dev->of_node;
	struct device_node *np_modes;
	int ret;
	char *propname;
	struct property *prop;
	const __be32 *list;
	int size, config;
	phandle phandle;
	struct device_node *np_config;

	ATMEL_TRACE();

	np_modes = of_find_node_by_name(np, "touchstate_modes");
	if (!np_modes) {
		pr_err("can't find touchstate modes node\n");
		ret = -EINVAL;
		goto err;
	}

	pr_debug("processing mode %s\n", mode_name);
	propname = kasprintf(GFP_KERNEL, "touchmode-%s", mode_name);
	prop = of_find_property(np_modes, propname, &size);
	kfree(propname);
	of_node_put(np_modes);
	if (!prop) {
		pr_err("can't find mode %s\n", mode_name);
		ret = -EINVAL;
		goto err;
	}
	list = prop->value;
	size /= sizeof(*list);

	if (size > MAX_NUM_STATES) {
		pr_err("unexpected number of states %d\n", size);
		ret = -EINVAL;
		goto err;
	}

	for (config = 0; config < size; config++) {
		phandle = be32_to_cpup(list++);

		/* Look up the touchstate configuration node */
		np_config = of_find_node_by_phandle(phandle);
		if (!np_config) {
			dev_err(dev,
				"prop %s index %i invalid phandle\n",
				prop->name, config);
			ret = -EINVAL;
			goto err;
		}

		/* Parse the node */
		ret = mxt_dt_parse_state(data, np_config,
				 &mode->patch_data[config]);
		of_node_put(np_config);
		if (ret < 0)
			goto err;
	}
err:
	return ret;
}

static int mxt_parse_dt(struct mxt_data *data)
{
	unsigned resolution[4], key_codes[MXT_MAX_BUTTONS];
	struct device *dev = &data->client->dev;
	struct mxt_platform_data *pdata = data->pdata;
	int error = 0;
	u32 value;
	struct device_node *config, *np = dev->of_node;

	ATMEL_TRACE();

	/*
	error = mxt_dt_parse_mode(data, "default", data->default_mode);
	if (error) {
		pr_err("failed to load default mode\n");
		goto exit_parser;
	}

	error = mxt_dt_parse_mode(data, "alternate", data->alternate_mode);
	if (error) {
		pr_warn("alternate mode not found; using default instead\n");
		data->alternate_mode = data->default_mode;
	}
	*/

	data->use_regulator = of_property_read_bool(np,
					"atmel,suspend-power-off");
	if (data->use_regulator)
		pr_info("using suspend method: power off\n");

	error = of_property_read_u32_array(np, "atmel,panel-coords", resolution, 4);
	if (!error) {
		pdata->res.x_min = resolution[0];
		pdata->res.y_min = resolution[1];
		pdata->res.x_max = resolution[2];
		pdata->res.y_max = resolution[3];
		pr_debug("panel coords X:%d-%d,Y:%d-%d\n",
					pdata->res.x_min, pdata->res.x_max,
					pdata->res.y_min, pdata->res.y_max);
	}

	/* pinctrl */
	pdata->ts_pinctrl = devm_pinctrl_get(dev);
	if (!IS_ERR_OR_NULL(pdata->ts_pinctrl)) {
		pdata->pinctrl_state_active
			= pinctrl_lookup_state(pdata->ts_pinctrl, PINCTRL_STATE_ACTIVE);
		pdata->pinctrl_state_suspend
			= pinctrl_lookup_state(pdata->ts_pinctrl, PINCTRL_STATE_SUSPEND);
	}
	ATMEL_DBG("ts_pinctrl=%p, active=%p, suspend=%p\n",
		  pdata->ts_pinctrl, pdata->pinctrl_state_active, pdata->pinctrl_state_suspend);

	/* gpio info */
	pdata->gpio_reset = of_get_named_gpio_flags(np, "atmel,reset-gpio",
						  0, &pdata->gpio_reset_flags);
	pdata->gpio_irq = of_get_named_gpio_flags(np, "atmel,irq-gpio",
						  0, &pdata->gpio_irq_flags);
	pdata->gpio_dcdc_en = of_get_named_gpio_flags(np, "atmel,vdd-dcdc_en-gpio",
						      0, &pdata->gpio_dcdc_en_flags);
	pdata->gpio_io_ldo_en = of_get_named_gpio_flags(np, "atmel,vdd-io-ldo_en-gpio",
							0, &pdata->gpio_io_ldo_en_flags);
	pdata->gpio_ldo_en = of_get_named_gpio_flags(np, "atmel,vdd-ldo_en-gpio",
						     0, &pdata->gpio_ldo_en_flags);
	ATMEL_DBG("GPIO(%s) valid=%d, gpio=%d\n", "reset", gpio_is_valid(pdata->gpio_reset), pdata->gpio_reset);
	ATMEL_DBG("GPIO(%s) valid=%d, gpio=%d\n", "irq", gpio_is_valid(pdata->gpio_irq), pdata->gpio_irq);
	ATMEL_DBG("GPIO(%s) valid=%d, gpio=%d\n", "dcdc_en", gpio_is_valid(pdata->gpio_dcdc_en), pdata->gpio_dcdc_en);
	ATMEL_DBG("GPIO(%s) valid=%d, gpio=%d\n", "io_ldo_en", gpio_is_valid(pdata->gpio_io_ldo_en), pdata->gpio_io_ldo_en);
	ATMEL_DBG("GPIO(%s) valid=%d, gpio=%d\n", "ldo_en", gpio_is_valid(pdata->gpio_ldo_en), pdata->gpio_ldo_en);

	/* defaults in case DTS is not populated */
	pdata->dt_info.family_id = 0xa4;
	pdata->dt_info.variant_id = 0x2;
	pdata->dt_info.version = 0x5;
	pdata->dt_info.build = 0x2;

	memset(key_codes, 0, sizeof(key_codes));
	error = of_property_read_u32_array(np, "atmel,key-buttons",
			 key_codes, MXT_MAX_BUTTONS);
	if (!error) {
		int i, keys;

		pdata->t15_num_keys = MXT_MAX_BUTTONS;
		pdata->t15_keymap = devm_kzalloc(dev,
			sizeof(unsigned) * pdata->t15_num_keys, GFP_KERNEL);
		if (!pdata->t15_keymap) {
			dev_err(dev, "T15 keymap allocation failure\n");
			goto exit_parser;
		}

		for (i = 0, keys = 0; i < pdata->t15_num_keys; i++)
			if (key_codes[i]) {
				*(pdata->t15_keymap + i) = key_codes[i];
				keys++;
			}

		data->buttons_enabled = true;
		pr_info("T15 has %d buttons\n", keys);
	}

	config = of_find_node_by_name(np, "atmel,cfg");
	if (!config) {
		dev_err(dev, "can't find default config node\n");
		error = -EINVAL;
		goto exit_parser;
	}

	error = of_property_read_u32(config, "atmel,family-id", &value);
	if (error) {
		dev_err(dev, "Unable to read family id\n");
		goto exit_parser;
	}
	pdata->dt_info.family_id = (u8)value;

	error = of_property_read_u32(config, "atmel,variant-id", &value);
	if (error) {
		dev_err(dev, "Unable to read variant id\n");
		goto exit_parser;
	}
	pdata->dt_info.variant_id = (u8)value;

	error = of_property_read_u32(config, "atmel,version", &value);
	if (error) {
		dev_err(dev, "Unable to read controller version\n");
		goto exit_parser;
	}
	pdata->dt_info.version = (u8)value;

	error = of_property_read_u32(config, "atmel,build", &value);
	if (error) {
		dev_err(dev, "Unable to read build id\n");
		goto exit_parser;
	}
	pdata->dt_info.build = (u8)value;

	error = of_property_read_u32(config, "atmel,revision-id", &value);
	if (error) {
		dev_err(dev, "Unable to read revision id\n");
		goto exit_parser;
	}
	data->revision_id = (u8)value;

exit_parser:
	if (!error)
		dev_dbg(dev, "DT parsed successfully\n");
	return 0;
}
#else
static int mxt_parse_dt(struct device *dev, struct mxt_platform_data *pdata)
{
	return -ENODEV;
}
#endif

static uint8_t* strnch( uint8_t *buf, char ch, uint32_t num)
{
	uint32_t n =0;
	uint8_t *pos = buf;
	if(NULL == buf) return NULL;
	while(*pos){
		if(*pos == ch) n++;
		if(num == n )return pos;
		pos++;
	}
	return NULL;
}

static int mxt_parse_register_address(struct mxt_data *data,uint8_t* buf)
{
	uint8_t obj,instance,byte;
	struct mxt_object *object;
	uint16_t address;
	uint8_t* pos;
	uint16_t num = data->tp_conf.conf_register_num;
	int i;
	int error;

	ATMEL_TRACE();
	data->tp_conf.address = (uint16_t*)kmalloc(num*sizeof(uint16_t),GFP_KERNEL);

	for(i = 0 ; i != num; i++)
	{
		pos = strnch(buf,',',i*3+1);
		if(NULL == pos) return 1;
		error = sscanf(pos+1,"%hhx,%hhx,%hhx",&obj,&instance,&byte);
		ATMEL_DBG("Read object = %X,instance = %X,byte = %X\n",obj,instance,byte);
		if(-1 == error) return 1;
		else
		{
			object = mxt_get_object(data, obj);
			if (object == NULL) {
				ATMEL_ERR("T%d object not found\n", obj);
				return 1;
			}
			address = object->start_address + mxt_obj_size(object)*instance+byte;
			data->tp_conf.address[i] = address;
		}
		ATMEL_DBG("Calculated address [%d] = %d .\n",i,data->tp_conf.address[i]);
	}

	return 0;
}

static int mxt_parse_mode_value(struct mxt_data *data,uint8_t *buf)
{
	struct mxt_tp_mode* tp_mode ;
	uint8_t *pos;
	uint16_t value_num = data->tp_conf.conf_register_num;
	int i ;
	int error;

	ATMEL_TRACE();

	tp_mode = (struct mxt_tp_mode *)kmalloc(sizeof(struct mxt_tp_mode),GFP_KERNEL);
	pos = strnch(buf,',',1);
	error = sscanf(pos+1,"%hx",&(tp_mode->mode_id));
	ATMEL_DBG("Read mode id = %X.\n",tp_mode->mode_id);

	for(i=0;  i != value_num; i++)
	{
		pos = strnch(buf,',',i+2);
		error = sscanf(pos+1 ,"%hhx",(tp_mode->value)+i);
	}

	for(i=0;i != value_num; i++)
	ATMEL_DBG("Value [%d] = %x.\n",i,tp_mode->value[i]);

	list_add_tail(&(tp_mode->list_node),&( data->tp_conf.mode_conf_list));

	return 0;
}

#define CONFIG_FILE_TYPEA "/vendor/firmware/touchdev_typeA.cfg"
#define CONFIG_FILE_TYPEB "/vendor/firmware/touchdev_typeB.cfg"
#define CONFIG_FILE_TYPEC "/vendor/firmware/touchdev_typeC.cfg"
#define TMP_BUF_SIZE 512
static void  mxt_parse_mode_conf(struct mxt_data *data)
{
	const int CFG_SIZE = 65535;
	struct file *filep = NULL;
	u8 *cfgbuf = NULL;
	int num_read = 0;
	int cfgsize;
	uint8_t *ptr = NULL;
	uint8_t *pos = NULL;
	uint8_t *tmp_buf = NULL;
	int error;
	int ret;
	u8 tp_cfg;
	uint8_t type;
	char panel_name[256] = {'\0'};
	char cfg_path[256] = {'\0'};

	ATMEL_TRACE();
	mdss_get_panel_name(panel_name);
	if (!strcmp(TYPEA,panel_name)) {
		strcpy(cfg_path, CONFIG_FILE_TYPEA);
		pr_info("cfg_path:%s.\n",cfg_path);
	}else if (!strcmp(TYPEB,panel_name)) {
		ret = __mxt_read_reg(data->client,
					data->T38_address + MXT_T38_VENDOR_VER,
					sizeof(tp_cfg), &tp_cfg);
		if (ret) {
			ATMEL_ERR("Unable to read config id\n");
		}
		if (tp_cfg != 0x0A){
			strcpy(cfg_path, CONFIG_FILE_TYPEB);
			pr_info("cfg_path:%s.\n",cfg_path);
		} else {
			strcpy(cfg_path, CONFIG_FILE_TYPEC);
			pr_info("cfg_path:%s.\n",cfg_path);
		}
	}else {
		ATMEL_ERR("Failed to compatible LCD panel.\n");
	}

	filep = filp_open(cfg_path, O_RDONLY|O_LARGEFILE, 0);
	if (IS_ERR(filep)){
		ATMEL_ERR("can't open %s",cfg_path);
		return ;
	}

	cfgsize = i_size_read(file_inode(filep));
	if ((cfgsize <= 0) || (cfgsize > CFG_SIZE)) {
		fput(filep);
		return;
	}

	cfgbuf = kzalloc(CFG_SIZE, GFP_KERNEL);
	if (cfgbuf == NULL) {
		fput(filep);
		return;
	}

	tmp_buf = (uint8_t *)kzalloc(sizeof(uint8_t)*TMP_BUF_SIZE,GFP_KERNEL);

	num_read = kernel_read(filep, 1, cfgbuf, cfgsize);
	fput(filep);
	filp_close(filep,NULL);

	pos = cfgbuf;

	INIT_LIST_HEAD(&(data->tp_conf.mode_conf_list));

	for(; *pos !='\0' ;)
	{
		memset(tmp_buf,0,sizeof(uint8_t)*TMP_BUF_SIZE);
		ptr = strstr(pos,"\r\n");
		memcpy(tmp_buf, pos, ptr-pos);
		error = sscanf(tmp_buf,"%hhx",&type);
		switch(type){
			case 0:
				error = sscanf(tmp_buf,"%hhx,%hx,%hx",&type,\
						&(data->tp_conf.conf_register_num),&(data->tp_conf.mode_num));
				ATMEL_DBG("Parse data/register number = %X mode_num =%X.\n",data->tp_conf.conf_register_num,data->tp_conf.mode_num);
				break;
			case 1:
				mxt_parse_register_address(data,tmp_buf);
				break;
			case 2:
				 mxt_parse_mode_value(data,tmp_buf);
				break;
			default:
				ATMEL_DBG("Error configure data type.\n");
		}
		pos = ptr +2 ;
	}

	return;
}

static int mxt_set_detect_mode(struct mxt_data *data)
{
	int i ;
	struct mxt_tp_mode *tp_mode_node;
	int state;

	ATMEL_TRACE();
	state = mxt_get_sensor_state(data);
	if(STATE_SUSPEND == state)
	{
		ATMEL_ERR("Device is on suspend state.\n");
		return 1;
	}

	ATMEL_DBG("Set detect mode = %X.\n",data->mxt_tp_mode);
	list_for_each_entry(tp_mode_node,&(data->tp_conf.mode_conf_list),list_node){
		if(tp_mode_node->mode_id == data->mxt_tp_mode)
		{
			for(i = 0 ; i != data->tp_conf.conf_register_num;i++){
				mxt_write_reg(data->client,
					      data->tp_conf.address[i],
					      tp_mode_node->value[i]);
			}
			return 0;
		}
	}
	return 1;
}

static int mxt_handle_pdata(struct mxt_data *data)
{
	int error;
	data->pdata = dev_get_platdata(&data->client->dev);

	ATMEL_TRACE();

	/* Use provided platform data if present */
	if (data->pdata) {
		if (data->pdata->cfg_name)
			mxt_update_file_name(&data->client->dev,
					     &data->cfg_name,
					     data->pdata->cfg_name,
					     strlen(data->pdata->cfg_name));
		return 0;
	}

	data->pdata = kzalloc(sizeof(*data->pdata), GFP_KERNEL);
	if (!data->pdata) {
		dev_err(&data->client->dev, "Failed to allocate pdata\n");
		return -ENOMEM;
	}

	/* Set default parameters */
	data->pdata->irqflags = IRQF_TRIGGER_LOW | IRQF_ONESHOT;

	if (data->client->dev.of_node) {
		error = mxt_parse_dt(data);
		if (error)
			return error;
	}

	return 0;
}

static int mxt_initialize_t9_input_device(struct mxt_data *data)
{
	struct device *dev = &data->client->dev;
	const struct mxt_platform_data *pdata = data->pdata;
	struct input_dev *input_dev;
	int error;
	unsigned int num_mt_slots;
	int i;

	ATMEL_TRACE();

	error = mxt_read_t9_resolution(data);
	if (error)
		dev_warn(dev, "Failed to initialize T9 resolution\n");

	input_dev = data->input_dev;
#ifdef MXT_DISTINCT_INPUT_NAME
	input_dev->name = "Atmel maXTouch T9 Touchscreen";
#endif
	if (pdata->t19_num_keys) {
		__set_bit(INPUT_PROP_BUTTONPAD, input_dev->propbit);

		for (i = 0; i < pdata->t19_num_keys; i++)
			if (pdata->t19_keymap[i] != KEY_RESERVED)
				input_set_capability(input_dev, EV_KEY,
						     pdata->t19_keymap[i]);

		input_abs_set_res(input_dev, ABS_X, MXT_PIXELS_PER_MM);
		input_abs_set_res(input_dev, ABS_Y, MXT_PIXELS_PER_MM);
		input_abs_set_res(input_dev, ABS_MT_POSITION_X,
				  MXT_PIXELS_PER_MM);
		input_abs_set_res(input_dev, ABS_MT_POSITION_Y,
				  MXT_PIXELS_PER_MM);
#ifdef MXT_DISTINCT_INPUT_NAME
		input_dev->name = "Atmel maXTouch Touchpad";
#endif
	} else
		set_bit(INPUT_PROP_DIRECT, input_dev->propbit);


	/* For single touch */
	input_set_abs_params(input_dev, ABS_X,
			     0, data->max_x, 0, 0);
	input_set_abs_params(input_dev, ABS_Y,
			     0, data->max_y, 0, 0);
	input_set_abs_params(input_dev, ABS_PRESSURE,
			     0, 255, 0, 0);

	/* For multi touch */
	num_mt_slots = data->num_touchids + data->num_stylusids;

#ifdef INPUT_MT_POINTER
	/* new version of input_mt */
	input_mt_init_slots(input_dev, num_mt_slots, 0);
#else
	/* old version of input_mt */
	input_mt_init_slots(input_dev, num_mt_slots);
#endif

	input_set_abs_params(input_dev, ABS_MT_TOUCH_MAJOR,
			     0, MXT_MAX_AREA, 0, 0);
	input_set_abs_params(input_dev, ABS_MT_POSITION_X,
			     0, data->max_x, 0, 0);
	input_set_abs_params(input_dev, ABS_MT_POSITION_Y,
			     0, data->max_y, 0, 0);
	input_set_abs_params(input_dev, ABS_MT_PRESSURE,
			     0, 255, 0, 0);
	input_set_abs_params(input_dev, ABS_MT_ORIENTATION,
			     0, 255, 0, 0);

	/* For T63 active stylus */
	if (data->T63_reportid_min) {
		input_set_capability(input_dev, EV_KEY, BTN_STYLUS);
		input_set_capability(input_dev, EV_KEY, BTN_STYLUS2);
		input_set_abs_params(input_dev, ABS_MT_TOOL_TYPE,
			0, MT_TOOL_MAX, 0, 0);
	}

	/* For T15 key array */
	if (data->T15_reportid_min) {
		data->t15_keystatus = 0;

		for (i = 0; i < data->pdata->t15_num_keys; i++)
			input_set_capability(input_dev, EV_KEY,
					     data->pdata->t15_keymap[i]);
	}

	return 0;
}

static int mxt_alloc_input_device(struct mxt_data *data)
{
	ATMEL_TRACE();

	data->input_dev = input_allocate_device();
	if (IS_ERR_OR_NULL(data->input_dev))
		return -ENOMEM;

#ifndef MXT_DISTINCT_INPUT_NAME
	data->input_dev->name = DRIVER_NAME;
#endif
	data->input_dev->id.bustype = BUS_I2C;
	data->input_dev->dev.parent = &data->client->dev;
#ifndef CONFIG_FB
	input_dev->open = mxt_input_open;
	input_dev->close = mxt_input_close;
#endif
	set_bit(EV_SYN, data->input_dev->evbit);
	set_bit(EV_ABS, data->input_dev->evbit);
	set_bit(EV_KEY, data->input_dev->evbit);
	set_bit(BTN_TOUCH, data->input_dev->keybit);

	input_set_drvdata(data->input_dev, data);

	return 0;
}

#if 0
static int mxt_init_mode(struct mxt_data *data, struct mxt_patchset **pmode)
{
	int i;
	struct mxt_patchset *mode = *pmode =
		kzalloc(sizeof(struct mxt_patchset), GFP_KERNEL);

	ATMEL_TRACE();

	if (!mode)
		return -ENOMEM;

	mode->patch_num = MAX_NUM_STATES;
	mode->patch_data = kzalloc(sizeof(struct mxt_patch) *
		mode->patch_num, GFP_KERNEL);
	if (!mode->patch_data) {
		kfree(mode);
		return -ENOMEM;
	}

	for (i = 0; i < mode->patch_num; i++)
		INIT_LIST_HEAD(&mode->patch_data[i].cfg_head);
	return 0;
}
#endif

static int mxt_probe(struct i2c_client *client,
		const struct i2c_device_id *id)
{
	struct mxt_data *data;
	int error;

	ATMEL_TRACE();

	data = kzalloc(sizeof(struct mxt_data), GFP_KERNEL);
	if (!data) {
		dev_err(&client->dev, "Failed to allocate memory\n");
		return -ENOMEM;
	}

	snprintf(data->phys, sizeof(data->phys), "i2c-%u-%04x/input0",
		 client->adapter->nr, client->addr);

	data->client = client;
	data->irq = client->irq;
	i2c_set_clientdata(client, data);

	/* mxt_init_mode(data, &data->default_mode); */
	/* mxt_init_mode(data, &data->alternate_mode); */

	error = mxt_handle_pdata(data);
	if (error)
		goto err_free_mem;

	data->current_mode = data->default_mode;

	init_completion(&data->bl_completion);
	init_completion(&data->reset_completion);
	init_completion(&data->crc_completion);
	init_completion(&data->selftest_completion);
	mutex_init(&data->debug_msg_lock);

	data->enable_fw_update = 0;
	mxt_create_task(data);

	mxt_set_sensor_state(data, STATE_INIT);

	atomic_set(&data->suspended, 0);
	data->poweron = true;

	error = mxt_gpio_configure(data);
	if (error)
		goto err_free_pdata;

	/* Handle power up in edge triggered IRQ mode */
	/* mxt_request_irq(data, IRQF_TRIGGER_FALLING | IRQF_ONESHOT); */

	error = mxt_power_init(data);
	if (error)
		goto err_free_object;

	mxt_regulator_enable(data);

#if 0
	error = mxt_initialize(data);
	if (error)
		goto err_disable_reg;
#endif

	sema_init(&data->crit_section_lock, 1);

#ifdef CONFIG_FB
	data->fb_notif.notifier_call = fb_notifier_callback;
	error = fb_register_client(&data->fb_notif);
	if (error)
		dev_err(&client->dev, "Error registering fb_notifier: %d\n",
			error);
#endif

	data->nfc_notifier.notifier_call = nfc_notifier_callback;
	error = register_nfc_notifier(&data->nfc_notifier);
	if(error)
		ATMEL_ERR("Error registering nfc_notifier.\n");

	error = kfifo_alloc(&nfc_event_fifo,PAGE_SIZE,GFP_KERNEL);
	if (error){
		ATMEL_ERR("Error kfifo_alloc for nfc_event_fifo.\n");
		kfifo_free(&nfc_event_fifo);
	}

	atomic_set(&data->nfc_enable_tp,TP_ENABLE);
	atomic_set(&data->enable_tp,TP_ENABLE);

	data->mxt_tp_mode = MXT_DEFAULT_TP_MODE;

	error = sysfs_create_group(&client->dev.kobj, &mxt_attr_group);
	if (error) {
		dev_err(&client->dev, "Failure %d creating sysfs group\n",
			error);
		goto err_free_object;
	}

	sysfs_bin_attr_init(&data->mem_access_attr);
	data->mem_access_attr.attr.name = "mem_access";
	data->mem_access_attr.attr.mode = S_IRUGO | S_IWUSR | S_IWGRP;
	data->mem_access_attr.read = mxt_mem_access_read;
	data->mem_access_attr.write = mxt_mem_access_write;
	data->mem_access_attr.size = data->mem_size;

	if (sysfs_create_bin_file(&client->dev.kobj,
				  &data->mem_access_attr) < 0) {
		dev_err(&client->dev, "Failed to create %s\n",
			data->mem_access_attr.attr.name);
		goto err_remove_sysfs_group;
	}

	data->mem_access_created = true;
	data->mode_is_persistent = true;

	error = mxt_initialize(data);
	if (error) {
		mxt_regulator_disable(data);
		msleep(MXT_REBOOT_TIME);
		mxt_regulator_enable(data);
		msleep(MXT_RESET_TIME);
		error = mxt_initialize(data);
		if (error) {
			data->init_completed = 0;
		} else {
			data->init_completed = 1;
		}
	} else {
		data->init_completed = 1;
	}

	TpIf_init(data);

	return 0;

err_remove_sysfs_group:
	sysfs_remove_group(&client->dev.kobj, &mxt_attr_group);
err_free_object:
	mxt_free_object_table(data);
#if 0
err_disable_reg:
#endif
	mxt_regulator_disable(data);
err_free_pdata:
	if (!dev_get_platdata(&data->client->dev))
		kfree(data->pdata);
err_free_mem:
	kfree(data);
	return error;
}

static int mxt_remove(struct i2c_client *client)
{
	struct mxt_data *data = i2c_get_clientdata(client);

	ATMEL_TRACE();

	mxt_suspend(&client->dev, MXT_SUSPEND_SYNC);

	mxt_destroy_task(data);

	TpIf_deinit(data);

	fb_unregister_client(&data->fb_notif);

	unregister_nfc_notifier(&data->nfc_notifier);

	if (data->mem_access_created)
		sysfs_remove_bin_file(&client->dev.kobj,
				      &data->mem_access_attr);

	sysfs_remove_group(&client->dev.kobj, &mxt_attr_group);
	free_irq(data->irq, data);
	mxt_power_deinit(data);
	mxt_gpio_free(data);
	mxt_free_object_table(data);
	if (!dev_get_platdata(&data->client->dev))
		kfree(data->pdata);
	kfree(data);
		kfifo_free(&nfc_event_fifo);

	return 0;
}

static void mxt_create_task(struct mxt_data *data)
{
	ATMEL_TRACE();

	if (data->main_wq == NULL) {
		data->main_wq =	create_singlethread_workqueue("mxt_main_wq");
	}

	INIT_DELAYED_WORK(&data->irq_work, mxt_work_irq);
#ifdef CONFIG_FB
	INIT_DELAYED_WORK(&data->resume_work, mxt_work_resume);
#endif
	INIT_DELAYED_WORK(&data->healthcheck_work, mxt_work_healthcheck);
	INIT_DELAYED_WORK(&data->nfc_sensing_work, mxt_work_nfc_sensing);
	INIT_DELAYED_WORK(&data->plugin_work, mxt_work_plugin);
	INIT_DELAYED_WORK(&data->fwupdate_work, mxt_work_fwupdate);
	INIT_DELAYED_WORK(&data->parse_mode_conf_work,mxt_work_parse_conf_work);

	return;
}

static void mxt_stop_all_task(struct mxt_data *data)
{
	ATMEL_TRACE();

	data->running_healthcheck = 0;

	cancel_delayed_work(&data->irq_work);
#ifdef CONFIG_FB
	cancel_delayed_work(&data->resume_work);
#endif
	cancel_delayed_work(&data->healthcheck_work);
	cancel_delayed_work(&data->plugin_work);
	cancel_delayed_work(&data->fwupdate_work);
	cancel_delayed_work(&data->nfc_sensing_work);

	if (data->main_wq != NULL) {
		flush_workqueue(data->main_wq);
	}

	return;
}

static void mxt_destroy_task(struct mxt_data *data)
{
	ATMEL_TRACE();

	mxt_stop_all_task(data);

	if (data->main_wq != NULL) {
		destroy_workqueue(data->main_wq);
	}

	return;
}

static void mxt_queue_irq(struct mxt_data *data)
{
	ATMEL_TRACE();

	if (data->main_wq != NULL) {
		queue_delayed_work(data->main_wq, &data->irq_work, 0);
	}

	return;
}

static void mxt_work_irq(struct work_struct *work)
{
	struct mxt_data *data =	container_of((struct delayed_work *)work,
					     struct mxt_data, irq_work);
	int state = mxt_get_sensor_state(data);

	ATMEL_TRACE();

	if (data->in_bootloader  ||
	    (!data->poweron && !data->mode_is_wakeable) ||
	    (state == STATE_UNKNOWN)) {

		ATMEL_DBG("interruption is ignored, "
			  "in_bootloader=%d, poweron=%d, mode_is_wakable=%d\n",
			  data->in_bootloader, data->poweron, data->mode_is_wakeable);

		/* bootloader state transition completion */
		complete(&data->bl_completion);
		goto end;
	}

	if (!data->object_table) {
		goto end;
	}

	if (data->T44_address)
		(void)mxt_process_messages_t44(data);
	else
		(void)mxt_process_messages(data);

end:
	return;
}

#ifdef CONFIG_FB
static void mxt_queue_resume(struct mxt_data *data)
{
	ATMEL_TRACE();

	if (data->main_wq != NULL) {
		queue_delayed_work(data->main_wq, &data->resume_work, 0);
	}

	return;
}

static void mxt_work_resume(struct work_struct *work)
{
	struct mxt_data *data =	container_of((struct delayed_work *)work,
					     struct mxt_data, resume_work);

	ATMEL_TRACE();

	mxt_resume(&data->client->dev);
	ATMEL_DBG("DISPLAY-ON\n");

	return;
}
#endif

static void mxt_queue_healthcheck(struct mxt_data *data)
{
	unsigned long expire = msecs_to_jiffies(MXT_HEALTHCHK_INTERVAL);

	ATMEL_TRACE();

	data->running_healthcheck = 1;
	if (data->main_wq != NULL) {
		queue_delayed_work(data->main_wq, &data->healthcheck_work,
				   expire);
	}
	return;
}

static void mxt_stop_healthcheck(struct mxt_data *data)
{
	ATMEL_TRACE();

	if (data->running_healthcheck == 1) {
		data->running_healthcheck = 0;
		cancel_delayed_work_sync(&data->healthcheck_work);
	}

	return;
}

static void mxt_work_healthcheck(struct work_struct *work)
{
	struct mxt_data *data =	container_of((struct delayed_work *)work,
					     struct mxt_data, healthcheck_work);

	ATMEL_TRACE();

	mxt_healthcheck(data);

	return;
}


static void mxt_queue_nfc_sening(struct mxt_data *data)
{
	unsigned long expire = msecs_to_jiffies(MXT_NFCSENING_INTERVAL);

	//ATMEL_TRACE();

	if (data->main_wq != NULL) {
		queue_delayed_work(data->main_wq, &data->nfc_sensing_work, expire);
	}
	return;
}

static void mxt_stop_nfc_sensing(struct mxt_data *data)
{
	ATMEL_TRACE();

	cancel_delayed_work_sync(&data->nfc_sensing_work);
	atomic_set(&data->nfc_enable_tp,TP_ENABLE);
	kfifo_reset(&nfc_event_fifo);

	return;
}

static void mxt_work_nfc_sensing(struct work_struct *work)
{
	struct mxt_data *data =	container_of((struct delayed_work *)work,
					     struct mxt_data,nfc_sensing_work);

	//ATMEL_TRACE();

	mxt_nfc_sensing(data);

	return;
}

static void mxt_nfc_sensing(struct mxt_data *data)
{
	unsigned long nfc_status ;
	unsigned long last_detect_time = 0;
	unsigned long not_detect_time = 0 ;
	int error ;

	ATMEL_TRACE();
	if((data->force_restart_time <= 0)||(data->restart_wait_time <= 0))
	{
		goto end;
	}

	error = kfifo_out(&nfc_event_fifo,&nfc_status,sizeof(unsigned long));
	if((error != sizeof(unsigned long))||(nfc_status != NFC_MSG_RF_INTF_ACTIVATED))
	{
		//ATMEL_DBG("The first event is not NFC_MSG_RF_INTF_ACTIVATED , error = %d nfc_status = %lu.\n",error,nfc_status);
		goto end;
	}

	last_detect_time = jiffies ;

	/*stop reporting the coordinate*/
	atomic_set(&data->nfc_enable_tp,TP_DISABLE);
	ATMEL_DBG("Stop reporting coordinate.\n");

receive_not_detect:
	error = kfifo_out(&nfc_event_fifo,&nfc_status,sizeof(unsigned long));
	if((error != sizeof(unsigned long)) ||(NFC_MSG_RF_DEACTIVATE != nfc_status)){
		if(time_after(jiffies,last_detect_time + msecs_to_jiffies(data->force_restart_time))){
			/* restart reporting the coordinate*/
			atomic_set(&data->nfc_enable_tp,TP_ENABLE);
			ATMEL_DBG("Elapsed time larger than force_restart_time ,restart reporting coordinate.\n");
			goto end;
		}
		else{
			goto receive_not_detect;
		}
	}
	else{
		 not_detect_time = jiffies;
receive_detect: 
		error = kfifo_out(&nfc_event_fifo,&nfc_status,sizeof(unsigned long));
		 if((error == sizeof(unsigned long)) && (NFC_MSG_RF_INTF_ACTIVATED == nfc_status)){
			last_detect_time = jiffies ;
			goto receive_not_detect;
		 }
		 else{
			 if(time_after(jiffies,not_detect_time + msecs_to_jiffies(data->restart_wait_time))){
			        /* restart reporting the coordinate*/
				atomic_set(&data->nfc_enable_tp,TP_ENABLE);
				ATMEL_DBG("Elapsed time larger than restart_wait_time,restart reporting coordinate.\n");
			 }
			 else goto receive_detect;
		 }
	}
end:
	mxt_queue_nfc_sening(data);
	return;

}


static void mxt_work_parse_conf_work(struct work_struct *work)
{

	struct mxt_data *data =	container_of((struct delayed_work *)work,
					     struct mxt_data,parse_mode_conf_work);
	ATMEL_TRACE();

	mxt_parse_mode_conf(data);

	return;
}

static void mxt_queue_plugin(struct mxt_data *data)
{
	ATMEL_TRACE();

	if (data->main_wq != NULL) {
		queue_delayed_work(data->main_wq, &data->plugin_work, 0);
	}
	return;
}

static void mxt_work_plugin(struct work_struct *work)
{
	ATMEL_TRACE();

	TpIf_queued_plugin_cb_func(work);

	return;
}

static int mxt_update_firmware(struct mxt_data *data, char *filename)
{
	int error;
	struct device *dev = &data->client->dev;

	ATMEL_TRACE();

	error = mxt_update_file_name(dev, &data->fw_name,
				     filename, strlen(filename));
	if (error) {
		ATMEL_ERR("mxt_update_file_name() failed(%d)\n", error);
		return error;
	}

	error = mxt_load_fw(dev);
	if (error) {
		ATMEL_ERR("The firmware update failed(%d)\n", error);
	} else {
		ATMEL_MSG("The firmware update succeeded\n");

		/* interrupt handler is released to switch irq in mxt_initialize() */
		mxt_irq_enable(data, false);
		free_irq(data->irq, data);
		mxt_free_input_device(data);

		mxt_set_sensor_state(data, STATE_QUERY);

		error = mxt_initialize(data);
		if (error) {
			ATMEL_ERR("mxt_initialize() after updating fw is failed(%d)\n", error);
			data->init_completed = 0;
			return error;
		} else {
			data->init_completed = 1;
		}
	}

	return 0;
}

static int mxt_update_config(struct mxt_data *data, char *filename)
{
	int ret;
	struct device *dev = &data->client->dev;

	ATMEL_TRACE();

	if (data->in_bootloader) {
		dev_err(dev, "Not in appmode\n");
		return -EINVAL;
	}

	ret = mxt_update_file_name(dev, &data->cfg_name,
				   filename, strlen(filename));
	if (ret)
		return ret;

	mxt_free_input_device(data);

	ret = mxt_configure_objects(data);
	if (ret)
		goto out;

out:
	return ret;
}

static void mxt_queue_fwupdate(struct mxt_data *data)
{
	ATMEL_TRACE();

	if (data->main_wq != NULL) {
		queue_delayed_work(data->main_wq, &data->fwupdate_work, 0);
	}

	return;
}

static void mxt_work_fwupdate(struct work_struct *work)
{
	int suspended = 0;
	int ret_fw = 0, ret_cfg = 0;
	int ret = 0;
	struct mxt_data *data =	container_of((struct delayed_work *)work,
					     struct mxt_data, fwupdate_work);
	struct TouchPanelFwUpdateInfo_t *info = mTpCmd.fwUpdate;

	ATMEL_TRACE();

	if (info == NULL) {
		ATMEL_ERR("fw update info error\n");
		return;
	}

	if (atomic_read(&data->suspended)) {
		suspended = 1;
		mxt_resume(&data->client->dev);
	}
	cancel_delayed_work(&data->healthcheck_work);
	cancel_delayed_work(&data->nfc_sensing_work);

	mxt_lock(&data->crit_section_lock);

	force_update = info->force;
	if (info->target == TPIF_FW_UPDATE_ALL) {
		/* update firmware */
		ret_fw = mxt_update_firmware(data, info->fw_name);
		if (ret_fw == 0) {
			/* update config */
			ret_cfg = mxt_update_config(data, info->cfg_name);
		}

		if ((ret_fw != 0) || (ret_cfg != 0)) {
			ATMEL_ERR("fw update(all) failed, fw=%d,cfg=%d\n",
				  ret_fw, ret_cfg);
			ret = -1;
		}
	} else if (info->target == TPIF_FW_UPDATE_CFG) {
		/* update config */
		ret_cfg = mxt_update_config(data, info->cfg_name);
		if (ret_cfg != 0) {
			ATMEL_ERR("fw update(cfg) failed, cfg=%d\n", ret_cfg);
			ret = -1;
		}
	} else {
		ATMEL_ERR("unknown update target\n");
		ret = -1;
	}
	force_update = 0;

	mxt_unlock(&data->crit_section_lock);

	if (ret == 0) {
		info->result = 0;
	} else {
		ATMEL_ERR("firmware update failed\n");
		info->result = -1;
	}

	if (suspended == 1) {
		mxt_suspend(&data->client->dev, MXT_SUSPEND_ASYNC);
	} else {
		mxt_resume(&data->client->dev);
	}

	return;
}


#ifdef CONFIG_PM
static int mxt_suspend(struct device *dev, int awaitable)
{
	struct i2c_client *client = to_i2c_client(dev);
	struct mxt_data *data = i2c_get_clientdata(client);
	static char ud_stats[PAGE_SIZE];

	ATMEL_TRACE();

	if ((data->pdata->ts_pinctrl != NULL)&&(1 != data->enable_fw_update)) {
		int retval = pinctrl_select_state(data->pdata->ts_pinctrl,
						  data->pdata->pinctrl_state_suspend);
		if (retval < 0) {
			ATMEL_ERR("Failed to select %s pinstate %d\n",PINCTRL_STATE_SUSPEND, retval);
			return -1;
		}
	}

	if (atomic_cmpxchg(&data->suspended, 0, 1) == 0) {
		mxt_lock(&data->crit_section_lock);

		mxt_set_sensor_state(data, STATE_SUSPEND);

		if (awaitable == MXT_SUSPEND_SYNC) {
			cancel_delayed_work_sync(&data->irq_work);
			cancel_delayed_work_sync(&data->plugin_work);
		} else {
			cancel_delayed_work(&data->irq_work);
			cancel_delayed_work(&data->plugin_work);
		}
		mxt_stop_healthcheck(data);
		mxt_stop_nfc_sensing(data);

		atomic_set(&data->enable_tp,TP_ENABLE);
		mxt_reset_slots(data);

		if (data->use_regulator)
			mxt_regulator_disable(data);
	}

	data->poweron = false;

	if (mPlugin != NULL) {
		mPlugin->reset();
	}

	mxt_ud_stat(ud_stats, sizeof(ud_stats));
	pr_info("%s\n", ud_stats);

	return 0;
}

static int mxt_resume(struct device *dev)
{
	struct i2c_client *client = to_i2c_client(dev);
	struct mxt_data *data = i2c_get_clientdata(client);
	int state = mxt_get_sensor_state(data);
	int has_error = 0;

	ATMEL_TRACE();


	if (data->pdata->ts_pinctrl != NULL) {
		has_error = pinctrl_select_state(data->pdata->ts_pinctrl,
					     data->pdata->pinctrl_state_active);
		if (has_error < 0) {
			ATMEL_ERR("Failed to select %s pinstate %d\n", PINCTRL_STATE_ACTIVE, has_error);
			return -1;
		}
	}

	if (atomic_cmpxchg(&data->suspended, 1, 0) == 1) {
		if (data->use_regulator) {
			mxt_regulator_enable(data);

			if (data->init_completed == 0) {
				int error;
				error = mxt_initialize(data);
				if (error) {
					mxt_regulator_disable(data);
					msleep(MXT_REBOOT_TIME);
					mxt_regulator_enable(data);
					msleep(MXT_RESET_TIME);
					error = mxt_initialize(data);
					if (error) {
						data->init_completed = 0;
					} else {
						data->init_completed = 1;
					}
				} else {
					data->init_completed = 1;
				}
			}

			if (data->init_completed == 1) {
				if (mxt_acquire_irq(data) == -1) {
					if (mxt_hw_reset(data) == -1) {
						has_error = 1;
					}
				}
			}
			if ((data->init_completed == 1) && (has_error == 0)) {
				mxt_set_t7_power_cfg(data, MXT_POWER_CFG_RUN);
				mxt_t6_command(data, MXT_COMMAND_CALIBRATE, 1, false);
			} else {
				mxt_regulator_disable(data);
			}
		} else if (!data->in_bootloader)
			(void)mxt_hw_reset(data);

		mxt_unlock(&data->crit_section_lock);
	}

	if ((data->init_completed == 1) && (has_error == 0)) {
		if (data->in_bootloader) {
			state = (state == STATE_INIT ||
				 state == STATE_FLASH) ? state : STATE_BL;
		} else {
			state = STATE_ACTIVE;
		}

		data->poweron = true;
		mxt_set_sensor_state(data, state);

		mxt_init_feature(data);
		mxt_queue_healthcheck(data);
		mxt_queue_nfc_sening(data);
	}

	return 0;
}

#ifdef CONFIG_FB
static int fb_notifier_callback(struct notifier_block *self,
				 unsigned long event, void *data)
{
	struct fb_event *evdata = data;
	int *blank;
	struct mxt_data *mxt_dev_data =
		container_of(self, struct mxt_data, fb_notif);

	ATMEL_TRACE();

	if (evdata && evdata->data && event == FB_EVENT_BLANK && mxt_dev_data &&
			mxt_dev_data->client) {
		blank = evdata->data;
		if (*blank == FB_BLANK_UNBLANK ||
				(*blank == FB_BLANK_VSYNC_SUSPEND &&
				atomic_read(&mxt_dev_data->suspended))) {
			mxt_queue_resume(mxt_dev_data);
			ATMEL_DBG("queued RESUME, blank=%d\n", *blank);
		} else if (*blank == FB_BLANK_POWERDOWN) {
			/* ensure no work left in queue */
			cancel_delayed_work_sync(&mxt_dev_data->resume_work);

			mxt_suspend(&mxt_dev_data->client->dev, MXT_SUSPEND_SYNC);
			ATMEL_DBG("DISPLAY-OFF\n");
		}
	}

	return 0;
}
#endif

static int nfc_notifier_callback(struct notifier_block *this,
				unsigned long event, void *ptr)
{
	ATMEL_TRACE();

	ATMEL_DBG("Receive NFC event : %ld.\n",event);
	if (istpenable){
		return 0;
	}
	if((NFC_MSG_RF_INTF_ACTIVATED == event) \
			||(NFC_MSG_RF_DEACTIVATE == event))
		kfifo_in(&nfc_event_fifo,&event,sizeof(unsigned long));

	return 0;
}

static void TpIf_init(struct mxt_data *data)
{
	int ret = 0;
	uint8_t nvm_data;
	int retry_cnt = 0;

	ATMEL_TRACE();

	ret = misc_register(&tpif_dev);
	if (ret != 0)
	{
		ATMEL_ERR("misc_register() failed, ret=%d\n", ret);
	}

	mId = -1;
	mPlugin = NULL;

	memset(&mCurrentEvent, 0, sizeof(mCurrentEvent));

	mTpCmd.pdata = data;
	sema_init(&mTpCmd.lock, 1);
	mTpCmd.gloveMode = -1;
	mTpCmd.gloveSens = -1;
	mTpCmd.waterProofMode = -1;
	mTpCmd.waterResistance = -1;
	mTpCmd.stylusMode = -1;
	mTpCmd.stylusType = -1;
	mTpCmd.stylusActivePressure = TP_PEN_ACTIVE_PRESSURE_NORMAL;
	mTpCmd.fwUpdate = NULL;

	mTpCmd.pdata->glove_conf = TP_GLOVE_SENS_MID;
	mTpCmd.pdata->waterproof_conf = TP_WATER_RESISTANCE_HIGH;
	mTpCmd.pdata->stylus_conf = TP_PEN_TYPE_PASSIVE;

#ifdef ENABLE_CFGDRV
	/* read flag from nvm */
	do {
		ret = cfgdrv_read(D_HCM_E_TP_FW_REFLASH_FLAG,
				  sizeof(nvm_data), &nvm_data);

		if (ret == D_HCM_NOT_INITIALIZE) {
			ATMEL_MSG("nvm is not initialized (%d)\n", retry_cnt);

			retry_cnt++;
			msleep(TP_NVM_INIT_WAIT);
		}
	} while((ret != 0) && (retry_cnt < TP_MAX_RETRY_NVM_INIT));
#else
	nvm_data = retry_cnt = 0;
#endif

	/* create procfs for factory command */
	tpif_procfs_dir = proc_mkdir(TPIF_PROCFS_DIR, NULL);

	tpif_pin_fault_test_entry = proc_create(TPIF_PROCFS_SELFTEST12,
						(S_IFREG|S_IRUSR|S_IWUSR),
						tpif_procfs_dir,
						&tpif_pin_fault_test_fops);
	tpif_signal_limit_test_entry = proc_create(TPIF_PROCFS_SELFTEST17,
						   (S_IFREG|S_IRUSR|S_IWUSR),
						   tpif_procfs_dir,
						   &tpif_signal_limit_test_fops);

}

static void TpIf_deinit(struct mxt_data *data)
{
	ATMEL_TRACE();

	remove_proc_entry(TPIF_PROCFS_SELFTEST17, tpif_procfs_dir);
	remove_proc_entry(TPIF_PROCFS_SELFTEST12, tpif_procfs_dir);
	remove_proc_entry(TPIF_PROCFS_DIR, NULL);

	misc_deregister(&tpif_dev);
}

static void TpIf_queued_plugin_cb_func(struct work_struct *w)
{

	struct mxt_data *data = mTpCmd.pdata;

	ATMEL_TRACE();

	/* Update config flag in struct mxt_data */
	mxt_lock(&mTpCmd.lock);
	if(mTpCmd.fingerSens != -1){
		data->finger_sensitivity = mTpCmd.fingerSens;
		mTpCmd.fingerSens = -1;
	}

	if (mTpCmd.gloveMode != -1) {
		data->glove_enabled = mTpCmd.gloveMode;
		mTpCmd.gloveMode = -1;
	}

	if (mTpCmd.gloveSens != -1) {
		data->glove_sensitivity = mTpCmd.gloveSens;
		mTpCmd.gloveSens = -1;
	}

	if (mTpCmd.waterProofMode != -1) {
		data->waterproof_enabled = mTpCmd.waterProofMode;
		mTpCmd.waterProofMode = -1;
	}

	if (mTpCmd.waterResistance != -1) {
		data->waterResistance = mTpCmd.waterResistance;
		mTpCmd.waterResistance = -1;
	}

	if (mTpCmd.stylusMode != -1) {
		switch (mTpCmd.pdata->stylus_conf)
		{
		case TP_PEN_TYPE_ACTIVE:
			data->passive_stylus_enabled = 0;
			data->active_stylus_enabled = mTpCmd.stylusMode;
			break;
		case TP_PEN_TYPE_PASSIVE:
			data->passive_stylus_enabled = mTpCmd.stylusMode;
			data->active_stylus_enabled = 0;
			break;
		case TP_PEN_TYPE_ACTIVE_PASSIVE:
			data->passive_stylus_enabled = mTpCmd.stylusMode;
			data->active_stylus_enabled = mTpCmd.stylusMode;
			break;
		default:
			break;
		}
		mTpCmd.stylusMode = -1;
	}

	if (mTpCmd.stylusType != -1) {
		int currentMode = ((mTpCmd.pdata->passive_stylus_enabled
				    |mTpCmd.pdata->active_stylus_enabled) == 0) ? 0 : 1;
		data->stylus_conf = mTpCmd.stylusType;

		switch (mTpCmd.pdata->stylus_conf) {
		case TP_PEN_TYPE_ACTIVE:
			data->passive_stylus_enabled = 0;
			data->active_stylus_enabled = currentMode;
			break;

		case TP_PEN_TYPE_PASSIVE:
			data->active_stylus_enabled = 0;
			data->passive_stylus_enabled = currentMode;
			break;

		case TP_PEN_TYPE_ACTIVE_PASSIVE:
			data->active_stylus_enabled =currentMode;
			data->passive_stylus_enabled = currentMode;
			break;
		default:
			break;
		}
		mTpCmd.stylusType = -1;
	}

	if (mTpCmd.stylusActivePressure != -1) {
		data->active_stylus_pressure = mTpCmd.stylusActivePressure;
		mTpCmd.stylusActivePressure = -1;
	}

	/*Calculate mxt_tp_mode */

	data->mxt_tp_mode = data->mxt_tp_mode & 0xFFF0;
	switch(data->finger_sensitivity){
		case TP_FINGER_SENS_LOW:
			data->mxt_tp_mode = data->mxt_tp_mode | 0x0000;
			break;
		case TP_FINGER_SENS_HIGH:
			data->mxt_tp_mode = data->mxt_tp_mode | 0x0002;
			break;
		case TP_FINGER_SENS_MID:
		default:
			data->mxt_tp_mode = data->mxt_tp_mode | 0x0001;
			break;
	}

	data->mxt_tp_mode = data->mxt_tp_mode & 0xFF0F;
	if(data->glove_enabled){
		switch(data->glove_sensitivity)
		{
			case TP_GLOVE_SENS_HIGH:
				data->mxt_tp_mode = data->mxt_tp_mode | 0x0030;
				break;
			case TP_GLOVE_SENS_LOW:
				data->mxt_tp_mode = data->mxt_tp_mode | 0x0010;
				break;
			case TP_GLOVE_SENS_MID:
			default:
				data->mxt_tp_mode = data->mxt_tp_mode | 0x0020;
				break;
		}
	}

	data->mxt_tp_mode = data->mxt_tp_mode & 0xF0FF;
	if(data->waterproof_enabled){
		switch(data->waterResistance){
			case TP_WATER_RESISTANCE_NORMAL:
			 	data->mxt_tp_mode = data->mxt_tp_mode | 0x0100;
				break;
			case TP_WATER_RESISTANCE_HIGH:
			default:
				data->mxt_tp_mode = data->mxt_tp_mode | 0x0200;
				break;
		}
	}

	data->mxt_tp_mode = data->mxt_tp_mode & 0xCFFF;
	if(data->passive_stylus_enabled)
		data->mxt_tp_mode = data->mxt_tp_mode | 0x1000;

	data->mxt_tp_mode = data->mxt_tp_mode & 0x3FFF;
	if(data->active_stylus_enabled){
		switch(data->active_stylus_pressure){
			case TP_PEN_ACTIVE_PRESSURE_LIGHT:
				data->mxt_tp_mode = data->mxt_tp_mode | 0x8000;
				break;
			case TP_PEN_ACTIVE_PRESSURE_NORMAL:
			default:
				data->mxt_tp_mode = data->mxt_tp_mode | 0x4000;
				break;
		}
	}

	/*Update detect mode*/
	mxt_set_detect_mode(data);
	mxt_unlock(&mTpCmd.lock);
}

static void TpIf_convertPos(struct TpPos_t *pos, int slot, u8 *message)
{
	u8 status;
	u8 tool_type;
	int x, y;
	u8 exp;
	u8 active_stylus_status;

	ATMEL_TRACE();

	pos->slot = slot;

	status = message[1];
	tool_type = (status & MXT_T100_TYPE_MASK);
	x = (message[3] << 8) | message[2];
	y = (message[5] << 8) | message[4];

	if (status & MXT_T100_DETECT) {
		pos->x = x;
		pos->y = y;

		switch (tool_type) {
		case MXT_T100_TYPE_FINGER:
			pos->option |= TP_POS_FIN;
			exp = (message[8] & 0x60) >> 5;
			pos->z = message[6] << TP_Z_SHIFT;
			pos->wx = TP_MM2DOT(((message[9] & 0xF0) >> 4) << exp);
			pos->wy = TP_MM2DOT((message[9] & 0x0F) << exp);
			pos->area = ((message[8] & 0x1F) << exp);
			pos->misc = message[7];
			ATMEL_DAT("[%d] x=%d,y=%d,z=%d,wx=%d,wy=%d,area=%d,misc=%d,exp=%u,finger\n",
				  slot, pos->x, pos->y, pos->z, pos->wx, pos->wy, pos->area, pos->misc, exp);
			break;
		case MXT_T100_TYPE_PASSIVE_STYLUS:
			pos->option |= TP_POS_PEN;
			exp = (message[8] & 0x60) >> 5;
			pos->z = message[6] << TP_Z_SHIFT;
			pos->wx = TP_MM2DOT(((message[9] & 0xF0) >> 4) << exp);
			pos->wy = TP_MM2DOT((message[9] & 0x0F) << exp);
			pos->area = ((message[8] & 0x1F) << exp);
			pos->misc = message[7];
			ATMEL_DAT("[%d] x=%d,y=%d,z=%d,wx=%d,wy=%d,area=%d,misc=%d,exp=%u,ppen\n",
				  slot, pos->x, pos->y, pos->z, pos->wx, pos->wy, pos->area, pos->misc, exp);
			break;
		case MXT_T100_TYPE_GLOVE:
			pos->option |= TP_POS_GLOVE;
			exp = (message[8] & 0x60) >> 5;
			pos->z = message[6];
			pos->wx = TP_MM2DOT(((message[9] & 0xF0) >> 4) << exp);
			pos->wy = TP_MM2DOT((message[9] & 0x0F) << exp);
			pos->area = ((message[8] & 0x1F) << exp);
			pos->misc = message[7];
			ATMEL_DAT("[%d] x=%d,y=%d,z=%d,wx=%d,wy=%d,area=%d,misc=%d,exp=%u,glove\n",
				  slot, pos->x, pos->y, pos->z, pos->wx, pos->wy, pos->area, pos->misc, exp);
			break;
		case MXT_T100_TYPE_ACTIVE_STYLUS:
			active_stylus_status = message[6];
			if (active_stylus_status & MXT_T100_ACTIVE_STYLUS_HOVER) {
				pos->option |= TP_POS_APEN;

				/* this event is available */
				if (active_stylus_status & MXT_T100_ACTIVE_STYLUS_TIPSWITH) {
					/* stylus is touching */
					pos->z = (message[7] << 2) + (0x3 & (message[6] >> 6));
					if (pos->z == 0) {
						/* non-hovering event must have valid z value */
						pos->z = 1;
					}
					pos->wx = pos->wy = TP_MM2DOT(1);
				} else {
					/* stylus is hovering */
					pos->option |= TP_POS_HOVER;
					pos->z = 0;
					pos->wx = pos->wy = 0;
				}

				/* battery level */
				pos->misc = (short)message[8];

				/* buttons */
				if (message[6] & MXT_T100_ACTIVE_STYLUS_BUTTON0) {
					pos->option |= TP_BTN_BTN1;
				}
				if (message[6] & MXT_T100_ACTIVE_STYLUS_BUTTON1) {
					pos->option |= TP_BTN_BTN2;
				}
				ATMEL_DAT("[%d] x=%d,y=%d,z=%d,wx=%d,wy=%d,op=%08X,batt=%d,apen\n",
					  slot, pos->x, pos->y, pos->z, pos->wx, pos->wy, 
					  (unsigned int)pos->option, pos->misc);

				/* ignore force calib when active stylus is detected */
				mForceCalibCnt = touch_dev_cfg.calib_cfg.force_calib_cnt;
			} else {
				pos->option = 0;
				pos->z = 0;
				pos->wx = pos->wy = 0;
				ATMEL_DAT("[%d] x=%d,y=%d,z=%d,wx=%d,wy=%d,apen(out)\n",
					  slot, pos->x, pos->y, pos->z, pos->wx, pos->wy);
			}
			break;
		default:
			pos->option = 0;
			pos->z = 0;
			pos->wx = pos->wy = 0;
			break;
		}
	} else {
		pos->option = 0;
		pos->x = pos->y = 0;
		pos->z = 0;
		pos->wx = pos->wy = 0;
	}
}

static void InputSubsystem_store(const struct TpPos_t *pos)
{
	int slot = pos->slot;
	struct TpPos_t *currPos = &mCurrentEvent.pos[slot];

	ATMEL_TRACE();

	if (TP_POS_IS_VALID(pos->option)) {
		currPos->x = pos->x;
		currPos->y = pos->y;
		currPos->z = pos->z;
		currPos->wx = pos->wx;
		currPos->wy = pos->wy;
		currPos->area = pos->area;
		currPos->misc = pos->misc;
		currPos->option = pos->option;
		mCurrentEvent.posState |= (1 << slot);
	} else {
		currPos->x = 0;
		currPos->y = 0;
		currPos->z = 0;
		currPos->wx = 0;
		currPos->wy = 0;
		currPos->area = 0;
		currPos->misc = 0;
		currPos->option = 0;
		mCurrentEvent.posState &= ~(1 << slot);
	}
}

static void InputSubsystem_update(void)
{
	int i;

	ATMEL_TRACE();

	mCurrentEvent.time = jiffies;
	mCurrentEvent.numFinger = 0;

	for (i = 0; i < TP_MAX_POS; i++) {
		if (TP_POS_IS_VALID(mCurrentEvent.pos[i].option)) {
			if ((mCurrentEvent.pos[i].z != 0)
			    && !(mCurrentEvent.pos[i].option & TP_POS_HOVER)) {
				mCurrentEvent.numFinger++;
			}
		}
	}

	return;
}

static void InputSubsystem_enqueue(struct input_dev *input_dev,
				   const struct TpEvent_t *ev)
{
	int i;

	ATMEL_TRACE();

	mCurrentEvent.numFinger = 0;

	for (i = 0; i < TP_MAX_POS; i++) {
		int slot_state = 0;
		const struct TpPos_t *pos = &ev->pos[i];
		int x = pos->x;
		int y = pos->y;
		int z = pos->z;

		input_mt_slot(input_dev, i);
		if (TP_POS_IS_VALID(pos->option)) {
			/* const int isFinger     = (pos->option & TP_POS_FIN); */
			/* const int isGlove      = (pos->option & TP_POS_GLOVE); */
			const int isPassivePen = (pos->option & TP_POS_PEN);
			const int isActivePen  = (pos->option & TP_POS_APEN);
			const int isHover      = (pos->option & TP_POS_HOVER);
			const int tool = (isPassivePen|isActivePen) ? MT_TOOL_PEN : MT_TOOL_FINGER;
			const int major = (pos->wx + pos->wy) / 2;
			/* touch event */
			input_report_abs(input_dev, ABS_MT_POSITION_X, x);
			input_report_abs(input_dev, ABS_MT_POSITION_Y, y);
			input_report_abs(input_dev, ABS_MT_PRESSURE, z);
			input_report_abs(input_dev, ABS_MT_TOUCH_MAJOR, major);
			/* input_report_abs(input_dev, ABS_MT_TOUCH_MINOR, minor); */
			if (isActivePen != 0) {
				u8 button1 = (pos->option & TP_BTN_BTN1) ? 1 : 0;
				u8 button2 = (pos->option & TP_BTN_BTN2) ? 1 : 0;
				input_report_key(input_dev, BTN_STYLUS, button1);
				input_report_key(input_dev, BTN_STYLUS2, button2);
			}
			/* update slot state */
			if (isHover == 0) {
				if (z > 0) {
					/* non-hovering, valid position detected */
					mCurrentEvent.numFinger++;
					slot_state = 1;
				} else {
					/* release */
					slot_state = 0;
				}
			} else {
				if (pos->wx > 0) {
					/* hovering position detected */
					slot_state = 1;
				} else {
					/* hover out */
					slot_state = 0;
				}
			}
			input_mt_report_slot_state(input_dev, tool, slot_state);
		} else {
			if ((x != 0) || (y != 0)) {
				slot_state = -1; /* ignore data, keep slot state */
			} else {
				/* release */
				input_mt_report_slot_state(input_dev, MT_TOOL_FINGER, 0);
				slot_state = 0;
			}
		}

		if (slot_state == 0) {
			memset(&mCurrentEvent.pos[i], 0, sizeof(mCurrentEvent.pos[i]));
		}
	}

	mxt_input_sync(input_dev, mCurrentEvent.numFinger);

	if (mCurrentEvent.posState == 0) {
		/* reset force calib count when no touch is detected */
		mForceCalibCnt = touch_dev_cfg.calib_cfg.force_calib_cnt;
	}

	return;
}

static const struct DefaultParams_t *TpIfMainAndroid_getDefault(void)
{
	static const struct DefaultParams_t params = {
		{
			{0,  719, 0, 0},
			{0, 1279, 0, 0},
			{0, MXT_MAX_PRESSURE, 0, 0},
			{0, MXT_MAX_AREA, 0, 0}
		}
	};

	return &params;
}

static void TpIfMainAndroid_getFeature(int type, int *mode)
{
	ATMEL_TRACE();

	switch (type) {
        case TP_TYPE_GLOVE:
		*mode = mTpCmd.pdata->glove_enabled;
		break;
        case TP_TYPE_WATER_PROOF:
		*mode = mTpCmd.pdata->waterproof_enabled;
		break;
        case TP_TYPE_PEN:
		*mode = ((mTpCmd.pdata->passive_stylus_enabled|mTpCmd.pdata->active_stylus_enabled) == 0) ? 0 : 1;
		break;
	default:
		*mode = 0;
		break;
	}
}

static void TpIfMainAndroid_setFeature(int type, int mode)
{
	int queued = 0;

	ATMEL_TRACE();

	switch (type) {
        case TP_TYPE_GLOVE:
		mTpCmd.gloveMode = mode;
		queued = 1;
		break;
        case TP_TYPE_WATER_PROOF:
		mTpCmd.waterProofMode = mode;
		queued = 1;
		break;
        case TP_TYPE_PEN:
		mTpCmd.stylusMode = mode;
		queued = 1;
		break;
        default:
		break;
	}

	if (queued != 0) {
		mxt_queue_plugin(mTpCmd.pdata);
	}
}

static int TpIfMainAndroid_funcFeature(int func, unsigned long arg)
{
	int queued = 0;

	ATMEL_TRACE();

	switch (func) {
	case TP_FUNC_SET_GLOVE_SENS:
		mTpCmd.gloveSens = (int)arg;
		queued = 1;
		break;
        case TP_FUNC_SET_WATER_RESISTANCE:
		mTpCmd.waterResistance = (int)arg;
		queued = 1;
		break;
        case TP_FUNC_SET_PEN_TYPE:
		mTpCmd.stylusType = (int)arg;
		queued = 1;
		break;
        case TP_FUNC_SET_PEN_ACTIVE_PRESSURE:
		mTpCmd.stylusActivePressure = (int)arg;
		queued = 1;
		break;
	case TP_FUNC_SET_FINGER_SENS:
		mTpCmd.fingerSens = (int)arg;
		queued = 1;
        default:
		break;
	}

	if (queued != 0) {
		mxt_queue_plugin(mTpCmd.pdata);
	}

	return 0;
}

static int TpIfMain_install(struct TouchPanelPlugin *plugin,
                             struct TouchPanelHandler **handler)
{
	ATMEL_TRACE();

	if (mId == -1) {
		mId = (int)(0x7FFFFFFF & jiffies);
		mPlugin = plugin;
		*handler = &mHandler;
	}

	return mId;
}

static void TpIfMain_uninstall(int id)
{
	ATMEL_TRACE();

	if (mId == id) {
		mPlugin = NULL;
		mId = -1;
	}
}

static unsigned int tpif_fops_poll(struct file *file, poll_table *wait)
{
	return POLLERR;
}

static int tpif_fops_open(struct inode *inode, struct file *file)
{
	return nonseekable_open(inode, file);
}

static int tpif_fops_close(struct inode *inode, struct file *file)
{
	return 0;
}

static long tpif_dev_ioctl(struct file *filp, unsigned int cmd,
			   unsigned long arg)
{
	int ret = 0;
	struct mxt_data *data = mTpCmd.pdata;

	ATMEL_TRACE();

	switch (cmd) {
	case TPIF_IOCTL_GET_VERSION:
	{
		struct TouchPanelFwVersion_t v;
		int ret_fw = 0, ret_cfg = 0;

		int suspended = 0;
		if (atomic_read(&data->suspended)) {
			suspended = 1;
			mxt_resume(&data->client->dev);
			mxt_stop_healthcheck(data);
			mxt_stop_nfc_sensing(data);
		}

		memset(&v, 0, sizeof(v));

		mxt_lock(&data->crit_section_lock);
		ret_fw = mxt_read_fw_version(data, &v.base);
		ret_cfg = mxt_read_cfg_version(data, &v.config, &v.config_rev);
		mxt_unlock(&data->crit_section_lock);

		if (ret_fw != 0 || ret_cfg != 0) {
			ret = -1;
		} else {
			ret = copy_to_user((void *)arg, &v, sizeof(v));
			if (ret != 0) {
				ATMEL_ERR("copy_to_user failed\n");
				ret = -1;
			}
		}

		if (suspended == 1) {
			mxt_suspend(&data->client->dev, MXT_SUSPEND_SYNC);
		}
		break;
	}
        case TPIF_IOCTL_UPDATE_FW:
        {
                struct TouchPanelFwUpdateInfo_t *info;

		if (mTpCmd.fwUpdate != NULL) {
			ATMEL_ERR("another update is processing\n");
			ret = -EAGAIN;
			break;
		}

		info = (struct TouchPanelFwUpdateInfo_t *)
			kzalloc(sizeof(struct TouchPanelFwUpdateInfo_t),
				GFP_KERNEL);
		if (info == NULL) {
			ATMEL_ERR("memory allocation failed\n");
			ret = -1;
			break;
		}

                ret = copy_from_user(info, (void *)arg, sizeof(*info));
                if (ret != 0) {
			ATMEL_ERR("copy_from_user failed\n");
			ret = -1;
			kfree(info);
			break;
                }
		info->fw_name[ARRAY_SIZE(info->fw_name)-1] = '\0';
		info->cfg_name[ARRAY_SIZE(info->cfg_name)-1] = '\0';

		data->enable_fw_update = 1;
		mTpCmd.fwUpdate = info;
		if (flush_delayed_work(&data->irq_work) != false) {
			ATMEL_ERR("irq work pended but force flush from queue\n");
		}
		mxt_queue_fwupdate(data);
		flush_delayed_work(&data->fwupdate_work);
		data->enable_fw_update = 0;

		ATMEL_MSG("fwupdate work complete\n");

                ret = copy_to_user((void *)arg, info, sizeof(*info));
                if (ret != 0) {
			ATMEL_ERR("copy_to_user failed\n");
			ret = -1;
                }

		kfree(info);
		mTpCmd.fwUpdate = NULL;
		break;
	}
	case TPIF_IOCTL_GET_PANEL:
	{
		char panel[256] = {'\0'};
		struct TouchPanelFwUpdateInfo_t *plinfo;
		plinfo = (struct TouchPanelFwUpdateInfo_t *)kzalloc(sizeof(struct TouchPanelFwUpdateInfo_t),
				GFP_KERNEL);
		if (plinfo == NULL) {
			ATMEL_ERR("memory allocation failed\n");
			ret = -1;
			break;
		}
		mdss_get_panel_name(panel);
		strcpy(plinfo->panel_name, panel);
		ret = copy_to_user((void*)arg, plinfo, sizeof(*plinfo));
                if (ret != 0) {
			ATMEL_ERR("copy_to_user failed\n");
			ret = -1;
                }
		kfree(plinfo);
		break;
	}
	case TPIF_IOCTL_GET_TOUCH:
	{
		char old_touch[16] = "0x01";
		char new_touch[16] = "0x02";
		u8 tp_cfg;
		struct TouchPanelFwUpdateInfo_t *plinfo;

		plinfo = (struct TouchPanelFwUpdateInfo_t *)kzalloc(sizeof(struct TouchPanelFwUpdateInfo_t),
				GFP_KERNEL);
		if (plinfo == NULL) {
			ATMEL_ERR("memory allocation failed\n");
			ret = -1;
			break;
		}

		if(data->info == NULL){
			pr_err("data->info is NULL\n");
			ret = -EAGAIN;
			kfree(plinfo);
			break;
		}
		ret = __mxt_read_reg(data->client,
					data->T38_address + MXT_T38_VENDOR_VER,
					sizeof(tp_cfg), &tp_cfg);
		if (ret) {
			ATMEL_ERR("Unable to read config id\n");
			ret = -1;
		}

		if(tp_cfg != 0x0A){
			strncpy(plinfo->touch_name,old_touch,sizeof(old_touch));
			ATMEL_MSG("touch_name:%s\n",plinfo->touch_name);
		} else {
			strncpy(plinfo->touch_name,new_touch,sizeof(new_touch));
			ATMEL_MSG("touch_name:%s\n",plinfo->touch_name);
		}

		ret = copy_to_user((void*)arg, plinfo, sizeof(*plinfo));
		if (ret != 0) {
			ATMEL_ERR("copy_to_user failed\n");
			ret = -1;
		}
		kfree(plinfo);
		break;
	}
	}

	return ret;
}

static int print_t25_message(struct mxt_data *data, char *page, size_t size)
{
	int len = 0;
	int i;
	int num_data = sizeof(data->T25_result) / sizeof(data->T25_result[0]);
	int num_str_per_data = 3;

	ATMEL_TRACE();

	for (i = 0; i < num_data; i++) {
		if (len + num_str_per_data > size) {
			/* buffer will overflow at this iteration */
			break;
		}
		len += sprintf((page + len), "%02X%c",
			       data->T25_result[i],
			       (i == (num_data - 1)) ? '\n' : ' ');
	}

	return len;
}

static ssize_t tpif_read_pin_fault_test(struct file *file,
					char __user *buf, size_t count,
					loff_t *off)
{
	struct mxt_data *pdata = mTpCmd.pdata;
	struct i2c_client *client = pdata->client;
	struct mxt_object *object;
	int len = 0;
	int error;
	u8 cmd;
	char page[64];

	ATMEL_TRACE();

	if (*off > 0) {
		return 0;
	}

	if (count < sizeof(page)) {
		return 0;
	}

	memset(pdata->T25_result, 0, sizeof(pdata->T25_result));
	reinit_completion(&pdata->selftest_completion);

	object = mxt_get_object(pdata, MXT_SPT_SELFTEST_T25);
	if (!object) {
		len = -EINVAL;
		goto selftest_end;
	}

	/* run selftest */
	cmd = MXT_T25_SELFTEST_PIN_FAULT;
	error = __mxt_write_reg(client,
				object->start_address + MXT_T25_CMD,
				sizeof(cmd), &cmd);
	if (error != 0) {
		len = error;
		goto selftest_end;
	}

	error = mxt_wait_for_completion(pdata, &pdata->selftest_completion,
					MXT_SELFTEST_TIMEOUT);
	if (error != 0) {
		ATMEL_ERR("selftest timeout\n");
	}

	memset(page, 0, sizeof(page));
	len = print_t25_message(pdata, page, sizeof(page));
	if (len > (sizeof(page) - 1)) {
		len = sizeof(page) - 1;
	}
	page[len] = '\0';
	if (copy_to_user(buf, page, len) != 0) {
		len = -EFAULT;
		goto selftest_end;
	}

	*off += len;

selftest_end:
	return len;
}

static ssize_t tpif_write_pin_fault_test(struct file *file,
					 const char __user *buf, size_t count,
					 loff_t *off)
{
	ATMEL_TRACE();

	/* nop */

	return (int)count;
}

static ssize_t tpif_read_signal_limit_test(struct file *file,
					   char __user *buf, size_t count,
					   loff_t *off)
{
	struct mxt_data *pdata = mTpCmd.pdata;
	int len = 0;
	char page[64];

	ATMEL_TRACE();

	if (*off > 0) {
		return 0;
	}

	if (count < sizeof(page)) {
		return 0;
	}

	memset(page, 0, sizeof(page));
	len = print_t25_message(pdata, page, sizeof(page));
	if (len > (sizeof(page) - 1)) {
		len = sizeof(page) - 1;
	}
	page[len] = '\0';
	if (copy_to_user(buf, page, len) != 0) {
		return -EFAULT;
	}

	*off += len;

	return len;
}

static ssize_t tpif_write_signal_limit_test(struct file *file,
					    const char __user *buf, size_t count,
					    loff_t *off)
{
	struct mxt_data *pdata = mTpCmd.pdata;
	struct i2c_client *client = pdata->client;
	struct mxt_object *object;
	int len = 0;
	int error;
	u8 cmd;
	unsigned int upper_limit[2], lower_limit[2], range[2]; /* index 0=LSB, 1=MSB */
	u8 param[2]; /* index 0=LSB, 1=MSB */
	char buffer[128];

	ATMEL_TRACE();

	if (buf == NULL) {
		return -EINVAL;
	}

	if (count > sizeof(buffer) - 1) {
		return -EINVAL;
	}

	memset(buffer, 0, sizeof(buffer));
	if (copy_from_user(buffer, buf, count) != 0) {
		return -EFAULT;
	}

	len = sscanf(buffer, "%x %x %x %x %x %x",
		     &upper_limit[0], &upper_limit[1],
		     &lower_limit[0], &lower_limit[1],
		     &range[0], &range[1]);
	if (len != 6) {
		ATMEL_ERR("cannot parse input data\n");
		goto selftest_end;
	}

	ATMEL_DBG("selftest signal limit parameters: "
		  "%02X %02X %02X %02X %02X %02X\n",
		  upper_limit[0], upper_limit[1],
		  lower_limit[0], lower_limit[1],
		  range[0], range[1]);

	memset(pdata->T25_result, 0, sizeof(pdata->T25_result));
	reinit_completion(&pdata->selftest_completion);

	object = mxt_get_object(pdata, MXT_SPT_SELFTEST_T25);
	if (!object) {
		len = -EINVAL;
		goto selftest_end;
	}

	/* setup signal limit test */
	param[0] = (u8)upper_limit[0];
	param[1] = (u8)upper_limit[1];
	error = __mxt_write_reg(client,
				object->start_address + MXT_T25_SIGNAL_UPPER_LSB,
				sizeof(param), param);
	if (error != 0) {
		ATMEL_ERR("signal upper limit write failed\n");
		goto selftest_end;
	}

	param[0] = (u8)lower_limit[0];
	param[1] = (u8)lower_limit[1];
	error = __mxt_write_reg(client,
				object->start_address + MXT_T25_SIGNAL_LOWER_LSB,
				sizeof(param), param);
	if (error != 0) {
		ATMEL_ERR("signal lower limit write failed\n");
		goto selftest_end;
	}

	param[0] = (u8)range[0];
	param[1] = (u8)range[1];
	error = __mxt_write_reg(client,
				object->start_address + MXT_T25_SIGNAL_RANGE_LSB,
				sizeof(param), param);
	if (error != 0) {
		ATMEL_ERR("signal range write failed\n");
		goto selftest_end;
	}

	/* confirmation */
	(void)__mxt_read_reg(client,
			     object->start_address + MXT_T25_SIGNAL_UPPER_LSB,
			     sizeof(param), param);
	upper_limit[0] = param[0];
	upper_limit[1] = param[1];

	(void)__mxt_read_reg(client,
			     object->start_address + MXT_T25_SIGNAL_LOWER_LSB,
			     sizeof(param), param);
	lower_limit[0] = param[0];
	lower_limit[1] = param[1];

	(void)__mxt_read_reg(client,
			     object->start_address + MXT_T25_SIGNAL_RANGE_LSB,
			     sizeof(param), param);
	range[0] = param[0];
	range[1] = param[1];

	ATMEL_DBG("selftest signal limit parameters (confirmation): "
		  "%02X %02X %02X %02X %02X %02X\n",
		  upper_limit[0], upper_limit[1],
		  lower_limit[0], lower_limit[1],
		  range[0], range[1]);

	/* run selftest */
	cmd = MXT_T25_SELFTEST_SIGNAL_LIMIT;
	error = __mxt_write_reg(client,
				object->start_address + MXT_T25_CMD,
				sizeof(cmd), &cmd);
	if (error != 0) {
		len = error;
		goto selftest_end;
	}

	error = mxt_wait_for_completion(pdata, &pdata->selftest_completion,
					MXT_SELFTEST_TIMEOUT);
	if (error != 0) {
		ATMEL_ERR("selftest timeout\n");
	}

selftest_end:
	return (int)count;
}


static const struct dev_pm_ops mxt_pm_ops = {
#if !defined(CONFIG_FB)
	.suspend	= mxt_suspend,
	.resume		= mxt_resume,
#endif
};
#endif

static void mxt_shutdown(struct i2c_client *client)
{
	struct mxt_data *data = i2c_get_clientdata(client);

	disable_irq(data->irq);
}

static const struct i2c_device_id mxt_id[] = {
	{ "qt602240_ts", 0 },
	{ "atmel_mxt_ts", 0 },
	{ "atmel_mxt_tp", 0 },
	{ "mXT224", 0 },
	{ }
};
MODULE_DEVICE_TABLE(i2c, mxt_id);

#ifdef CONFIG_OF
static struct of_device_id mxt_match_table[] = {
	{ .compatible = "atmel,mxt-ts",},
	{ },
};
#else
#define mxt_match_table NULL
#endif

static struct i2c_driver mxt_driver = {
	.driver = {
		.name	= DRIVER_NAME,
		.owner	= THIS_MODULE,
		.of_match_table = mxt_match_table,
#ifdef CONFIG_PM
		.pm	= &mxt_pm_ops,
#endif
	},
	.probe		= mxt_probe,
	.remove		= mxt_remove,
	.shutdown	= mxt_shutdown,
	.id_table	= mxt_id,
};

module_i2c_driver(mxt_driver);

/* Module information */
MODULE_AUTHOR("Joonyoung Shim <jy0922.shim@samsung.com>");
MODULE_DESCRIPTION("Atmel maXTouch Touchscreen driver");
MODULE_LICENSE("GPL");
